/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

/**
 * Base class for every type of attribute.
 */
public abstract class Attribute
{
    /**
     * Empty array of attributes.
     */
    public static final Attribute[] EMPTY = new Attribute[0];

    /**
     * Constructs an attribute from a class file stream.
     *
     * @param in
     * The stream to read from.
     * @param pool
     * The constant pool.
     * @exception IOException
     * An io problem has occured
     * @exception ClassFileFormatException
     * Attempt to read from a bad .class file.
     */
    public static Attribute read(DataInput in, ConstantPool pool)
        throws IOException, ClassFileFormatException
    {
        int nameIndex = in.readUnsignedShort();
        UTF8Constant name = (UTF8Constant)pool.getEntryAt(nameIndex);

        String tag = name.getValue();

        switch (tag.charAt(0)) {
        case 'C':
            if (tag.equals("Code")) {
                return new CodeInfo(in, pool);
            }
            else if (tag.equals("ConstantValue")) {
                return new ConstantValueAttribute(in, pool);
            }
            break;

        case 'D':
            if (tag.equals("Deprecated")) {
                return new DeprecatedAttribute(in, pool);
            }
            break;

        case 'E':
            if (tag.equals("Exceptions")) {
                return new ExceptionsAttribute(in, pool);
            }
            break;

        case 'I':
            if (tag.equals("InnerClasses")) {
                return new InnerClassTable(in, pool);
            }
            break;

        case 'L':
            if (tag.equals("LineNumberTable")) {
                throw new ClassFileFormatException
                    ("Attribute \"LineNumberTable\" illegal outside " +
                     "of Code Attribute");
            }
            else if (tag.equals("LocalVariableTable")) {
                throw new ClassFileFormatException
                    ("Attribute \"LocalVariableTable\" illegal outside " +
                     "of Code Attribute");
            }
            break;

        case 'S':
            if (tag.equals("SourceFile")) {
                return new SourceFileAttribute(in, pool);
            }
            else if (tag.equals("Synthetic")) {
                return new SyntheticAttribute(in, pool);
            }
            break;

        default:
            break;
        }

        return new GenericAttribute(name, in, pool);
    }

    /**
     * Constructs an attribute from a class file stream.
     *
     * @param in
     * The stream to read from.
     * @param pool
     * The constant pool.
     * @exception IOException
     * An io problem has occured
     * @exception ClassFileFormatException
     * Attempt to read from a bad .class file.
     */
    public static Attribute readInterfaceOnly(DataInput in, ConstantPool pool)
        throws IOException, ClassFileFormatException
    {
        int nameIndex = in.readUnsignedShort();
        UTF8Constant name = (UTF8Constant)pool.getEntryAt(nameIndex);
        String tag = name.getValue();

        switch (tag.charAt(0)) {
        case 'C':
            if (tag.equals("Code")) {
                return new SkippedCodeInfo(in, pool);
            }
            else if (tag.equals("ConstantValue")) {
                return new ConstantValueAttribute(in, pool);
            }
            break;

        case 'D':
            if (tag.equals("Deprecated")) {
                return new DeprecatedAttribute(in, pool);
            }
            break;

        case 'E':
            if (tag.equals("Exceptions")) {
                return new ExceptionsAttribute(in, pool);
            }
            break;

        case 'I':
            if (tag.equals("InnerClasses")) {
                return new InnerClassTable(in, pool);
            }
            break;

        case 'L':
            if (tag.equals("LineNumberTable")) {
                throw new ClassFileFormatException
                    ("Attribute \"LineNumberTable\" illegal outside of "
                     + " Code Attribute");
            }
            else if (tag.equals("LocalVariableTable")) {
                throw new ClassFileFormatException
                    ("Attribute \"LocalVariableTable\" illegal outside of "
                     + "Code Attribute");
            }
            break;

        case 'S':
            if (tag.equals("SourceFile")) {
                return new SourceFileAttribute(in, pool);
            }
            else if (tag.equals("Synthetic")) {
                return new SyntheticAttribute(in, pool);
            }
            break;

        default:
            break;
        }

        return new GenericAttribute(name, in, pool);
    }

    /**
     * Constructs an sub-attribute of CodeInfo from a class file stream.
     *
     * @param in
     * The stream to read from.
     * @param pool
     * The constant pool.
     * @param instructions
     * (sparse) array of instructions
     *
     * @exception IOException
     * An io problem has occured.
     * @exception ClassFileFormatException
     * Attempt to read from a bad .class file.
     */
    public static Attribute readCodeInfoAttribute(DataInput in,
                                                  ConstantPool pool,
                                                  Instruction[] instructions)
        throws IOException, ClassFileFormatException
    {
        int nameIndex = in.readUnsignedShort();
        UTF8Constant name = (UTF8Constant)pool.getEntryAt(nameIndex);
        String tag = name.getValue();

        switch (tag.charAt(0)) {
        case 'C':
            if (tag.equals("Code")) {
                throw new ClassFileFormatException
                    ("Attribute \"Code\" illegal as sub-attribute of "
                     + "Code Attribute");
            }
            else if (tag.equals("ConstantValue")) {
                throw new ClassFileFormatException
                    ("Attribute \"ConstantValue\" illegal as sub-attribute "
                     + "of Code Attribute");
            }
            break;

        case 'D':
            if (tag.equals("Deprecated")) {
                throw new ClassFileFormatException
                    ("Attribute \"Deprecated\" illegal as sub-attribute "
                     + "of Code Attribute");
            }
            break;

        case 'E':
            if (tag.equals("Exceptions")) {
                throw new ClassFileFormatException
                    ("Attribute \"Exceptions\" illegal as sub-attribute of "
                     + "Code Attribute");
            }
            break;

        case 'I':
            if (tag.equals("InnerClasses")) {
                throw new ClassFileFormatException
                    ("Attribute \"InnerClasses\" illegal as sub-attribute "
                     + "of Code Attribute");
            }
            break;

        case 'L':
            if (tag.equals("LineNumberTable")) {
                return new LineNumberTable(in, pool, instructions);
            }
            else if (tag.equals("LocalVariableTable")) {
                return new LocalVariableTable(in, pool, instructions);
            }
            break;

        case 'S':
            if (tag.equals("SourceFile")) {
                throw new ClassFileFormatException
                    ("Attribute \"SourceFile\" illegal as sub-attribute "
                     + "of Code Attribute");
            }
            else if (tag.equals("Synthetic")) {
                throw new ClassFileFormatException
                    ("Attribute \"Synthetic\" illegal as sub-attribute of "
                     + "Code Attribute");
            }
            break;

        default:
            break;
        }

        return new GenericAttribute(name, in, pool);
    }

    /**
     * Returns the attribute's tag
     */
    public abstract int getTag();

    /**
     * Returns the space in bytes used by this attribute in the classfile
     */
    public abstract int getSize();

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param pool
     * The constant pool for this class.
     */
    public abstract void resolveConstants(ConstantPool pool)
        throws ClassFileFormatException;

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool.
     *
     * @param pool
     * The constant pool that contain all data.
     * @param out
     * The file where to write this object info.
     *
     * @exception IOException
     * An io problem has occured.
     * @exception ClassFileFormatException
     * Attempt to write a bad classfile info.
     */
    public abstract void write(ConstantPool pool, DataOutput out)
        throws IOException, ClassFileFormatException;
}

