/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

/**
 * Exceptions Attribute. See VMS 4.7.5.
 *
 * This attribute indicates which checked exceptions a method may throw.
 */
public class ExceptionsAttribute
    extends Attribute
{
    /**
     * Name for this type of attribute.
     */
    private static UTF8Constant attr = new UTF8Constant("Exceptions");

    /**
     * The exceptions in this attribute.
     */
    private ClassConstant[] exceptions;

    /**
     * Create an exceptions attribute.
     */
    public ExceptionsAttribute(String[] names)
    {
        exceptions = new ClassConstant[names.length];

        for (int i = 0; i < exceptions.length; i++) {
            exceptions[i] = new ClassConstant(names[i]);
        }
    }

    /**
     * Constructs a exceptions attribute from a class file stream.
     *
     * @param in
     * The stream to read from.
     * @param pool
     * The constant pool.
     * @exception IOException
     * An io problem has occured.
     */
    public ExceptionsAttribute(DataInput in, ConstantPool pool)
        throws IOException
    {
        // ignore the length
        in.readInt();

        exceptions = new ClassConstant[in.readUnsignedShort()];
        for (int i = 0; i < exceptions.length; i++) {
            exceptions[i] =
                (ClassConstant)pool.getEntryAt(in.readUnsignedShort());
        }
    }

    /**
     * Returns the attribute's tag.
     */
    public int getTag()
    {
        return ClassFileConstants.ATT_EXCEPTIONS;
    }

    /**
     * Returns the space in bytes used by this attribute in the classfile
     */
    public int getSize()
    {
        return 2 + 4 + 2 + 2*exceptions.length;
    }

    /**
     * Returns the exceptions.
     */
    public String[] getExceptions()
    {
        String[] names = new String[exceptions.length];

        for (int i = 0; i < exceptions.length; i++) {
            names[i] = exceptions[i].getName();
        }

        return names;
    }

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param pool
     * The constant pool for this class.
     */
    public void resolveConstants(ConstantPool pool)
    {
        pool.addItem(attr);

        for (int i = 0; i < exceptions.length; i++) {
            pool.addItem(exceptions[i]);
        }
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool.
     *
     * @param pool
     * The constant pool that contain all data.
     * @param out
     * The file where to write this object info.
     *
     * @exception IOException 
     * An io problem has occured.
     */
    public void write(ConstantPool pool, DataOutput out)
        throws IOException
    {
        out.writeShort(attr.getIndex());
        out.writeInt(exceptions.length*2 + 2);
        out.writeShort(exceptions.length);
        for (int i = 0; i < exceptions.length; i++) {
            out.writeShort(exceptions[i].getIndex());
        }
    }
}
