/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

/**
 * Line Number Table Attribute. See VMS 4.7.6
 *
 * This attribute represents a mapping between he Java Virtual Machine code
 * array and the line number in the original Java source file.
 */
public class LineNumberTable
    extends Attribute
{
    /**
     * ...
     */
    private static UTF8Constant attr = new UTF8Constant("LineNumberTable");

    /**
     * ...
     */
    private LineNumberInfo[] entries;

    /**
     * Create a line number table attribute.
     */
    public LineNumberTable(LineNumberInfo[] entries)
    {
        this.entries = entries;
    }

    /**
     * Constructs a line number table attribute from a class file stream.
     *
     * @param in
     * The stream to read from.
     * @param pool
     * The constant pool.
     * @param instructions
     * (sparse) array of instructions.
     * @exception IOException
     * An io problem has occured.
     */
    public LineNumberTable(DataInput in,
                           ConstantPool pool,
                           Instruction[] instructions)
        throws IOException
    {
        in.readInt();   // ignore

        this.entries = new LineNumberInfo[in.readUnsignedShort()];
        for (int i = 0; i < this.entries.length; i++) {
            this.entries[i] = new LineNumberInfo(in, pool, instructions);
        }
    }

    /**
     * Returns the attribute's tag.
     */
    public int getTag()
    {
        return ClassFileConstants.ATT_LINENUMBERTABLE;
    }

    /**
     * Returns the space in bytes used by this attribute in the classfile.
     */
    public int getSize()
    {
        return 2 + 4 + 2 + 4*entries.length;
    }

    /**
     * Returns line number information.
     */
    public LineNumberInfo[] getLineNumbers()
    {
        return entries;
    }

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param pool
     * The constant pool for this class.
     */
    public void resolveConstants(ConstantPool pool)
    {
        pool.addItem(attr);
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool.
     *
     * @param pool
     * The constant pool that contain all data.
     * @param out
     * The file where to write this object info.
     * @exception IOException
     * An io problem has occured.
     */
    public void write(ConstantPool pool, DataOutput out)
        throws IOException
    {
        out.writeShort(attr.getIndex());
        out.writeInt(2 + 4 * entries.length);
        out.writeShort(entries.length);
        for (int i = 0; i < entries.length; i++) {
            entries[i].write(out);
        }
    }
}

