/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.classfile;

import java.io.DataOutput;
import java.io.IOException;

import at.dms.util.InternalError;

/**
 * This class represents instructions that take a local variable as argument.
 *
 * Note that an extra wide prefix is automatically added for these
 * instructions if the numeric argument is larger than 255.
 */
public class NewarrayInstruction
    extends Instruction
{
    /**
     * Array type.
     */
    private byte type;

    /**
     * Constructs a new instruction that takes a local variable as argument.
     *
     * @param type
     * The code of the base type.
     */
    public NewarrayInstruction(byte type)
    {
        super(opc_newarray);

        this.type = type;
    }

    /**
     * Returns true iff control flow can reach the next instruction
     * in textual order.
     */
    public boolean canComplete()
    {
        return true;
    }

    /**
     * Returns the number of bytes used by the the instruction in the code
     * array.
     */
    public int getSize()
    {
        return 1 + 1;
    }

    /**
     * Return the type of this array.
     */
    public String getType()
    {
        switch (type) {
        case 4:
            return "boolean";
        case 5:
            return "char";
        case 6:
            return "float";
        case 7:
            return "double";
        case 8:
            return "byte";
        case 9:
            return "short";
        case 10:
            return "int";
        case 11:
            return "long";
        default:
            throw new InternalError("invalid type code " + type);
        }
    }

    /**
     * Returns the type pushed on the stack.
     */
    public byte getReturnType()
    {
        return TYP_REFERENCE;
    }

    /**
     * Return the amount of stack (positive or negative) used by this
     * instruction.
     */
    public int getStack()
    {
        return 0;
    }

    /**
     * Returns the size of data pushed on the stack by this instruction.
     */
    public int getPushedOnStack()
    {
        return 1;
    }

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param pool      the constant pool for this class.
     */
    void resolveConstants(ConstantPool pool)
    {
    }

    /**
     * Write this instruction into a file.
     *
     * @param pool      the constant pool that contain all data.
     * @param out     the file where to write this object info.
     *
     * @exception IOException
     * An io problem has occured.
     */
    public void write(ConstantPool pool, DataOutput out)
        throws IOException
    {
        out.writeByte((byte)getOpcode());
        out.writeByte(type);
    }
}

