/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.backend;

import java.util.ArrayList;
import java.util.List;

import at.dms.util.Utils;
import at.dms.classfile.SwitchInstruction;
import at.dms.classfile.HandlerInfo;
import at.dms.backend.InstructionHandle;

/**
 * This is the entry point of the backend, this class constructs the
 * control flow graf and applies optimizations
 */
public class ControlFlow
{
    private MethodEnv env;
    private BasicBlock[] bblocks; // List of basic blocks
    private BasicBlock[] eblocks; // List of exception handler

    /**
     * Creates a new instruction handle.
     *
     * @param instruction the instruction.
     * @param prev the handle of the next instruction in textual order.
     */
    public ControlFlow(MethodEnv env,
                       InstructionHandle start, HandlerInfo[] handlers)
    {
        this.env = env;

        buildBasicBlocks(start, handlers);

        setMarked(false);
        bblocks[0].buildQuadruples(env);
    }

    /**
     * Prints a trace of the control flow structure.
     */
    public void trace()
    {
        new TraceControlFlow(bblocks, eblocks).run();
    }

    public void optimize()
    {
        // HIGH LEVEL OPTIMIZATION

        // LIVENESS ANALYSIS
        LivenessAnalysis    live = new LivenessAnalysis(bblocks, eblocks);

        live.run();
        trace();

        // CSE, COPY PROPAGATION, ....
        new DeadcodeElimination(bblocks, eblocks).run();
        new StackScheduler(bblocks, eblocks).run();

        // REGISTER ALLOCATION    
        RegisterAllocation  alloc =
            new RegisterAllocation(env, bblocks, eblocks, live);

        alloc.run();
        new TraceInferenceGraph(alloc.getInferenceGraph()).run();
    }

    public InstructionHandle getInstructions() {
        setMarked(false);

        CodeSequence    seq = new CodeSequence();
    
        // Goes all over the tree
        seq.plantBasicBlock(bblocks[0]);
        seq.close();

        // Generates exceptions
        for (int i = 0; i < eblocks.length; i++) {
            seq.plantBasicBlock(eblocks[i]);
        }

        resolveAccessors(seq.getCodeStart());
        for (int i = 0; i < eblocks.length; i++) {
            resolveAccessors(eblocks[i].getFirstInstruction());
        }

        return seq.getCodeStart();
    }

    public static BasicBlock findBasicBlock(InstructionHandle handle)
    {
        for (int pos = 0; ; pos++) {
            if (handle.getAccessor(pos) instanceof BasicBlock) {
                return (BasicBlock)handle.getAccessor(pos);
            }
        }
    }

    private void buildBasicBlocks(InstructionHandle start,
                                  HandlerInfo[] handlers)
    {
        List  bblocks = new ArrayList();
        List  body = new ArrayList();
        BasicBlock  current = null;
        boolean startBB = true;
        int     countBB = 0;

        for ( InstructionHandle handle = start;
              handle != null;
              handle = handle.getNext()) {

            if (handle.isTarget() || startBB) {
                // new basic block
                if (current != null && !current.isMarked()) {
                    closeBasicBlock(current, body, bblocks);
                }
                current = new BasicBlock(countBB++);

                handle.addAccessor(current);
                body.add(handle);
                startBB = isEndOfBasicBlock(handle);
            }
            else if (isEndOfBasicBlock(handle)) {
                // end of basic block
                body.add(handle);
                closeBasicBlock(current, body, bblocks);

                startBB = true;
            }
            else {
                // add to basic block
                body.add(handle);
            }
        }
        if (body.size() != 0) {
            closeBasicBlock(current, body, bblocks);
        }

        clean(handlers, bblocks);
    }

    private void closeBasicBlock(BasicBlock current, List body, List bblocks)
    {
        current.setBody((InstructionHandle[])
                        body.toArray(new InstructionHandle[body.size()]));

        body.clear();
        bblocks.add(current);
        current.setMarked(true);
    }

    private void clean(HandlerInfo[] handlers, List vbblocks)
    {
        bblocks = (BasicBlock[])
            vbblocks.toArray(new BasicBlock[vbblocks.size()]);

        for (int i = 0; i < bblocks.length; i++ ) {
            bblocks[i].resolveJump();
        }

        // Handlers
        eblocks = new BasicBlock[handlers.length];
        for (int i  = 0; i < handlers.length; i++) {
            eblocks[i] = findBasicBlock((InstructionHandle)handlers[i]
                                        .getHandler());
        }

        // Clean all
        for ( InstructionHandle handle = bblocks[0].getFirstInstruction();
              handle != null;) {

            InstructionHandle last = handle;

            handle.removeAccessors();
            handle = handle.getNext();
            last.setNext(null);
        }
    }
 
    private static boolean isEndOfBasicBlock(InstructionHandle handle)
    {
        return handle.isJump() || !handle.getInstruction().canComplete();
    }

    private void setMarked(boolean marked)
    {
        setMarked(bblocks, marked);
    }

    protected final static void setMarked(BasicBlock[] bblocks, boolean marked)
    {
        for (int i = 0; i < bblocks.length; i++) {
            bblocks[i].setMarked(marked);
        }
    }

    private void resolveAccessors(InstructionHandle handlex)
    {
        for (; handlex != null; handlex = handlex.getNext()) {
            if (handlex.isJump()) {
                handlex.setTarget(((BasicBlock)handlex.getJump()
                                   .getTarget()).getFirstInstruction());
            }
            else if (handlex.getInstruction() instanceof SwitchInstruction) {
                SwitchInstruction instruction =
                    (SwitchInstruction)handlex.getInstruction();
    
                for (int i = -1; i < instruction.getSwitchCount(); i++) {
                    instruction.setTarget
                        (((BasicBlock)instruction
                          .getTarget(i)).getFirstInstruction(), i);
                }
            }
        }
    }
}
