/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.backend;

import java.util.ArrayList;
import java.util.List;
import java.util.LinkedList;

import at.dms.classfile.ClassFileConstants;
import at.dms.classfile.IincInstruction;
import at.dms.classfile.LocalVarInstruction;
import at.dms.classfile.PushLiteralInstruction;

/**
 * This class represents a quadruple: a source and a destinationination.
 */
class QQuadruple
    extends QNode
    implements QOrigin
{
    /**
     *
     */
    private QDestination destination;

    /**
     *
     */
    private QOrigin origin;

    public QQuadruple(QDestination destination, QOrigin origin)
    {
        this.destination = destination;
        this.origin = origin;    
    }

    /**
     * isStore
     */
    public boolean isStore()
    {
        return true;
    }

    /**
     * Returns this node.
     */
    public QQuadruple getStore()
    {
        return this;
    }

    /**
     * Human readable form.
     */
    public String toString()
    {
        return destination + " <- " + origin;
    }

    /**
     * The type of this instruction.
     */
    public int getType()
    {
        return ClassFileConstants.TYP_VOID;
    }

    /**
     * Duplicate this node: not implemented.
     */
    public QOrigin duplicate()
    {
        throw new RuntimeException("NYI");
    }

    /**
     * Uses the stack instead of a local var.
     */
    public void useStack()
    {
        destination = new QStack(((QTemporary)destination).getType());
    }

    /**
     * returns the parameters of this instruction
     */
    public QOrigin[] getOrigins()
    {
        return origin.getOrigins();
    }

    /**
     * Sets the parameters of this instruction.
     */
    public void setOrigin(QOrigin origin, int i)
    {
        if (this.origin instanceof QOperator) {
            this.origin.setOrigin(origin, i);
        }
        else if (i == 0) {
            this.origin = origin;
        }
        else {
            throw new RuntimeException();
        }
    }

    /**
     * Returns the defined temporary.
     */
    public QTemporary getDef()
    {
        return (QTemporary)destination;
    }

    /**
     * Returns the used temporaries.
     */
    public QTemporary[] getUses() {
        return origin.getUses();
    }

    /**
     * Generates instructions for this quadruple.
     *
     * @param seq
     * The code sequence of instruction.
     */
    public void generate(CodeSequence seq)
    {
        origin.generate(seq);
        if (destination instanceof QTemporary) {
            destination.store(seq, isLive((QTemporary)destination));
        }
        else {
            destination.store(seq, true);
        }
    }

    /**
     * Builds the quadruple list.
     *
     * @param instructions
     * The list of instructions.
     *
     * @param entryStack
     * The stack in entry.
     *
     * @param list
     * A place holder for quadruples.
     *
     * @return
     * The stack at the end of this basic block
     */
    public static QTemporary[] buildQuadruples(MethodEnv env,
                                               BasicBlock block,
                                               InstructionHandle[] instructions, 
                                               QTemporary[] entryStack,
                                               List list)
    {
        LinkedList stack = new LinkedList();

        for (int i = 0; entryStack != null && i < entryStack.length; i++) {
            stack.addFirst(entryStack[i]);
        }

        for (int i = 0; i < instructions.length; i++) {
            InstructionHandle instruction = instructions[i];
            //instruction.getInstruction().dump();

            if (instruction.getInstruction() instanceof LocalVarInstruction) {
                if (instruction.getLocal().isLoad()) {
                    QTemporary var = env.getLocalVar(instruction);

                    stack.addFirst(var);
                }
                else {
                    QOrigin value = (QOrigin)stack.removeFirst();
                    QTemporary var = env.getLocalVar(instruction);
      
                    list.add(new QQuadruple(var, value));
                }
            }
            else if ( instruction.getInstruction()
                      instanceof PushLiteralInstruction) {
                stack.addFirst(new QLiteral(instruction));
            }
            else if (instruction.isPop()) {
                list.add(new QQuadruple(new QTemporary(1),
                                        (QOrigin)stack.removeFirst()));
            }
            else if (instruction.isDup()) {
                Object  temp1 = stack.removeFirst();

                stack.addFirst(temp1);
                stack.addFirst(((QOrigin)temp1).duplicate());   
            }
            else if (instruction.isDupX2()) {
                Object temp1 = stack.removeFirst();
                Object temp2 = stack.removeFirst();

                stack.addFirst(temp2);
                stack.addFirst(temp1);  
                stack.addFirst(((QOrigin)temp2).duplicate());
            }
            else if (instruction.isSwap()) {
                Object  temp1 = stack.removeFirst();
                Object  temp2 = stack.removeFirst();

                stack.addFirst(temp1);
                stack.addFirst(temp2);
            }
            else if (instruction.getInstruction() instanceof IincInstruction) {
                IincInstruction increment = (IincInstruction)
                    instruction.getInstruction();

                list.add(new QIinc(instruction.getInstruction(),
                                   env.getLocalVar(increment.getVariable())));
            }
            else {
                // OPER
                QOperand[] operands =
                    readOperands(stack, instruction.countOperands());
    
                if (instruction.isJump()) {
                    list.add(new QJump(new QOperator(instruction, operands)));
                }
                else if (!instruction.hasReturnValue()) {
                    list.add(new QVoid(new QOperator(instruction, operands)));
                }
                else if (i != instructions.length - 1
                         && ( instructions[i+1].getInstruction()
                              instanceof LocalVarInstruction )
                         && instructions[i+1].getLocal().isStore()) {

                    QOperator oper = new QOperator(instruction, operands);
                    QTemporary temp = env.getLocalVar(instructions[i + 1]);
        
                    list.add(new QQuadruple(temp, oper));
                    i++;
                }
                else {
                    QOperator oper = new QOperator(instruction, operands);
                    QTemporary temp = new QTemporary(oper.getType());
                    
                    list.add(new QQuadruple(temp, oper));
                    stack.addFirst(temp);
                }
            }
        }

        int pos = 0;
        QNode last = ( list.size() == 0
                       ? null : (QNode)list.get(list.size() - 1) );

        if (last instanceof QJump) {
            list.remove(list.size() - 1);
        }

        QTemporary[] remainingLinkedList = new QTemporary[stack.size()];
        while (stack.size() > 0) {
            QOrigin value = (QOrigin)stack.removeFirst();
      
            if (value instanceof QTemporary) {
                remainingLinkedList[pos++] = (QTemporary)value;
            }
            else {
                remainingLinkedList[pos] = new QTemporary(value.getType());
                list.add(new QQuadruple(remainingLinkedList[pos++], value));
            }
        }
        if (last instanceof QJump) {
            list.add(last);
        }

        return remainingLinkedList;
    }

    private static QOperand[] readOperands(LinkedList stack, int size)
    {
        List list = new ArrayList();

        for (int i = 0; size > 0; i++) {
            QOperand oper = (QOperand)stack.removeFirst();

            size -= InstructionHandle.getTypeSize(oper.getType());
            list.add(oper);
        }

        return (QOperand[])list.toArray(new QOperand[list.size()]);
    }
}
