/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.backend;

import at.dms.classfile.ClassFileConstants;
import at.dms.classfile.LocalVarInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.Opcodes;

/**
 * This class represents a temporary.
 */
class QTemporary
    extends QOperand
    implements QDestination, ClassFileConstants, Opcodes
{
    /**
     * Constant for UNINITIALIZED.
     */
    public static final int UNINITIALIZED = -1;

    /**
     * Constant for UNUSED.
     */
    public static final int UNUSED = -2;

    /**
     * An empty array of temporaries.
     */
    public static final QTemporary[] EMPTY = new QTemporary[0];

    private static int instanceCount = 0;

    private int type;
    private BasicBlock defBlock;
    private int defPos;
    private BasicBlock useBlock;
    private int usePos;
    private int countUses;

    private int register = UNINITIALIZED;

    private int precolor = Integer.MAX_VALUE;
    private boolean usePrecolor;
  
    private int pos;

    QTemporary(int type)
    {
        this.type = type;
        /* For trace */
        this.pos = instanceCount++;
    }

    QTemporary(int type, int precolor)
    {
        this(type);
        this.precolor = precolor; // is it a parameter, we will see after...
    }

    /**
     * Returns the defined temporary.
     */
    public QTemporary getDef()
    {
        return null;
    }

    /**
     * Returns the used temporaries.
     */
    public QTemporary[] getUses() {
        return new QTemporary[]{this};
    }

    /**
     * Returns the size (the number of slot) of this temporary.
     */
    public int getSize()
    {
        return InstructionHandle.getTypeSize(type);
    }

    /**
     * Returns the register number of this temporary.
     */
    public int getRegister()
    {
        return register;
    }

    /**
     * Returns the pre-color of this temporary.
     */
    public int getPrecolor()
    {
        if (usePrecolor && precolor != Integer.MAX_VALUE) {
            return precolor;
        }
        else {
            return UNINITIALIZED;
        }
    }

    /**
     * Enforces the precolor.
     */
    public void enforceColor()
    {
        usePrecolor = true;
    }

    /**
     * The type of the temporary.
     */
    public int getType()
    {
        return type;
    }

    /**
     * Human readable form.
     */
    public String toString()
    {
        int name = ( (register == UNINITIALIZED) ? pos : register );

        if (precolor != Integer.MAX_VALUE) {
            if (usePrecolor) {
                return "P_" + precolor;
            }
            else {
                return "L_" + precolor;
            }
        }
        else {
            return "t_" + name;
        }
    }

    /**
     * Duplicate this node.
     */
    public QOrigin duplicate()
    {
        return this;
    }

    /**
     * Uses a temporary.
     */
    public void use(BasicBlock block, int pos)
    {
        useBlock = block;
        usePos = pos;
        countUses++;
    }

    /**
     * Defines? a temporary.
     */
    public void def(BasicBlock block, int pos)
    {
        defBlock = block;
        defPos = pos;
    }

    /**
     * Returns the number of uses.
     */
    public int getUseCount()
    {
        return countUses;
    }

    /**
     * Sets position.
     */
    public void setPosition(int pos)
    {
        this.pos = pos;
    }

    /**
     * Gets position.
     */
    public int getPosition()
    {
        return pos;
    }

    /**
     * ???
     */
    public void setRegister(int reg)
    {
        if (countUses == 0) {
            register = UNUSED;
        }
        else {
            register = reg;
        }
    }

    /**
     * Generates instructions for this quadruple.
     *
     * @param seq
     * The code sequence of instruction.
     */
    public void generate(CodeSequence seq)
    {
        int opcode;

        if (register != UNUSED) {
            switch (getType()) {
            case TYP_REFERENCE:
                opcode = Opcodes.opc_aload;
                break;
            case TYP_DOUBLE:
                opcode = Opcodes.opc_dload;
                break;
            case TYP_FLOAT:
                opcode = Opcodes.opc_fload;
                break;
            case TYP_INT:
                opcode = Opcodes.opc_iload;
                break;
            case TYP_LONG:
                opcode = Opcodes.opc_lload;
                break;
            default:
                throw new RuntimeException();
            }
      
            seq.plantInstruction(new LocalVarInstruction(opcode, register));
        }
        else if (countUses == 0) {
            throw new RuntimeException();      
        }
    }

    /**
     * Generates instructions for storing.
     *
     * @param seq
     * The code sequence of instruction.
     */
    public void store(CodeSequence seq, boolean isLive)
    {
        int opcode;

        if (register != UNUSED && isLive) {
            switch (getType()) {
            case TYP_REFERENCE:
                opcode = opc_astore;
                break;
            case TYP_DOUBLE:
                opcode = opc_dstore;
                break;
            case TYP_FLOAT:
                opcode = opc_fstore;
                break;
            case TYP_INT:
                opcode = opc_istore;
                break;
            case TYP_LONG:
                opcode = opc_lstore;
                break;
            default:
                throw new RuntimeException();
            }
      
            seq.plantInstruction(new LocalVarInstruction(opcode, register));
        }
        else if (countUses == 0 || !isLive) {
            int size = InstructionHandle.getTypeSize(type);

            seq.plantInstruction(new NoArgInstruction
                (size == 1 ? opc_pop : opc_pop2));
        }
    }
}
