/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.codegen.jvm;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.Opcodes;
import at.dms.kjc.CodeSequence;
import at.dms.type.NumericType;
import at.dms.type.StandardType;
import at.dms.type.Type;

/**
 * A collection of utility methods to help code generation concerning types.
 *
 * @author Erwin Bolwidt
 */
public class Types
    implements Opcodes
{
    /**
     * Returns the JVM opcode to load a local variable of the specified type.
     *
     * @param type
     * The type of the local variable for which to generate a load instruction.
     */
    public static int getLoadOpcode(Type type)
    {
        switch (type.getTypeID()) {
        case Type.TID_CLASS:
        case Type.TID_ARRAY:
            return opc_aload;
        case Type.TID_DOUBLE:
            return opc_dload;
        case Type.TID_FLOAT:
            return opc_fload;
        case Type.TID_LONG:
            return opc_lload;
        case Type.TID_BYTE:
        case Type.TID_SHORT:
        case Type.TID_CHAR:
        case Type.TID_INT:
        case Type.TID_BOOLEAN:
            return opc_iload;
        default:
            throw new RuntimeException("INTERNAL ERROR: " + type);
        }
    }

    /**
     * Returns the JVM opcode to store a local variable of this type.
     */
    public static int getStoreOpcode(Type type)
    {
        switch (type.getTypeID()) {
        case Type.TID_CLASS:
        case Type.TID_ARRAY:
            return opc_astore;
        case Type.TID_DOUBLE:
            return opc_dstore;
        case Type.TID_FLOAT:
            return opc_fstore;
        case Type.TID_LONG:
            return opc_lstore;
        case Type.TID_BYTE:
        case Type.TID_SHORT:
        case Type.TID_CHAR:
        case Type.TID_INT:
        case Type.TID_BOOLEAN:
            return opc_istore;
        default:
            throw new RuntimeException("INTERNAL ERROR: " + type);
        }
    }

    /**
     * Returns the JVM opcode to return a value of this type.
     */
    public static int getReturnOpcode(Type type)
    {
        switch (type.getTypeID()) {
        case Type.TID_VOID:
            return opc_return;
        case Type.TID_CLASS:
        case Type.TID_ARRAY:
            return opc_areturn;
        case Type.TID_DOUBLE:
            return opc_dreturn;
        case Type.TID_FLOAT:
            return opc_freturn;
        case Type.TID_LONG:
            return opc_lreturn;
        case Type.TID_BYTE:
        case Type.TID_SHORT:
        case Type.TID_CHAR:
        case Type.TID_INT:
        case Type.TID_BOOLEAN:
            return opc_ireturn;
        default:
            throw new RuntimeException("INTERNAL ERROR: " + type);
        }
    }

    /**
     * Generates a sequence of bytecodes that casts the top of the JVM
     * stack from one numeric type to another.
     *
     * @param from
     * The source type.
     * @param to
     * The destination type.
     * @param code
     * The code list.
     */
    public static void genCast(NumericType from, NumericType to,
                               CodeSequence code)
    {
        if (!to.equalsType(from)) {
            switch (to.getTypeID()) {
            case Type.TID_BYTE:
                genCast(from, StandardType.INTEGER, code);
                code.plantInstruction(new NoArgInstruction(opc_i2b));
                break;

            case Type.TID_CHAR:
                genCast(from, StandardType.INTEGER, code);
                code.plantInstruction(new NoArgInstruction(opc_i2c));
                break;

            case Type.TID_SHORT:
                genCast(from, StandardType.INTEGER, code);
                code.plantInstruction(new NoArgInstruction(opc_i2s));
                break;

            case Type.TID_INT:
                switch (from.getTypeID()) {
                case Type.TID_BYTE:
                case Type.TID_CHAR:
                case Type.TID_SHORT:
                case Type.TID_INT:
                    break;
                case Type.TID_LONG:
                    code.plantInstruction(new NoArgInstruction(opc_l2i));
                    break;
                case Type.TID_FLOAT:
                    code.plantInstruction(new NoArgInstruction(opc_f2i));
                    break;
                case Type.TID_DOUBLE:
                    code.plantInstruction(new NoArgInstruction(opc_d2i));
                    break;
                }
                break;

            case Type.TID_LONG:
                switch (from.getTypeID()) {
                case Type.TID_BYTE:
                case Type.TID_CHAR:
                case Type.TID_SHORT:
                case Type.TID_INT:
                    code.plantInstruction(new NoArgInstruction(opc_i2l));
                    break;
                case Type.TID_LONG:
                    break;
                case Type.TID_FLOAT:
                    code.plantInstruction(new NoArgInstruction(opc_f2l));
                    break;
                case Type.TID_DOUBLE:
                    code.plantInstruction(new NoArgInstruction(opc_d2l));
                    break;
                }
                break;
 
            case Type.TID_FLOAT:
                switch (from.getTypeID()) {
                case Type.TID_BYTE:
                case Type.TID_CHAR:
                case Type.TID_SHORT:
                case Type.TID_INT:
                    code.plantInstruction(new NoArgInstruction(opc_i2f));
                    break;
                case Type.TID_LONG:
                    code.plantInstruction(new NoArgInstruction(opc_l2f));
                    break;
                case Type.TID_FLOAT:
                    break;
                case Type.TID_DOUBLE:
                    code.plantInstruction(new NoArgInstruction(opc_d2f));
                    break;
                }
                break;
 
            case Type.TID_DOUBLE:
                switch (from.getTypeID()) {
                case Type.TID_BYTE:
                case Type.TID_CHAR:
                case Type.TID_SHORT:
                case Type.TID_INT:
                    code.plantInstruction(new NoArgInstruction(opc_i2d));
                    break;
                case Type.TID_LONG:
                    code.plantInstruction(new NoArgInstruction(opc_l2d));
                    break;
                case Type.TID_FLOAT:
                    code.plantInstruction(new NoArgInstruction(opc_f2d));
                    break;
                case Type.TID_DOUBLE:
                    break;
                }
                break;
            }
        }
    }

    /**
     * Returns the size used on the JVM stack by a value of this type.
     */
    public static int getSize(Type type)
    {
        switch (type.getTypeID()) {
        case Type.TID_VOID:
            return 0;
        case Type.TID_LONG:
        case Type.TID_DOUBLE:
            return 2;
        default:
            return 1;
        }
    }
}
