/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.compiler;

import java.io.BufferedReader;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PushbackReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;

import at.dms.classfile.SourceDescription;

/**
 * Buffer that holds a whole input file.
 */
public class InputBuffer
{
    /**
     * Description of source file.
     */
    private String name;

    /**
     * Whether it is possible to write a class file to the same place as
     * the source file.
     */
    private boolean writableLocation;

    /**
     * Characters in file.
     */
    public char[] data;

    /**
     * Size of file.
     */
    public int size;

    /**
     * Cursor in file.
     */
    public int pos;

    /**
     * Line number in file.
     */
    public int line;

    /**
     * Reader used while initializing buffer.
     */
    private Reader initialReader;

    /**
     * Temporary buffer used to speed up when initially reading the source
     * file into the input buffer.
     */
    private char[] initialBuffer = new char[16384];

    /**
     * The number of valid characters in the buffer. If less than the
     * size of the array, then this is the last portion of the source file;
     * no more characters are available from the source file when the
     * character in the buffer are all used.
     */
    private int initialBufferSize;

    /**
     * From where in initialBuffer to read the next character.
     */
    private int initialIndex;

    /**
     * Constructs a new input buffer.
     *
     * @param sourceFile
     * Description of source file.
     * @exception IOException
     * An I/O exception occurred.
     */
    public InputBuffer(SourceDescription sourceFile, String encoding)
        throws IOException
    {
        this.name = sourceFile.getName();
        this.writableLocation = sourceFile.isWritableLocation();

        try {
            createInitialReader(sourceFile.getInput(), encoding);
            readFile(sourceFile.getLength());
        }
        finally {
            clearInitialReader();
            sourceFile.getInput().close();
        }
    }

    /**
     * Creates the reader that will read the file in the constructor.
     */
    private void createInitialReader(InputStream input, String encoding)
        throws IOException
    {
        if (encoding == null) {
            initialReader = new InputStreamReader(input);
        }
        else {
            try {
                initialReader = new InputStreamReader(input, encoding);
            }
            catch (UnsupportedEncodingException e) {
                // The description from the Sun library is very poor.
                throw new UnsupportedEncodingException("Encoding '" +
                                                       encoding +
                                                       "' is not supported.");
            }
        }
    }

    /**
     * Clears all memory associated with the initial reader.
     */
    private void clearInitialReader()
    {
        initialReader = null;
        initialBuffer = null;
    }

    /**
     * Implement a simple buffered reader; private methods are fast since
     * they aren't virtual, and a good JIT compile will inline the method
     * too.
     */
    private int readChar()
        throws IOException
    {
        if (initialBufferSize == -1) {
            return -1;
        } else if (initialIndex == initialBufferSize) {
            initialBufferSize = initialReader.read(initialBuffer);
            if (initialBufferSize == -1) {
                return -1;
            }
            initialIndex = 0;
        }
        return initialBuffer[initialIndex++];
    }
    
    /**
     * Reads file into an input buffer.
     *
     * @param input
     * The input stream to read.
     * @exception IOException
     * An I/O exception occurred.
     */
    private void readFile(long fileLength)
        throws IOException
    {
        if (fileLength > Integer.MAX_VALUE) {
            throw new IOException("File " + getName() + " is too big.");
        }
        this.size = (int)fileLength;
        this.data = new char[this.size + 1];
        char[] unicode = new char[4];

        int readPos = 0;
        int tmp;

    fileLoop:
        while (true) {
            tmp = readChar();
            switch (tmp) {
            case -1:
                break fileLoop;
            default:
                this.data[readPos++] = (char)tmp;
                break;
            case '\\':
                /* Possible start of a unicode escape. Read one more
                 * character
                 */
                tmp = readChar();
                switch (tmp) {
                case 'u':
                    /* Yes, \ u means unicode */

                    /* The spec says any number of u characters may appear
                     * after the initial \ u; these extra u characters must
                     * be ignored.
                     */
                    do {;} while ((tmp = readChar()) == 'u');

                    /* After the \ u+, four hexadecimal digits must follow.
                     */
                    unicode[0] = (char)tmp;
                    for ( int i = 1; tmp != -1 && i < 4; i++ ) {

                        tmp = readChar();
                        unicode[i] = (char)tmp;
                    }

                    if (tmp == -1) {
                        throw new EOFException("Invalid unicode escape: " +
                                               "end of file encountered " +
                                               "within a unicode escape " +
                                               "sequence");
                    }

                    try {
                        this.data[readPos++] = (char)
                            Integer.parseInt(new String(unicode), 16);
                    }
                    catch (NumberFormatException e) {
                        throw new IOException("Invalid unicode escape: \\u"
                                              + new String(unicode));
                    }
                    break;
                case -1:
                    /* End-of-file just after the backslash; this means
                     * that the backslash is just a stand-alone backslash.
                     */
                    this.data[readPos++] = '\\';
                    break fileLoop;
                default:
                    /* Leave backslash in stream as-is, if it didn't mark the
                     * start of a unicode escape sequence. The next character
                     * is never a the start of a unicode escape sequence;
                     * \\u is just a backslash, a backslash and a u character.
                     */
                    this.data[readPos++] = '\\';
                    this.data[readPos++] = (char)tmp;
                    break;
                }
                break;
            }
        } 

        this.size = readPos;

        /* Ignore ^Z at end of file (see JLS 3.5); it is not part of
         * the file.
         */
        if (size > 0 && this.data[size - 1] == 0x1a) {
            size--;
        }

        this.data[this.size++] = ' '; // EOT


        this.pos = 0;
        this.line = 1;
    }

    /**
     * Returns the name of the file in this InputBuffer.
     */
    public String getName()
    {
        return name;
    }

    /**
     * Returns whether it is possible to write a class file to the same place
     * as the source file.
     */
    public boolean isWritableLocation()
    {
        return writableLocation;
    }

    /**
     * Releases cached buffer.
     */
    public void release()
    {
    }

    /**
     * Returns whether the input buffer has been read completely.
     */
    public final boolean isEndOfData()
    {
        return pos >= size;
    }

}
