/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * Abstract base class for different BodyContexts.
 *
 * It follows the control flow and maintains assignment information about
 * local variables and fields, which uncaught exceptions can occur at each
 * point in the block.
 * It also verify that code is still reachable.
 *
 * @see CContext
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CConstructorContext
 * @see CInitializerContext
 * @see CBodyContext
 * @see CBlockContext
 */
public class CAbstractBodyContext
    extends CContext
{
    /**
     * Code Unreachable at this point in control flow.
     * Binary constant for flowState.
     */
    public static final int FLOW_UNREACHABLE = 1 << 0;

    /**
     * Control flow is breaked from this point on.
     *
     * Code just after break in same flow is not reachable.
     * Binary constant for flowState.
     */
    public static final int FLOW_BREAKED = 1 << 1;

    /**
     * Control flow is continued from this point on.
     *
     * Code just after continue in same flow is not reachable.
     * Binary constant for flowState.
     */
    public static final int FLOW_CONTINUED = 1 << 2;

    /**
     * Control flow is inside a loop construct.
     * Binary constant for flowState.
     */
    public static final int FLOW_INLOOP = 1 << 3;

    /**
     * Control flow is inside a switch construct.
     * <em>Currently an alias for FLOW_INLOOP</em>.
     * Binary constant for flowState.
     *
     * @see #FLOW_INLOOP
     */
    public static final int FLOW_INSWITCH = FLOW_INLOOP;

    /**
     * Control flow analysis info.
     */
    protected int flowState;

    /**
     * Local variable assignment information.
     */
    protected CVariableInfo variableInfo;

    /**
     * Field assignment information.
     */
    protected CVariableInfo fieldInfo;

    /**
     * Exceptions that can be thrown from this at a point in the source code.
     */
    protected Set throwables;

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     *
     * This starts a new context in a method.
     *
     * @param parent
     * The parent context, it must be different than null except if called by
     * the top level.
     */
    protected CAbstractBodyContext(CAbstractBodyContext parent)
    {
        super(parent);

        flowState = parent.getFlowState();
        throwables = parent.getThrowables();
    }

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     *
     * This starts a new context in a method.
     *
     * @param parent
     * The parent context, it must be different than null except if called by
     * the top level.
     */
    protected CAbstractBodyContext(CMethodContext parent)
    {
        super(parent);

        //flowState = 0;
        throwables = parent.getThrowables();
        assert(this instanceof CBlockContext);
    }

    /**
     * Verify everything is okay at the end of this context.
     */
    public void close(TokenReference ref)
    {
        informParent();
    }

    /**
     * Verify everything is okay at the end of this context.
     */
    public void informParent()
    {
        /* Inform parent */
        if (parent instanceof CAbstractBodyContext) {
            CAbstractBodyContext parent = (CAbstractBodyContext)this.parent;
      
            if ( variableInfo != null ) {
                int parentPosition =
                    parent.getBlockContext().getMaxLocalsIndex();
 
                for (int i = 0; i < parentPosition; i++) {
                    int info = variableInfo.getInfo(i);
   
                    if (info != 0) {
                        parent.addVariableInfo(i, info);
                    }
                }
            }
            parent.flowState = flowState;
        }

        if ( fieldInfo != null ) {
            int parentPosition =
                parent.getClassContext().getCClass().getFieldCount();
      
            for (int i = 0; i < parentPosition; i++) {
                int info = fieldInfo.getInfo(i);
 
                if (info != 0) {
                    parent.addFieldInfo(i, info);
                }
            }
        }
    }

    /**
     * Adds a break from within this context to the specified statement.
     */
    public final void addBreak(JStatement breakTo, CBodyContext from)
        throws PositionedError
    {
        addForwardedBreak(breakTo, from.makeSnapshot());
    }

    /**
     * Adds a continue from within this context to the specified statement.
     */
    public final void addContinue(JStatement continueTo, CBodyContext from)
        throws PositionedError
    {
        addForwardedContinue(continueTo, from.makeSnapshot());
    }

    /**
     * Adds a break from within this context to the specified statement.
     */
    protected void addForwardedBreak(JStatement breakTo, CBodyContext from)
        throws PositionedError
    {
        if (parent instanceof CAbstractBodyContext) {
            ((CAbstractBodyContext)parent).addForwardedBreak(breakTo, from);
        }
        else {
            throw new RuntimeException("Assertion failed, no unforwarded "
                                       + "breaks should exist at top method "
                                       + "level.");
        }
    }

    /**
     * Adds a continue from within this context to the specified statement.
     */
    protected void addForwardedContinue(JStatement continueTo,
                                        CBodyContext from)
        throws PositionedError
    {
        if (parent instanceof CAbstractBodyContext) {
            ((CAbstractBodyContext)parent).addForwardedContinue(continueTo,
                                                                from);
        }
        else {
            throw new RuntimeException("Assertion failed, no unforwarded "
                                       + "continues should exist at top "
                                       + "method level.");
        }
    }

    /**
     * Returns true if current control flow is inside a loop construct.
     */
    public boolean isInLoop()
    {
        return (flowState & FLOW_INLOOP) != 0;
    }

    /**
     * Returns true if current control flow is ever reachable.
     */
    public void setReachable(boolean reachable)
    {
        if (reachable) {
            flowState &= ~FLOW_UNREACHABLE;
        }
        else {
            flowState |= FLOW_UNREACHABLE;
        }
    }

    /**
     * Makes code in this flow unreachable after this point.
     */
    public void addReturn()
    {
        setReachable(false);
    }

    /**
     * Makes code in this flow unreachable after this point, and marks the
     * flow as "breaked".
     */
    public void addBreak()
    {
        flowState |= FLOW_UNREACHABLE | FLOW_BREAKED;
    }

    /**
     * Makes code in this flow unreachable after this point, and marks the
     * flow as "continued".
     */
    public void addContinue()
    {
        flowState |= FLOW_UNREACHABLE | FLOW_CONTINUED;
    }

    /**
     * Returns true if current point in control flow is ever reachable.
     */
    public boolean isReachable()
    {
        return (flowState & FLOW_UNREACHABLE) == 0;
    }

    /**
     * Returns true if current point in control flow is after a break
     * statement.
     */
    public boolean isBreaked()
    {
        return (flowState & FLOW_BREAKED) != 0;
    }

    /**
     * Returns true if current point in control flow is after a continue
     * statement.
     */
    public boolean isContinued()
    {
        return (flowState & FLOW_CONTINUED) != 0;
    }

    /**
     * Return the current flow state.
     *
     * @return
     * A bit-structure containing different information in different bits.
     */
    public int getFlowState()
    {
        return flowState;
    }

    /**
     * ...
     */
    public int getMaxLocalsPosition()
    {
        return ((CAbstractBodyContext)parent).getMaxLocalsPosition();
    }

    /**
     * Returns the statement with the specified label.
     */
    public JLabeledStatement getLabeledStatement(String label)
    {
        if (parent instanceof CAbstractBodyContext) {
            return ((CAbstractBodyContext)parent).getLabeledStatement(label);
        }
        else {
            return null;
        }
    }

    /**
     * Returns the nearest breakable statement.
     */
    public JStatement getNearestBreakableStatement()
    {
        if (parent instanceof CAbstractBodyContext) {
            return ( ((CAbstractBodyContext)parent)
                     .getNearestBreakableStatement() );
        }
        else {
            return null;
        }
    }

    /**
     * Returns the nearest continuable statement.
     */
    public JStatement getNearestContinuableStatement()
    {
        return parent instanceof CAbstractBodyContext ?
            ((CAbstractBodyContext)parent).getNearestContinuableStatement() :
            null;
    }

    /**
     * Adds new assignment info to existing assignment info.
     *
     * @param index
     * The variable position in method array of local vars.
     *
     * @param info
     * The assignment information to set.
     *
     * We make it a local copy of this information and at the end of this
     * context we will transfer it to the parent context according to
     * controlFlow.
     */
    public void addVariableInfo(int index, int info)
    {
        if (variableInfo == null) {
            variableInfo = new CVariableInfo();
        }

        if (variableInfo.getInfo(index) != 0) {
            variableInfo.addInfo(index, info);
        }
        else {
            variableInfo.addInfo(index, info | getVariableInfo(index));
        }
    }

    /**
     * Replaces old assignment info with new info.
     *
     * @param index
     * The variable position in method array of local vars.
     *
     * @param info
     * The assignment information to set.
     *
     * We make it a local copy of this information and at the end of this
     * context we will transfer it to the parent context according to
     * controlFlow.
     */
    public void setVariableInfo(int index, int info)
    {
        if (variableInfo == null) {
            variableInfo = new CVariableInfo();
        }

        variableInfo.setInfo(index, info);
    }

    /**
     * @param var  the definition of a variable
     * @return all informations we have about this variable
     */
    public int getVariableInfo(int index)
    {
        if (variableInfo == null) {
            return ((CAbstractBodyContext)parent).getVariableInfo(index);
        }

        int info = variableInfo.getInfo(index);

        if (info == 0) {
            info = ((CAbstractBodyContext)parent).getVariableInfo(index);
        } 

        return info;
    }

    /**
     * Adds new assignment info to existing assignment info.
     *
     * @param index
     * The field position.
     *
     * @param info
     * The assignment information to add.
     *
     * We make it a local copy of this information and at the end of this
     * context we will transfert it to the parent context according to
     * controlFlow.
     */
    public void addFieldInfo(int index, int info)
    {
        if (fieldInfo == null) {
            fieldInfo = new CVariableInfo();
        }

        if (fieldInfo.getInfo(index) != 0) {
            fieldInfo.addInfo(index, info);
        }
        else {
            fieldInfo.addInfo(index, info | getFieldInfo(index));
        }
    }

    /**
     * Replaces old assignment info with new info.
     *
     * @param index
     * The field position.
     *
     * @param info
     * The assignment information to set.
     *
     * We make it a local copy of this information and at the end of this
     * context we will transfert it to the parent context according to
     * controlFlow.
     */
    public void setFieldInfo(int index, int info)
    {
        if (fieldInfo == null) {
            fieldInfo = new CVariableInfo();
        }
        fieldInfo.setInfo(index, info);
    }

    /**
     * @param var  the definition of a field
     * @return all informations we have about this field
     */
    public int getFieldInfo(int index)
    {
        if (fieldInfo == null) {
            return parent.getFieldInfo(index);
        }

        int info = fieldInfo.getInfo(index);

        if (info == 0) {
            return parent.getFieldInfo(index);
        } 
        else {
            return info;
        }
    }

    /**
     * lookupField
     * @param ident  the name of the field
     * @return a variable from an ident in current context
     * @exception UnpositionedError this error will be positioned soon
     */
    public CField lookupField(String ident)
        throws UnpositionedError
    {
        return getClassContext().lookupField(ident);
    }

    /**
     * lookupLocalVariable
     * @param ident  the name of the variable
     * @return a variable from an ident in current context
     */
    public JLocalVariable lookupLocalVariable(String ident)
    {
        return parent.lookupLocalVariable(ident);
    }
    
    /**
     * lookupOuterLocalVariable
     * @param ident  the name of the variable
     * @return a variable from an ident in upperclass context
     */
    public JExpression lookupOuterLocalVariable(TokenReference ref,
                                                String ident)
    {
        CClassContext owner = getClassContext();
        
        return owner.lookupOuterLocalVariable(ref, ident);
    }
    
    /**
     * Adds one throwable to the set that can be thrown from this context.
     *
     * @param throwable
     * The type of the new throwable.
     */
    public void addThrowable(CThrowableInfo throwable)
    {
        if (throwables == null) {
            throwables = new HashSet(1);
        }
        throwables.add(throwable);
    }

    /**
     * Adds throwables to the set that can be thrown from this context.
     */
    public void addThrowables(Set newThrowables)
    {
        if (throwables == null) {
            throwables = new HashSet(newThrowables.size());
        }
        throwables.addAll(newThrowables);
    }

    /**
     * @return the list of exception that may be thrown
     */
    public Set getThrowables()
    {
        return throwables == null ? Collections.EMPTY_SET : throwables;
    }

}
