/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.ClassInfo;
import at.dms.classfile.FieldInfo;
import at.dms.classfile.InnerClassInfo;
import at.dms.classfile.MethodInfo;
import at.dms.compiler.Compiler;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * A class that is pre-compiled and loaded from a .class file.
 *
 * This class represents the exported members of a class (inner classes,
 * methods and fields)
 */
public class CBinaryClass
    extends CClass
{
    /**
     * Extra information about this class if it is an inner class, else
     * <code>null</code>.
     */
    private InnerClassInfo extraInfo;

    /**
     * Constructs a class export from file.
     *
     * @param compiler
     * Compiler instance for binary files. (that are not really compiled now,
     * having been compiled earlier)
     * @param classInfo
     * Standard ClassInfo record from classfile for this class.
     * @param extraInfo
     * If this class is an inner class, then this contains extra information
     * about it, and this comes from the class of which this is a member
     * class, or from a class that uses this member class directly. If this
     * is not an inner class, then this parameter should be <code>null</code>.
     */
    public CBinaryClass(Compiler compiler,
                        ClassInfo classInfo,
                        InnerClassInfo extraInfo)
    {
        super(compiler,
              ( extraInfo == null || extraInfo.getOuterClass() == null
                ? getOwner(classInfo.getName())
                : lookupClassFileClass(extraInfo.getOuterClass().getName()) ),
              classInfo.getSourceFile(),
              false,
              ( extraInfo == null
                ? classInfo.getModifiers() : extraInfo.getModifiers() ),
              ( extraInfo == null || extraInfo.getSimpleName() == null
                ? getIdent(classInfo.getName()) : extraInfo.getSimpleName() ),
              classInfo.getName(),
              ( classInfo.getSuperClass() == null
                ? null
                : ClassOrInterfaceType.lookup(classInfo.getSuperClass()) ),
              classInfo.isDeprecated());

        this.extraInfo = extraInfo;
    }

    void close(ClassInfo classInfo, InnerClassInfo extraInfo)
        throws UnpositionedError
    {

        setInnerClasses(loadInnerClasses(classInfo.getInnerClasses()));

        ClassOrInterfaceType[] interfaces =
            loadInterfaces(classInfo.getInterfaces());

        FieldInfo[] fields = classInfo.getFields();
        CField[] fieldExports = null;

        fieldExports = new CField[fields.length];
        for (int i = 0; i < fields.length; i++) {
            fieldExports[i] = new CBinaryField(this, fields[i]);
        } 

        MethodInfo[] methods = classInfo.getMethods();
        CMethod[] methodExports = new CMethod[methods.length];

        for (int i = 0; i < methods.length; i++) {
            methodExports[i] = new CBinaryMethod(this, methods[i]);
        }

        close(interfaces, fieldExports, methodExports);
    }

    /**
     * Returns extra information about this class if it is an inner class, else
     * <code>null</code>.
     */
    public InnerClassInfo getExtraInfo()
    {
        return extraInfo;
    }

    /**
     * Loads the ClassType objects that represent the types of the
     * interfaces that are implemented by this class.
     */
    private ClassOrInterfaceType[] loadInterfaces(String[] interfaces)
        throws UnpositionedError
    {
        if (interfaces != null) {
            ClassOrInterfaceType[] ret;
            ret = new ClassOrInterfaceType[interfaces.length];
            for (int i = 0; i < interfaces.length; i++) {
                ret[i] = ClassOrInterfaceType.get(interfaces[i]);
            }
            return ret;
        }
        else {
            return ClassOrInterfaceType.EMPTY;
        }
    }

    /**
     * Loads the information for the inner that are used by this class.
     */
    private ClassOrInterfaceType[] loadInnerClasses(InnerClassInfo[] inners)
        throws UnpositionedError
    {
        if (inners != null) {
            ClassOrInterfaceType[] innerClasses =
                new ClassOrInterfaceType[inners.length];

            for (int i = 0; i < inners.length; i++) {
                innerClasses[i] =
                    ClassOrInterfaceType.get(inners[i].getQualifiedName());
            }
            return innerClasses;
        }
        else {
            return ClassOrInterfaceType.EMPTY;
        }
    }

    /**
     * ...
     */
    private static CClass lookupClassFileClass(String clazz)
                                               
    {
        ReferenceType type = ClassOrInterfaceType.get(clazz);
        return type.getCClass();
    }

    /**
     * Returns the owner of this class, if it is an inner class,
     * or null if it isn't.
     */
    private static CClass getOwner(String clazz)
    {
        int index = clazz.lastIndexOf("$");
    
        if ( index == -1 ) {
            return null;
        }
        else {
            ReferenceType type =
                ClassOrInterfaceType.get(clazz.substring(0, index));
            return type.getCClass();
        }
    }
}
