/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.Map;

import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * This class represents a compilation unit
 */
public class CCompilationUnit
{
    /**
     * Package name of this compilation unit.
     */
    private final String packageName;

    /**
     * The imported classes.
     */
    private final JClassImport[] importedClasses;

    /**
     * The imported packages (import statement ending in ".*")
     */
    private final JPackageImport[] importedPackages;

    /**
     * Construct a compilation unit context.
     */
    public CCompilationUnit(String packageName,
                            JClassImport[] importedClasses,
                            JPackageImport[] importedPackages)
    {
        this.packageName = packageName;
        this.importedClasses = importedClasses;
        this.importedPackages = importedPackages;
    }

    /**
     * Returns a class according to imports or null if error occur.
     *
     * @return
     * A class according to imports or null if error occur.
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public ClassOrInterfaceType lookupClass(CClass access, String name)
        throws UnpositionedError
    {
        if (name.lastIndexOf('/') == -1) {
            // 6.5.4.1 Simple Type Names

            ClassOrInterfaceType type = null;

            // First look for a type declared by a single-type-import of by a
            // type declaration

            // If type is declared by a single-type-import, mark it as
            // used (max. 1)
            for (int i = 0; i < importedClasses.length; i++) {
                if (name.equals(importedClasses[i].getClassIdent())) {
                    importedClasses[i].setUsed();
                    String qualifiedName = importedClasses[i].getClassName();
                    type = ClassOrInterfaceType.get(qualifiedName);
                    break;
                }
            }

            if (type != null) {
                if (access == null || type.getCClass().isAccessible(access)) {
                    return type;
                }
                else {
                    return null;
                }
            }

            // Otherwise, look for a type declared in another compilation unit
            // of this package
            if (packageName.length() == 0) {

                /* Package is default package. */
                if (CTopLevel.hasClassFile(name)) {
                    type = ClassOrInterfaceType.get(name);
                }
            }
            else {

                String temp = packageName + '/' + name;
                if (CTopLevel.hasClassFile(temp)) {
                    type = ClassOrInterfaceType.get(temp);
                }
            }

            if (type == null) {
                // Otherwise, look for a type declared by EXACTLY ONE
                // import-on-demand declaration
                boolean foundOnce = false;
                for (int i = 0; i < importedPackages.length; i++) {
                    String qualifiedName = (importedPackages[i].getName()
                                            + '/' + name).intern();

                    if (CTopLevel.hasClassFile(qualifiedName)) {
                        if (!foundOnce) {
                            foundOnce = true;
                        }
                        else {
                            // Oops, the name is ambiguous (declared by more
                            // than one import-on-demand declaration)
                            throw new UnpositionedError
                                (KjcMessages.CUNIT_RENAME2, name);
                        }

                        type = ClassOrInterfaceType.get(qualifiedName);
                        importedPackages[i].setClassUsed(name);
                        
                    }
                }
            }

            /* Now the name must be unique and found. */
            if (type == null) {
                // The javadoc says to return null.
                return null;
            }

            if (access == null || type.getCClass().isAccessible(access)) {
                return type;
            }
            else {
                return null;
            }
        }
        else {
            // 6.5.4.2 Qualified Type Names: look directly at top
            if (!CTopLevel.hasClassFile(name)) {
                // The javadoc says to return null.
                return null;
            }

            ClassOrInterfaceType result = ClassOrInterfaceType.get(name);
            if (access == null || result.getCClass().isAccessible(access)) {
                return result;
            }
            else {
                return null;
            }
            
        }
    }
}
