/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.Compiler;
import at.dms.compiler.PositionedError;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;
import at.dms.util.MessageDescription;
import at.dms.util.Utils;

/**
 * This class represents a local context during checkBody.
 * It follows the control flow and maintain informations about
 * variable (initialized, used, allocated) and exceptions (thrown, caught.)
 * It also verify that context is still reachable.
 *
 * There is a set of utilities method to access fields, methods and class
 * with the name by clamping the parsing tree.
 * @see CContext
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CBodyContext
 * @see CBlockContext
 */
public abstract class CContext
    extends Utils
    implements Constants
{
    /**
     * ...
     */
    protected CContext parent;

    /**
     * ...
     */
    protected CCompilationUnitContext cunit;

    /**
     * Construct a non-specific context.
     */
    public CContext()
    {
        cunit = (CCompilationUnitContext)this;
    }

    /**
     * Construct a non-specifick context. This constructor is for subclasses
     * only.
     *
     * @param parent
     * The parent context. May not be null except if called by the top level.
     */
    protected CContext(CContext parent)
    {
        this.parent = parent;
        cunit = parent.cunit;
    }

    /**
     * Search for a class with the provided name.
     *
     * @param access
     * @param caller
     * Referring class or <code>null</code> for 'able to access everything'.
     * @param name
     * Method name.
     * @return
     * @exception UnpositionedError
     * This error will be positioned soon.
     * The class if found, null otherwise.
     */
    public ClassOrInterfaceType lookupClass(CClass access, String name)
        throws UnpositionedError
    {
        return parent.lookupClass(access, name);
    }

    /**
     * Search for a matching method with the provided type parameters.
     *
     * @param caller
     * Caller class or <code>null</code> for 'able to access everything'
     * @param ident
     * Method name.
     * @param params
     * Method parameters
     * @return
     * The method if found, null otherwise.
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public CMethod lookupMethod(CClass caller, String ident, Type[] params)
        throws UnpositionedError
    {
        return parent.getClassContext().lookupMethod(caller, ident, params);
    }

    /**
     * lookupLocalVariable
     * @return a variable from an ident in current context
     * @exception UnpositionedError this error will be positioned soon
     */
    public CField lookupField(String ident) throws UnpositionedError
    {
        return getClassContext().lookupField(ident);
    }

    /**
     * lookupLocalVariable
     * @param ident  the name of the local variable
     * @return a variable from an ident in current context
     */
    public JLocalVariable lookupLocalVariable(String ident)
    {
        return parent.lookupLocalVariable(ident);
    }

    /**
     * @param field  the definition of a field
     * @return a field from a field definition in current context
     */
    public int getFieldInfo(int index)
    {
        return parent.getFieldInfo(index);
    }

    /**
     * @param field  the definition of a field
     * @return a field from a field definition in current context
     */
    public void addFieldInfo(int index, int info)
    {
        parent.addFieldInfo(index, info);
    }

    /**
     * @param field  the definition of a field
     * @return a field from a field definition in current context
    public void setFieldInfo(int index, int info)
    {
        parent.setFieldInfo(index, info);
    }
     */

    /**
     * getParentContext
     * @return the parent
     */
    public CContext getParentContext()
    {
        return parent;
    }

    /**
     * @return the compilation unit
     */
    public CCompilationUnitContext getCompilationUnitContext()
    {
        return cunit;
    }

    /**
     * Returns the nearest parent of type CClassContext.
     *
     * @return the nearest parent of type CClassContext.
     */
    public CClassContext getClassContext()
    {
        return parent.getClassContext();
    }

    /**
     * Returns the nearest parent of type CMethodContext.
     *
     * @return the nearest parent of type CMethodContext.
     */
    public CMethodContext getMethodContext()
    {
        return parent.getMethodContext();
    }

    /**
     * Returns the nearest block context.
     */
    public CBlockContext getBlockContext()
    {
        return parent.getBlockContext();
    }

    /**
     * Adds a class to generate.
     */
    public void addSourceClass(CSourceClass clazz)
    {
        getCompilationUnitContext().addSourceClass(clazz);
    }

    /**
     * Add an error into the list and eat it.
     * This method should be called after a try catch block after catching
     * exception or directly without exception thrown.
     *
     * @param trouble
     * The error
     */
    public void reportTrouble(PositionedError trouble)
    {
        getCompilationUnitContext().reportTrouble(trouble);
    }

    /**
     * Add an error into the list and eat it.
     * This method should be called after a try catch block after catching
     * exception or directly without exception thrown.
     *
     * @param mess
     * The error message.
     * @param param1
     * The first message parameter.
     * @param param2
     * The second message parameter.
     * @exception UnpositionedError
     * this error will be positioned soon.
     */
    public void fail(MessageDescription mess, Object param1, Object param2)
        throws UnpositionedError
    {
        throw new UnpositionedError(mess, param1, param2);
    }

    /**
     * Verify an expression and if false add an error.
     *
     * @param expr
     * Condition to verify.
     * @param mess
     * The message to be displayed.
     * @return
     * The value of expr.
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public boolean check(boolean expr,
                         MessageDescription mess)
        throws UnpositionedError
    {
        if (!expr) {
            fail(mess, null, null);
        }
        return expr;
    }

    /**
     * Verify an expression and if false add an error.
     *
     * @param expr
     * Condition to verify.
     * @param mess
     * The message to be displayed.
     * @param param1
     * The first parameter.
     * @return
     * The value of expr.
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public boolean check(boolean expr,
                         MessageDescription mess,
                         Object param1)
        throws UnpositionedError
    {
        if (!expr) {
            fail(mess, param1, null);
        }
        return expr;
    }

    /**
     * Verify an expression and if false add an error.
     *
     * @param expr  condition to verify
     * @param mess  the message to be displayed
     * @param param1  the first parameter
     * @param param2  the second parameter
     * @return the value of expr
     * @exception UnpositionedError this error will be positioned soon
     */
    public boolean check(boolean expr,
                         MessageDescription mess,
                         Object param1,
                         Object param2)
        throws UnpositionedError
    {
        if (!expr) {
            fail(mess, param1, param2);
        }
        return expr;
    }

    /**
     * Gets the compiler.
     */
    public Compiler getCompiler()
    {
        return parent.getCompiler();
    }
}
