/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.classfile.FieldInfo;
import at.dms.classfile.FieldRefInstruction;

/**
 * This class represents an exported field of a class.
 */
public abstract class CField
    extends CMember
{
    /**
     * Type of field.
     */
    private Type type;

    /**
     * Value of field.
     */
    private JExpression value;

    /**
     * Constructs a field export.
     *
     * @param owner
     * The owner of this field.
     * @param modifiers
     * The modifiers on this field.
     * @param ident
     * The name of this field.
     * @param type
     * The type of this field.
     * @param deprecated
     * Is this field deprecated?
     */
    public CField(CClass owner,
                  int modifiers,
                  String ident,
                  Type type,
                  boolean deprecated)
    {
        super(owner, modifiers, ident, deprecated);
        this.type = type;
    }

    /**
     * @return the interface.
     */
    public CField getField()
    {
        return this;
    }

    /**
     * @return the type of this field.
     */
    public Type getType()
    {
        return type;
    }

    /**
     * @param value the value known at third pass.
     */
    public void setValue(JExpression value)
    {
        this.value = value;
    }

    /**
     * @return the value of initializer or null.
     */
    public JExpression getValue()
    {
        //!!!TEST
        if (value != null && ! value.getType().equals(type)) {
            System.err.println("<<< get " + getPrefixName() + "."
                               + getIdent() + ": bad value type "
                               + value.getType()
                               + "(" + type + " expected)");
        }
        //!!!TEST
        return value;
    }

    /**
     * Generates a sequence of bytecodes to load.
     *
     * @param code
     * The code list.
     */
    public void genLoad(CodeSequence code)
    {
        code.plantInstruction
            (new FieldRefInstruction(isStatic() ? opc_getstatic : opc_getfield,
                                     getPrefixName(),
                                     getIdent(),
                                     getType().getSignature()));
    }

    /**
     * Generates a sequence of bytecodes to load.
     *
     * @param code
     * The code list.
     */
    public void genStore(CodeSequence code)
    {
        code.plantInstruction
            (new FieldRefInstruction(isStatic() ? opc_putstatic : opc_putfield,
                                     getPrefixName(),
                                     getIdent(),
                                     getType().getSignature()));
    }

    /**
     * Generate the code in a class file.
     */
    public FieldInfo genFieldInfo()
    {
        Object localValue = null;

        if (value != null
            && value.isConstant()
            && isFinal()
            && isStatic()
            && !(value instanceof JNullLiteral)) {

            localValue = value.getLiteral().getValue();

            if ( value.getType().isIntegral() &&
                 !(localValue instanceof Long) ) {

                localValue =
                    new Integer(value.getOrdinalLiteral().getNumberValue()
                                .intValue());
            }
            else if (value.getType() == StandardType.FLOAT) {
                localValue = new Float(((Number)localValue).floatValue());
            }
        }

        return new FieldInfo((short)getModifiers(),
                             getIdent(),
                             type.getSignature(),
                             localValue,
                             isDeprecated(),
                             isSynthetic());
    }
}
