/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.util.Utils;

/**
 * This class represents an exported member of a class.
 *
 * @see CMember
 * @see CField
 * @see CMethod
 * @see CClass
 */
public class CMember
    extends Utils
    implements Constants
{
    /**
     * Prefix for names of accessor methods.
     */
    public final static String ACCESSOR_PREFIX = "access$";

    /**
     * Class in which this member appears.
     */
    private CClass owner;

    /**
     * Java modifiers for member (like public, final, etc.)
     */
    private int modifiers;

    /**
     * Name of the member.
     */
    private String ident;

    /**
     * Whether this member is deprecated by the programmer.
     */
    private boolean deprecated;

    /**
     * Whether this is a synthetic (generated by the compiler) member.
     */
    private boolean synthetic;

    /**
     * Constructs a field export
     * @param owner  the owner of this member
     * @param modifiers the modifiers on this member
     * @param ident  the ident of this member
     * @param deprecated is this member deprecated
     */
    public CMember(CClass owner,
                   int modifiers, String ident, boolean deprecated)
    {
        this.owner = owner;
        this.modifiers = modifiers;
        this.ident = ident;
        this.deprecated = deprecated;
    }

    /**
     * Whether this is a synthetic member.
     */
    public boolean isSynthetic()
    {
        return synthetic;
    }

    /**
     * Sets the flag that indicates if this member is synthetic or not.
     */
    public void setSynthetic(boolean flag)
    {
        this.synthetic = flag;
    }

    /**
     * Returns the owner of this member.
     */
    public CClass getOwner()
    {
        return owner;
    }

    /**
     * Returns the ident of this member.
     *
     * @return the ident of this member.
     */
    public String getIdent()
    {
        return ident;
    }

    /**
     * @return the fully qualified name of this member
     */
    public String getQualifiedName()
    {
        return owner.getQualifiedName() + JAV_NAME_SEPARATOR + ident;
    }

    /**
     * @return the fully qualified name of this member
     */
    public String getPrefixName()
    {
        return owner.getQualifiedName();
    }

    /**
     * @return the fully qualified name of this member
     */
    public String getJavaName()
    {
        return getQualifiedName().replace('/' , '.');
    }

    /**
     * @return the modifiers of this member.
     */
    public int getModifiers()
    {
        return modifiers;
    }

    /**
     * (Re)sets the modifiers of this member.
     */
    public void setModifiers(int modifiers)
    {
        this.modifiers = modifiers;
    }

    /**
     * @return the interface
     */
    public CField getField()
    {
        fail();
        return null;
    }

    /**
     * @return the interface
     */
    public CMethod getMethod()
    {
        fail();
        return null;
    }

    /**
     * @return the interface
     */
    public CClass getCClass()
    {
        fail();
        return null;
    }

    /**
     * @return true if this member is static
     */
    public boolean isStatic()
    {
        return hasFlag(modifiers, ACC_STATIC);
    }

    /**
     * @return true if this member is public
     */
    public boolean isPublic()
    {
        return hasFlag(modifiers, ACC_PUBLIC);
    }

    /**
     * @return true if this member is protected
     */
    public boolean isProtected()
    {
        return hasFlag(modifiers, ACC_PROTECTED);
    }

    /**
     * @return true if this member is private
     */
    public boolean isPrivate()
    {
        return hasFlag(modifiers, ACC_PRIVATE);
    }

    /**
     * @return true if this member is final
     */
    public boolean isFinal()
    {
        return hasFlag(modifiers, ACC_FINAL);
    }

    /**
     * @return true if this member is deprecated
     */
    public boolean isDeprecated()
    {
        return deprecated;
    }

    /**
     * Returns true if it is necessary to use an accessor method to access
     * this member. Precondition: the member must be accessible for
     * <code>from</code> (use <code>isAccessible</code> method).
     */
    public boolean mustUseAccessorMethod(CClass from)
    {
        /* An accessor method is never needed to access private fields
         * inside a class itself.
         */
        if (from == getOwner()) {
            return false;
        }

        /* Is the accessing class defined inside the owner of this member?
         * Or is the owner of this field defined inside the class that wants
         * to access this field? Then if it is a private field, an 
         * accessor is required.
         */
        if (from.isDefinedInside(owner) || owner.isDefinedInside(from)) {
            return isPrivate();
        }

        else if (from.isInnerClass()) {
            if (mustUseAccessorMethod(from.getOwner())) {
                return isPrivate();
            }
        }

        return false;
    }

    /**
     * Returns true if this member is accessisble from the specified class
     * according to the JLS. This doesn't mean that it is accessible according
     * to the JVM. One must check using
     * <code>mustUseAccessorMethod(CClass)</code> if an accessor method is
     * required to let another class access the member.
     *
     * @return true if this member is accessible
     */
    public boolean isAccessible(CClass from)
    {
        if ( isPublic() ||
             from.isDefinedInside(owner) ||
             owner.isDefinedInside(from) ) {

            return true;
        }

        if (from.isInnerClass()) {
            if (isAccessible(from.getOwner())) {
                return true;
            }
        }

        if (!isPrivate() && owner.getPackage().equals(from.getPackage())) {
            return true;
        }

        if (isProtected() && from.descendsFrom(owner)) {
            return true;
        }

        return false;
    }

}
