/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.Set;
import java.util.Iterator;

import at.dms.classfile.InvokeinterfaceInstruction;
import at.dms.classfile.MethodInfo;
import at.dms.classfile.MethodRefInstruction;
import at.dms.codegen.jvm.Types;
import at.dms.type.*;

/**
 * This class represents a class method.
 */
public abstract class CMethod
    extends CMember
{
    /**
     * Return type of method.
     */
    private Type returnType;

    /**
     * Parameter types of method.
     */
    private Type[] parameters;

    /**
     * Types of exceptions that can be thrown from method.
     */
    private ClassOrInterfaceType[] exceptions;

    /**
     * Constructs a method member.
     *
     * @param owner
     * The owner of this method.
     * @param modifiers
     * The modifiers on this method.
     * @param ident
     * The ident of this method.
     * @param returnType
     * The return type of this method.
     * @param parameters
     * The parameter types of this method.
     * @param exceptions
     * A list of all exceptions of the throws list.
     * @param deprecated
     * Is this method deprecated.
     */
    public CMethod(CClass owner,
                   int modifiers,
                   String ident,
                   Type returnType,
                   Type[] parameters,
                   ClassOrInterfaceType[] exceptions,
                   boolean deprecated)
    {
        super(owner, modifiers, ident, deprecated);

        this.returnType = returnType;
        this.parameters = parameters;
        this.exceptions = exceptions;
    }

    /**
     * @return the interface
     */
    public CMethod getMethod()
    {
        return this;
    }

    /**
     * @return the type of this field
     */
    public final Type getReturnType()
    {
        return returnType;
    }

    /**
     * @return the type of this field
     */
    public final Type[] getParameters()
    {
        return parameters;
    }

    /**
     * @return the type of this field
     */
    public String getSignature()
    {
        return TypeSignature.genMethodSignature(returnType, parameters);
    }

    /**
     * @return the exceptions that can be thrown by this method
     */
    public ClassOrInterfaceType[] getThrowables()
    {
        return exceptions;
    }

    /**
     * This method is used by initializers that knows throwables exceptions
     * only after body was checked.
     *
     * @param throwables
     * The exceptions that can be thrown by this method.
     */
    public void setThrowables(Set throwables)
    {
        if (throwables != null) {
            Iterator iterator = throwables.iterator();
            int  count = 0;

            exceptions = new ClassOrInterfaceType[throwables.size()];
            while (iterator.hasNext()) {
                exceptions[count++] = ((CThrowableInfo)iterator.next())
                    .getThrowable();
            }
        }
    }

    /**
     * Returns true iff this method is native.
     */
    public boolean isNative()
    {
        return hasFlag(getModifiers(), ACC_NATIVE);
    }

    /**
     * Returns true iff this method is abstract.
     */
    public boolean isAbstract()
    {
        return hasFlag(getModifiers(), ACC_ABSTRACT);
    }

    /**
     * Returns true iff this method is a constructor.
     */
    public boolean isConstructor()
    {
        return getIdent() == JAV_CONSTRUCTOR;
    }

    /**
     * Implementation of java.lang.Object.equals(Object)
     */
    public boolean equals(Object o)
    {
        if (o instanceof CMethod) {
            return equals((CMethod)o);
        }
        else {
            return false;
        }
    }

    /**
     * Checks if two methods have same signature.
     *
     * @param method
     * Another method.
     */
    public boolean equals(CMethod method)
    {
        if (parameters.length != method.parameters.length) {
            return false;
        }

        for (int i = 0; i < parameters.length; i++) {
            if (!parameters[i].equals(method.parameters[i])) {
                return false;
            }
        }
        if (getIdent() != method.getIdent()) {
            return false;
        }

        return true;
    }
  
    /**
     * Checks if this matches with the provided name and type parameters.
     *
     * @param ident
     * Name of method.
     * @param actuals
     * Method parameter types.
     */
    public boolean isMatching(String ident, Type[] actuals)
    {
        return ( parameters.length == actuals.length &&
                 ident.equals(getIdent()) &&      
                 isMatchingParams(actuals) );
    }

    /**
     * Checks if this matches with the provided type parameters.
     *
     * @param actuals
     * Method parameter types.
     */
    public boolean isMatchingParams(Type[] actuals)
    {
        Type[] formals = getParameters();

        for (int i = 0; i < actuals.length; i++) {
            /* Method invocation conversion = assigment conversion without
             * literal narrowing we just look at the type and do not consider
             * literal special case.
             */
            if (!actuals[i].allowsAssignmentConversion(formals[i])) {
                return false;
            }
        }

        return true;
    }

    /**
     * Checks if this method is more specific than other.
     *
     * @param other
     * The method to compare to.
     * @return true
     * If this method is more specific than other.
     */
    public boolean isMoreSpecificThan(CMethod other)
    {
        Type[] theseFormals = getParameters();
        Type[] otherFormals = other.getParameters();

        for (int i = 0; i < otherFormals.length; i++) {
            /* method invocation conversion = assigment conversion without
             * literal narrowing we just look at the type and do not consider
             * literal special case.
             */
            if (!theseFormals[i].allowsAssignmentConversion(otherFormals[i])) {
                return false;
            }
        }

        return true;
    }

    /**
     * Returns nice string representing method name and number of parameters
     * for use in display output.
     */
    public String toString()
    {
        return getIdent() + "(" + parameters.length + ")";
    }

    /**
     * Generates a sequence of bytecode to call this method.
     *
     * @param code
     * The code sequence.
     * @param nonVirtual
     * Force non-virtual dispatching
     */
    public void genCallCode(CodeSequence code, boolean nonVirtual)
    {
        if (getOwner().isInterface()) {
            int size = 0;
      
            for (int i = 0; i < parameters.length; i++) {
                size += Types.getSize(parameters[i]);
            }
      
            code.plantInstruction
                (new InvokeinterfaceInstruction(getPrefixName(),
                                                getIdent(),
                                                getSignature(),
                                                size + 1));
        }
        else {
            int opcode;

            if (isStatic()) {
                opcode = opc_invokestatic;
            }
            else if (nonVirtual || isPrivate()) {
                // JDK1.2 does not like ... || isFinal()
                opcode = opc_invokespecial;
            }
            else {
                opcode = opc_invokevirtual;
            }

            code.plantInstruction(new MethodRefInstruction(opcode,
                                                           getPrefixName(),
                                                           getIdent(),
                                                           getSignature()));
        }
    }

    /**
     * Creates a MethodInfo object.
     */
    public MethodInfo genMethodInfo()
    {
        /* Uhh? fail? */
        fail();
        String[] exceptions = new String[this.exceptions.length];
        for (int i = 0; i < this.exceptions.length; i++) {
            exceptions[i] = this.exceptions[i].getSignature();
        }

        return new MethodInfo((short)getModifiers(),
                              getIdent(),
                              getSignature(),
                              exceptions,
                              null,
                              isDeprecated(),
                              false);
    }

}
