/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.MethodInfo;
import at.dms.classfile.CodeInfo;
import at.dms.codegen.jvm.Types;
import at.dms.type.*;

/**
 * This class represents an exported method declaration of a class or
 * interface.
 */
public class CSourceMethod
    extends CMethod
{
    /**
     * The syntax tree of the code of the method.
     */
    private JBlock body;

    /**
     * Whether this method is used by non-synthetic code.
     */
    private boolean used;

    /**
     * How heavily the code in this method must be optimized.
     */
    private int optimizingLevel;

    /**
     * If this is a private method, and an inner class of the owner wants
     * to call it, an accessor method must be generated and
     * is stored in this field so it can be re-used.
     */
    private CSourceMethod innerAccessor;

    /**
     * If a member class of the owner wants to call the super-method
     * of this method, an accessor method must be generated and
     * is stored in this field so it can be re-used.
     */
    private CSourceMethod superAccessor;

    /**
     * If this method is a constructor, and it contains an explicit constructor
     * invocation, then this field contains the constructor that was invoked.
     * It is used to detect cycles in explicit constructor invocations.
     */
    private CMethod explicitlyInvokedConstructor;

    /**
     * Constructs a source method export.
     *
     * @param owner
     * The owner of this method.
     * @param modifiers
     * The modifiers on this method.
     * @param ident
     * The ident of this method.
     * @param returnType
     * The return type of this method.
     * @param paramTypes
     * The parameter types of this method.
     * @param exceptions
     * A list of all exceptions in the throws list.
     * @param deprecated
     * Is this method deprecated.
     * @param body
     * The source code.
     * @param optimizingLevel
     * The optimization level.
     */
    public CSourceMethod(CClass owner,
                         int modifiers,
                         String ident,
                         Type returnType,
                         Type[] paramTypes,
                         ClassOrInterfaceType[] exceptions,
                         boolean deprecated,
                         JBlock body,
                         int optimizingLevel)
    {
        super(owner, modifiers, ident, returnType, paramTypes, exceptions,
              deprecated);

        this.body = body;
        this.optimizingLevel = optimizingLevel;
    }

    /**
     * Whether this method is used. All public and synthetic methods are
     * 'used' by default.
     */
    public boolean isUsed()
    {
        return used || !isPrivate() || isSynthetic();
    }

    /**
     * Marks a method as 'used' or not-used. Public and synthetic methods
     * are not affected by a call to this method.
     */
    public void setUsed(boolean flag)
    {
        this.used = flag;
    }

    public void setExplicitlyInvokedContructor(CMethod constructor)
    {
        this.explicitlyInvokedConstructor = constructor;
    }

    public CMethod getExplicitlyInvokedContructor()
    {
        return this.explicitlyInvokedConstructor;
    }

    /**
     * Generates a new MethodInfo object.
     */
    public MethodInfo genMethodInfo()
    {
        ClassOrInterfaceType[] excs = getThrowables();
        String[]  exceptions = new String[excs.length];
        for (int i = 0; i < excs.length; i++) {
            exceptions[i] = excs[i].getQualifiedName();
        }

        if (body != null) {
            return new CMethodInfo((short)getModifiers(),
                                   getIdent(),
                                   getSignature(),
                                   exceptions,
                                   this,
                                   isDeprecated(),
                                   false);
        }
        else {
            return new MethodInfo((short)getModifiers(),
                                  getIdent(),
                                  getSignature(),
                                  exceptions,
                                  null,
                                  isDeprecated(),
                                  false);
        }
    }

    /**
     * Generates or returns an already-generated method that can access
     * this method from classes that should be able to access this method
     * per the JLS, but can't because the JVM doesn't allow it.
     * <p>
     * This is the case for inner classes calling a private method in an
     * outer class, or an outer class calling a private method in an inner
     * class.
     * <p>
     * The accessor method will have package-only access rights, which is okay
     * for inner classes, since they always reside in the same package as
     * the class itself.
     */
    public CSourceMethod getAccessor()
        throws PositionedError
    {
        if (innerAccessor != null) {
            /* Accessor is already created. */
            return innerAccessor;
        }

        TokenReference where;
        if (body != null) {
            where = body.getTokenReference();
        }
        else {
            where = TokenReference.NO_REF;
        }
            

        CSourceClass owner = (CSourceClass)getOwner();
        if (isConstructor()) {
            innerAccessor = owner.addJMethod(makeConstructorAccessor(where));
        }
        else {
            innerAccessor = owner.addJMethod(makeNormalAccessor(where));
        }

        return innerAccessor;
    }

    /**
     * Generates or returns an already-generated method that can access
     * the super-method of this method from classes that should be able to
     * access this method per the JLS, but can't because the JVM doesn't
     * implement it.
     * <p>
     * This is the case for inner classes calling a super-method of an
     * outer class.
     * <p>
     * The accessor method will have package-only access rights, which is okay
     * for inner classes, since they always reside in the same package as
     * the class itself.
     */
    public CSourceMethod getSuperAccessor()
        throws PositionedError
    {
        if (superAccessor != null) {
            /* Accessor is already created. */
            return superAccessor;
        }

        TokenReference where = body.getTokenReference();
        CSourceClass owner = (CSourceClass)getOwner();
        return superAccessor = owner.addJMethod(makeSuperAccessor(where));
    }

    /**
     * Creates an accessor method to access this method normally.
     * Normally means: not as a constructor and not the super method of this
     * method.
     */
    private JMethodDeclaration makeNormalAccessor(TokenReference where)
        throws PositionedError
    {
        /* If the method is an ordinary method:
         *
         * static <method-return-type> access$xxx(<owner-type> instance,
         *                                        <params...>)
         * {
         *     return instance.<method-name>(<params....>);
         * }
         *
         * // or, for static methods:
         *
         * static <method-return-type> access$xxx(<params...>)
         * {
         *     return <method-name>(<params....>);
         * }
         */
        CSourceClass owner = (CSourceClass)getOwner();
        Type[] accessedParameters = getParameters();

        JFormalParameter[] newParameters;
        boolean hasInstance = false;
        JExpression instanceExpression = null;

        // Fixed: owner.isStatic() -> isStatic() 
        if (!isStatic()) {
            JFormalParameter instanceParam =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     owner.getType(),
                                     "access$instance",
                                     true);
            newParameters =
                new JFormalParameter[accessedParameters.length + 1];
            newParameters[0] = instanceParam;
            instanceExpression =
                new JLocalVariableExpression(where, newParameters[0]);
            hasInstance = true;
        }
        else {
            newParameters =
                new JFormalParameter[accessedParameters.length];
        }


        for (int i = 0; i < accessedParameters.length; i++) {
            int j = i;
            if (hasInstance) {
                j++;
            }
            newParameters[j] =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     accessedParameters[i], "p$" + i,
                                     true);
        }

        JExpression[] args = new JExpression[accessedParameters.length];
        for (int i = 0; i < args.length; i++) {
            int j = i;
            if (hasInstance) {
                j++;
            }
            args[i] = new JLocalVariableExpression(where,
                                                   newParameters[j]);
        }

        JExpression callMethod =
            new JMethodCallExpression
                (where,
                 new JNameExpression(where,
                                     instanceExpression, // <instance>
                                     getIdent()),
                 args);
                                      

        JStatement callMethodStatement;

        if (getReturnType() == StandardType.VOID) {
            callMethodStatement = new JExpressionStatement(where,
                                                           callMethod, null);
        }
        else {
            callMethodStatement = new JReturnStatement(where,
                                                       callMethod, null);
        }

        JMethodDeclaration method = 
            new JMethodDeclaration(where,
                                   ACC_STATIC,
                                   getReturnType(), // return type
                                   ACCESSOR_PREFIX
                                   + owner.getNextSyntheticIndex(),
                                   newParameters,
                                   getThrowables(),
                                   new JBlock(where,
                                              new JStatement[] {
                                                  callMethodStatement },
                                              null),
                                   null, null);

        return method;
    }

    /**
     * Creates an accessor method to access this the super-method of this
     * method. This construct is used for <code>ClassName.super</code>
     * constructs.
     */
    private JMethodDeclaration makeSuperAccessor(TokenReference where)
        throws PositionedError
    {
        /* If the method is an ordinary method:
         *
         * final <method-return-type> access$<owner-name>$xxx(<params...>)
         * {
         *     return super.<method-name>(<params....>);
         * }
         */
        CSourceClass owner = (CSourceClass)getOwner();
        Type[] accessedParameters = getParameters();

        /* The only reason for this parameter is to avoid clashes with
         * method signatures in subclasses.
         */
        JFormalParameter selectorParam =
            new JFormalParameter(where,
                                 ACC_FINAL,
                                 owner.getType(),
                                 "access$selector",
                                 true);
        JFormalParameter[] newParameters =
            new JFormalParameter[accessedParameters.length + 1];
        newParameters[0] = selectorParam;


        for (int i = 0; i < accessedParameters.length; i++) {
            newParameters[i + 1] =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     accessedParameters[i], "p$" + i,
                                     true);
        }

        JExpression[] args = new JExpression[accessedParameters.length];
        for (int i = 0; i < args.length; i++) {
            args[i] = new JLocalVariableExpression(where,
                                                   newParameters[i + 1]);
        }

        JExpression callMethod =
            new JMethodCallExpression
                (where,
                 new JNameExpression(where,
                                     new JSuperExpression(where),
                                     getIdent()),
                 args);
                                      

        JStatement callMethodStatement;

        if (getReturnType() == StandardType.VOID) {
            callMethodStatement = new JExpressionStatement(where,
                                                           callMethod, null);
        }
        else {
            callMethodStatement = new JReturnStatement(where,
                                                       callMethod, null);
        }

        JMethodDeclaration method = 
            new JMethodDeclaration(where,
                                   ACC_FINAL,
                                   getReturnType(), // return type
                                   ACCESSOR_PREFIX
                                   + owner.getNextSyntheticIndex(),
                                   newParameters,
                                   getThrowables(),
                                   new JBlock(where,
                                              new JStatement[] {
                                                  callMethodStatement },
                                              null),
                                   null, null);

        return method;
    }

    /**
     * Creates an accessor method to access this method as a constructor.
     * XXX-EJB: this doesn't work right; creating a static method sounds
     * nice, but it doesn't work if there is a subclass that needs to
     * access this private contstructor.
     * There are two solutions: make the constructor public or package-access,
     * or make the accessor a constructor with a special magic type as
     * an extra argument to ensure that no normal constructor signature could
     * clash with this one (this is how javac and jikes do it)
     */
    private JMethodDeclaration makeConstructorAccessor(TokenReference where)
        throws PositionedError
    {
        /* 'package' <init>(<outer-type> outerInstance, <params...>,
         *                  <fake-type> fake)
         * {
         *     this(outerInstance, params);
         * }
         * For static member classes, outerInstance is omitted.
         */

        CSourceClass owner = (CSourceClass)getOwner();
        CSourceClass outerClass = (CSourceClass)owner.getOwner();

        Type[] origParameters = getParameters();
        JFormalParameter[] syntheticParameters;

        syntheticParameters = new JFormalParameter[origParameters.length + 1];

        for (int i = 0; i < origParameters.length; i++) {
            int j = i;
            syntheticParameters[j] =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     origParameters[i], "p$" + i,
                                     true);
        }

        /* The fake class is used, like with javac, to differentiate between
         * synthetic constructors and any possible constructor that could
         * be declared in the source. Making up a new type that is not
         * assessible from the source is one such way, and it is used
         * by javac too.
         */
        CClass fakeClass = createFakeMemberClass(where);

        JFormalParameter fakeParam = 
            new JFormalParameter(where,
                                 ACC_FINAL,
                                 fakeClass.getType(),
                                 "access$fake", true);
        syntheticParameters[syntheticParameters.length-1] = fakeParam;

        JExpression[] args = new JExpression[origParameters.length];

        for (int i = 0; i < args.length; i++) {
            int j = i;
            args[i] = new JLocalVariableExpression(where,
                                                   syntheticParameters[j]);
        }

        JExpression callConstructor =
            new JExplicitConstructorInvocation(where,
                                               null,
                                               JAV_THIS,
                                               args);

        JStatement statement = 
            new JExpressionStatement(where, callConstructor, null);

        JConstructorBlock syntheticBody =
            new JConstructorBlock(where, new JStatement[] { statement });

        JConstructorDeclaration method = 
            new JConstructorDeclaration(where, 0,
                                        owner.getIdent(),
                                        syntheticParameters,
                                        getThrowables(),
                                        syntheticBody,
                                        null, null);
        
        return method;
    }

    /**
     * ...
     */
    private CClass createFakeMemberClass(TokenReference where)
    {
        CSourceClass owner = (CSourceClass)getOwner();
        String fakeID = String.valueOf(owner.getNextSyntheticIndex()).intern();
        String fakeQualified = owner.getQualifiedName() + "$" + fakeID;

        CSourceClass fakeClazz = new CSourceClass(owner.getCompiler(),
                                                  owner,
                                                  null,
                                                  where,
                                                  0,
                                                  fakeID,
                                                  fakeQualified,
                                                  false);
        fakeClazz.setSynthetic(true);
        fakeClazz.setInnerClasses(new ClassOrInterfaceType[0]);
        fakeClazz.close(new ClassOrInterfaceType[0], StandardType.OBJECT,
                        new CField[0], new CMethod[0]);
        CTopLevel.addSourceClass(fakeClazz);
        ((Main)owner.getCompiler()).addSourceClass(fakeClazz);
        return fakeClazz;
    }

    /**
     * @return the signature of this method.
     */
    public String getSignature()
    {
        Type[] params = getParameters();
        if (getOwner().isInnerClass() && isConstructor()) {
            params = ((CSourceClass)getOwner()).genConstructorArray(params);
        }
        return TypeSignature.genMethodSignature(getReturnType(), params);
    }

    /**
     * Generates code for this method.
     */
    public CodeInfo genBodyCode()
    {
        CodeSequence code = CodeSequence.getCodeSequence();
    
        code.setLineNumber(body.getTokenReference().getLine());
        body.genCode(code);
    
        if ( code.getPC() == 0 || 
             (code.getInstructionAt(code.getPC() - 1).getOpcode()
              != opc_return)
            && getReturnType() == StandardType.VOID) {

            code.plantInstruction(new NoArgInstruction(opc_return));
        }

        CodeInfo info =
            new CodeInfo(code.getInstructionArray(),
                         code.getHandlers(),
                         optimizingLevel >  1 ? null : code.getLineNumbers(),
                         null);    //seq.getLocalVariableInfos());
        code.release();
        body = null;

        Type[] parameters = getParameters();
        int  paramCount = 0;

        for (int i = 0; i < parameters.length; i++) {
            paramCount += Types.getSize(parameters[i]);
        }
        paramCount += Types.getSize(getReturnType());
        paramCount += isStatic() ? 0 : 1;

        info.setParameterCount(paramCount);

        if (optimizingLevel >= 10) {
            return at.dms.backend.Optimizer.optimize
                (null, info, new at.dms.backend.BackendOptions());
        }
        else if (optimizingLevel >= 1) {
            return at.dms.optimize.Optimizer.optimize(info, optimizingLevel);
        }
        else {
            return info;
        }
    }
}
