/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

import at.dms.classfile.ClassInfo;
import at.dms.classfile.ClassPath;
import at.dms.classfile.InnerClassInfo;
import at.dms.classfile.SourceDescription;
import at.dms.compiler.Compiler;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;
import at.dms.util.Utils;

/**
 * This class implements the conceptual directory structure for .class files.
 */
public final class CTopLevel
    extends Utils
{
    /**
     * The fake Compiler instance that is used as the 'Compiler' for binary
     * classes loaded from the classpath.
     */
    private static final Compiler BINARYCLASS_COMPILER = new EmptyCompiler();

    /**
     * Temporary placeholder for a class that does exist on the classpath
     * but hasn't been laoded yet.
     */
    private static CClass CLS_EXISTING =
        new CSourceClass(null, null, null, TokenReference.NO_REF, 0,
                         "<gen>", "<gen>", false);

    /**
     * The compiler in use.
     */
    private static Compiler compiler;

    /**
     * All classes loaded in a compilation session.
     */
    private static Map classCache = new HashMap(2000);

    /**
     * Maps from fully-qualified JVM class names to
     * at.dms.classfile.InnerClassInfo objects. As soon as one of
     * those objects is found for a class, it must be put here, so if the
     * class with the fully qualified name is actually loaded, it gets
     * the right instance of this information along. The most important
     * information in it are the real access flags of an inner class
     * (private, protected, static) that the JVM does not allow in the
     * access modifiers of the ClassInfo of the class.
     */
    private static Map classAnnotation = new HashMap(2000);

    public static void putExtraInfo(String className, InnerClassInfo extraInfo)
    {
        InnerClassInfo currentInfo =
            (InnerClassInfo)classAnnotation.get(className);
        if ( currentInfo == null ) {
            if ( classCache.containsKey(className) ) {
                throw new IllegalStateException("Class already loaded " +
                                                "without extra " +
                                                "InnerClassInfo");
            }
            else {
                classAnnotation.put(className, extraInfo);
            }
        }
        else if ( !currentInfo.equals(extraInfo) ) {
            throw new IllegalStateException("Extra InnerClassInfo placed " +
                                            "given earlier does not match " +
                                            "new extra info.");
        }
    }

    /**
     * Loads class definition from .class file.
     *
     * @param fqName
     * The fully qualified name of the class.
     */
    public static CClass loadClass(String fqName)
    {
        CClass clazz = (CClass)classCache.get(fqName);
        InnerClassInfo innerInfo = (InnerClassInfo)classAnnotation.get(fqName);

        if (clazz != null && clazz != CLS_EXISTING) {
            /* Some information was found in the cache. */
            if (clazz == CClass.CLS_UNDEFINED) {
                return null;
            }
            else if (clazz instanceof CBinaryClass) {
                InnerClassInfo clazzInfo =
                    ((CBinaryClass)clazz).getExtraInfo();
                
                if (clazzInfo == null && innerInfo == null) {
                    return clazz;
                }
                else if ( clazzInfo != null &&
                          clazzInfo.equals(innerInfo) ) {
                    return clazz;
                }
                else {
                    // XXX: use checked exception and be more informative.
                    System.err.println("InnerClassInfo does not " +
                                       "match InnerClassInfo " +
                                       "found earlier for the " +
                                       "same class: " + fqName
                                       + " early is " + clazzInfo +
                                       ", new is " + innerInfo);
                    throw new RuntimeException("InnerClassInfo does not " +
                                               "match InnerClassInfo " +
                                               "found earlier for the " +
                                               "same class.");
                }
            }
            else {
                return clazz;
            }
        }
        else {
            /* Either this class name wasn't heard of before (cache contained
             * null) or it exists but hasn't been loaded (CLS_EXISTING)
             */
            
            ClassInfo file = ClassPath.getClassInfo(fqName, true);
            if (file == null) {
                /* .class file wasn't found, now try to find a source file
                 * that could contain the class name.
                 */
                SourceDescription sourceFile = ClassPath.getSourceFile(fqName);
                if (sourceFile != null) {
                    compiler.loadSourceFile(sourceFile);
                    clazz = (CClass)classCache.get(fqName);
                    /* Hopefully, the result of loading the source file is
                     * that a class with the name that we were looking for
                     * now exists.
                     */
                    if (clazz == null) {
                        /* The source file didn't contain the right class. */
                        clazz = new CBadClass(fqName);
                    }
                    else {
                        /* Loading the source file automatically puts it in
                         * the cache!
                         */
                        return clazz;
                    }
                }
                else {
                    /* The class didn't exist on the classpath, and no
                     * source file could be found for it either.
                     */
                    clazz = new CBadClass(fqName);
                }
            }
            else {
                InnerClassInfo[] infos = file.getInnerClasses();
                if (infos != null) {
                    for (int i = 0; i < infos.length; i++) {
                        putExtraInfo(infos[i].getQualifiedName(), infos[i]);
                    }
                }

                try {
                    /* A .class file _was_ found on the class path. So, create
                     * it as a CBinaryClass.
                     */
                    CBinaryClass binClazz =
                        new CBinaryClass(BINARYCLASS_COMPILER, file,
                                         innerInfo);
                    /* Cache the result. If it was cached earlier as
                     * CLS_EXISTING, then that marker will be replaced with
                     * the real CBinaryClass, CSourceClass.
                     */
                    classCache.put(fqName, binClazz);
                    binClazz.close(file, innerInfo);
                    return binClazz;
                }
                catch (UnpositionedError e) {
                    e.printStackTrace();
                    clazz = new CBadClass(fqName);
                }
            }
            
            /* Cache the result. If it was cached earlier as CLS_EXISTING,
             * then that marker will be replaced with the real CBinaryClass,
             * CSourceClass.
             */
            classCache.put(fqName, clazz);
            
            return clazz;
        }
    }
    
    /**
     * Adds a source class to the set of all loaded classes.
     *
     * @param CClass
     * A class to add (must be a CSourceClass).
     * @return
     * false if name exists for source class as source class in an other file.
     */
    public static boolean addSourceClass(CSourceClass clazz)
    {
        CClass last = (CClass)classCache.put(clazz.getQualifiedName(), clazz);

        if (last == null || last == CLS_EXISTING) {
            return true;
        }
        else if ( !(last instanceof CSourceClass) ) {
            return true;
        }
        else if (clazz.getOwner() != null) {
            return true;
        }
        else {
            return false;
        }
    }

    /**
     * Returns whether a class file that contain the named class named can
     * be found.
     *
     * @param name
     * The name of the class file.
     * @return
     * Whether a class file that contain the named class named can be found.
     */
    public static boolean hasClassFile(String name)
    {
        // already loaded or parsed ?
        CClass clazz = (CClass)classCache.get(name);
        if (clazz == CClass.CLS_UNDEFINED) {
            /* The fact that the class does not exist was cached. */
            return false;
        }
        else if (clazz != null) {
            /* The class has been found before. */
            return true;
        }
        else if (ClassPath.hasClassFile(name)) {
            /* The class hasn't been found before, but it is found now. */

            /* Place a temporary marker in the cache; if hasClassFile is
             * called repeatedly for this same class without loading it
             * into memory, then at least the zip file isn't repeatedly
             * searched to see if it exists.
             */
            classCache.put(name, CLS_EXISTING);
            return true;
        }
        else if (ClassPath.getSourceFile(name) != null) {
            /* There is a source file that should contain the right class. */

            /* Place a temporary marker in the cache; if hasClassFile is
             * called repeatedly for this same class without loading it
             * into memory, then at least the zip file isn't repeatedly
             * searched to see if it exists.
             */
            classCache.put(name, CLS_EXISTING);
            return true;
        }
        else {
            /* The class really doesn't exist; cache this fact to reduce
             * the time necessary for future lookups.
             */
            classCache.put(name, CClass.CLS_UNDEFINED);
            return false;
        }
    }

    /**
     * Removes all source classes.
     */
    public static void initSession(Compiler compiler)
    {
        CTopLevel.compiler = compiler;

        Iterator iterator = classCache.values().iterator();
        while (iterator.hasNext()) {
            CClass clazz = (CClass)iterator.next();
            if (clazz instanceof CSourceClass) {
                iterator.remove();
            }
        }
    }

    /**
     * Ends session.
     */
    public static void endSession(Compiler compiler)
    {
        CTopLevel.compiler = null;
    }
}
