/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.util.ArrayOf2BitValues;

/**
 * Information about possible/definitive assignment of a set of variables.
 * For each variable, it is tracked whether the variable may have been
 * initialized, is definitely unassigned, or is definitely assigned.
 */
public final class CVariableInfo
{
    /**
     * Bit field for "definitely assigned". If an info bitfield has
     * this value in it and also INFO_MAYBE_INITIALIZED, then the meaning
     * is "definitely assigned".
     */
    private static final int DEFINITELY_ASSIGNED = 0x01;

    /**
     * Bit field for "possibly assigned".
     */
    private static final int NOT_DEFINITELY_UNASSIGNED = 0x02;

    /**
     * Assignment information for a whole array of variables.
     */
    private ArrayOf2BitValues infos;

    /**
     * Maintain assignment info for a set of variables.
     */
    public CVariableInfo()
    {
        this.infos = new ArrayOf2BitValues();
    }
  
    /**
     * Maintain assignment info for a set of variables.
     */
    private CVariableInfo(int[] bitArray)
    {
        this.infos = new ArrayOf2BitValues(bitArray);
    }

    /**
     * Returns a deep clone of this object.
     */
    public Object clone()
    {
        return new CVariableInfo((int[])infos.getInternalArray().clone());
    }

    /**
     * Returns assignment info for variable at position <code>pos</code>.
     */
    public int getInfo(int pos)
    {
        return infos.get(pos);
    }

    /**
     * Adds assignment info for variable at position <code>pos</code> to
     * <code>info</code>. The new info bits are merged (with the binary-or
     * operation) with the existing information.
     */
    public void addInfo(int pos, int info)
    {
        int currentInfo = infos.get(pos);
        infos.set(pos, currentInfo | info);
    }

    /**
     * Sets assignment info for variable at position <code>pos</code> to
     * <code>info</code>; any old information about the variables is
     * overwritten.
     */
    public void setInfo(int pos, int info)
    {
        infos.set(pos, info);
    }

    /**
     * Merge two variable info values. Algorithm: if either one is
     * possible (maybe) assigned, then the result is possibly assigned.
     * Only if both a definitely assigned, the result is definitely assigned.
     *
     * @param one
     * The first info bitfield.
     *
     * @param two
     * The second info bitfield.
     *
     * @return
     * The merged result of these information flags.
     */
    public static final int merge(int one, int two)
    {
        return ( ((one | two) & NOT_DEFINITELY_UNASSIGNED)
                 | ((one & two) & DEFINITELY_ASSIGNED) );
    }

    /**
     * Returns info bitfield that means "definitely assigned".
     */
    public static final int initialize()
    // XXX: rename to public field: ASSIGNED
    {
        return DEFINITELY_ASSIGNED | NOT_DEFINITELY_UNASSIGNED;
    }

    /**
     * Returns whether the info bitfield argument means "definitely assigned".
     */
    public static final boolean isDefinitelyAssigned(int info)
    {
        return (info & DEFINITELY_ASSIGNED) != 0;
    }

    /**
     * Returns whether the info bitfield argument means "definitely assigned".
     */
    public static final boolean isInitialized(int info)
    // XXX: rename to isDefinitelyAssigned
    {
        return (info & DEFINITELY_ASSIGNED) != 0;
    }

    /**
     * Returns whether the info bitfield argument means
     * "not definitely unassigned".
     */
    public static final boolean isDefinitelyUnassigned(int info)
    {
        return (info & NOT_DEFINITELY_UNASSIGNED) == 0;
    }

    /**
     * Returns whether the info bitfield argument means
     * "not definitely unassigned".
     */
    public static final boolean mayBeInitialized(int info)
    // XXX: rename to isDefinitelyUnAssigned and reverse condition.
    // (reverse both here and with callers of this method)
    {
        return (info & NOT_DEFINITELY_UNASSIGNED) != 0;
    }

    /**
     * Returns whether the info bitfield argument means both
     * "definitely assigned" <em>and</em> "definitely unassigned". According
     * to the JLS, this can only happen at a position in the code that
     * is unreachable.
     */
    public static final boolean isUnreachable(int info)
    {
        /* Read: check if only definitely assigned flag, and not the
         * not-definitely-unassigned flag, is set in the info value.
         */
        return (info & DEFINITELY_ASSIGNED) == DEFINITELY_ASSIGNED;
    }
}
