/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.type.*;

import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.ClassRefInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.MessageDescription;

/**
 * A representation of the addition operator. See JLS2 15.18.2 and 15.18.1
 * This class represents both numeric and string additions.
 *
 * @author DMS, Erwin Bolwidt
 */
public strictfp class JAddExpression
    extends JBinaryArithmeticExpression
{

    /**
     * Construct a node in the parsing tree
     *
     * @param where the line of this node in the source code
     * @param left the left operand
     * @param right the right operand
     */
    public JAddExpression(TokenReference where,
                          JExpression left, JExpression right)
    {
        super(where, left, right);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The context for evaluation of this expression.
     *
     * @return
     * A pure java expression including promote node.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        check(context,
              left.getType() != StandardType.VOID
              && right.getType() != StandardType.VOID,
              KjcMessages.ADD_BADTYPE,
              left.getType(), right.getType());

        check(context,
              !(left instanceof JTypeNameExpression)
              && !(right instanceof JTypeNameExpression),
              KjcMessages.ADD_BADTYPE,
              left.getType(), right.getType());

        JExpression expr = checkBinaryBody(context);

        if ((expr == this) && !type.equals(StandardType.STRING)) {
            left = left.convertType(type, context);
            right = right.convertType(type, context);
        }

        return expr;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The context for evaluation of this expression.
     *
     * @return
     * A pure java expression including promote node.
     *
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public JExpression checkBinaryBody(CExpressionContext context)
        throws PositionedError
    {
        try {
            type = computeType(left.getType(), right.getType(),
                               KjcMessages.ADD_BADTYPE);

            if (left.isConstant() && right.isConstant()) {
                return constantFolding();
            }
            else {
                return this;
            }
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }
    }

    /**
     * compute the type of this expression according to operands
     * @param leftType the type of left operand
     * @param rightType the type of right operand
     * @return the type computed for this binary operation
     * @exception UnpositionedError this error will be positioned soon
     */
    public static Type computeType(Type left, Type right,
                                    MessageDescription badTypeMessage)
        throws UnpositionedError
    {
        if (left.equals(StandardType.STRING) || right.equals(StandardType.STRING)) {
            if (left.equals(StandardType.VOID) || right.equals(StandardType.VOID)) {
                throw new UnpositionedError(badTypeMessage, left, right);
            }
            return StandardType.STRING;
        }
        else {
            return JBinaryArithmeticExpression.computeType
                (left, right, badTypeMessage);
        }
    }

    /**
     * Peforms the operation of the subclass implementation on constant
     * operands. This method must only be called when both operands are
     * constant.
     *
     * @return
     * A literal, being the result of addition on the constant operands.
     */
    public JExpression constantFolding()
        throws UnpositionedError
    {
        Type leftType = left.getType();
        Type rightType = right.getType();

        
        if (leftType.equals(StandardType.STRING)
            || rightType.equals(StandardType.STRING)) {

            /* Do String-constant folding. */
            String leftString = String.valueOf(left.getLiteral().getValue());
            String rightString = String.valueOf(right.getLiteral().getValue());

            if (leftString.length() + rightString.length() > 65535) {
                /* XXX: this is bad, it makes the expression a non-constant.
                 * Refactoring is needed: JLS constantness needs to be
                 * decoupled from compiler/optimization constantness.
                 */
                return this;
            }

            return new JStringLiteral(getTokenReference(),
                                      leftString.concat(rightString));
        }
        else {
            return super.constantFolding();
        }
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public int compute(int left, int right)
    {
        return left + right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public long compute(long left, long right)
    {
        return left + right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public float compute(float left, float right)
    {
        return left + right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public double compute(double left, double right)
    {
        return left + right;
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitBinaryExpression(this, "+", left, right);
    }

    /**
     * @param type the type of result
     * @return the type of opcode for this operation
     */
    public static int getOpcode(Type type)
    {
        switch (type.getTypeID()) {
        case TID_FLOAT:
            return opc_fadd;
        case TID_LONG:
            return opc_ladd;
        case TID_DOUBLE:
            return opc_dadd;
        default:
            return opc_iadd;
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param code the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (type.equals(StandardType.STRING)) {
            code.plantInstruction
                (new ClassRefInstruction(opc_new, "java/lang/StringBuffer"));
            code.plantInstruction(new NoArgInstruction(opc_dup));
            code.plantInstruction
                (new MethodRefInstruction(opc_invokespecial,
                                          "java/lang/StringBuffer",
                                          JAV_CONSTRUCTOR,
                                          "()V"));
            appendToStringBuffer(code, left);
            appendToStringBuffer(code, right);
            code.plantInstruction
                (new MethodRefInstruction(opc_invokevirtual,
                                          "java/lang/StringBuffer",
                                          "toString",
                                          "()Ljava/lang/String;"));
        }
        else {
            left.genCode(code);
            right.genCode(code);

            code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    private void appendToStringBuffer(CodeSequence code, JExpression expr)
    {
        if ((expr instanceof JAddExpression)
            && (expr.getType().equals(StandardType.STRING))) {
            
            JAddExpression addExpr = (JAddExpression)expr;
            addExpr.appendToStringBuffer(code, addExpr.left);
            addExpr.appendToStringBuffer(code, addExpr.right);
        }
        else {
            expr.genCode(code);

            Type type = expr.getType();
            if (!type.isReference()
                || type.equals(StandardType.STRING)) {
                // This special case does not appear in the JLS 2nd edition.
                // [EJB]
                // || type.equals(StandardType.CHARARRAY)) {

                if (type == StandardType.BYTE || type == StandardType.SHORT) {
                    type = StandardType.INTEGER;
                }

                code.plantInstruction
                    (new MethodRefInstruction(opc_invokevirtual,
                                              "java/lang/StringBuffer",
                                              "append",
                                              "(" + type.getSignature()
                                              + ")Ljava/lang/StringBuffer;"));
            }
            else {
                code.plantInstruction
                    (new MethodRefInstruction(opc_invokevirtual,
                                              "java/lang/StringBuffer",
                                              "append",
                                              "(Ljava/lang/Object;"
                                              + ")Ljava/lang/StringBuffer;"));
            }
        }
    }
}
