/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.codegen.jvm.Types;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * This class implements an access through an array
 * constant values may be folded at compile time
 */
public class JArrayAccessExpression
    extends JExpression
    implements CInitializable
{
    /**
     * Type of "java/lang/ArrayIndexOutOfBoundsException"
     */
    private ClassOrInterfaceType arrayBoundsException;

    /**
     * The array being accessed.
     */
    private JExpression prefix;

    /**
     * The index in the array.
     */
    private JExpression accessor;

    /**
     * The indexed type of the array; also, the result type of the expression.
     */
    private Type type;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param prefix
     * The array being accessed.
     * @param accessor
     * The index in the array.
     */
    public JArrayAccessExpression(TokenReference where,
                                  JExpression prefix,
                                  JExpression accessor)
    {
        super(where);

        this.prefix = prefix;
        this.accessor = accessor;
    }

    /**
     * Returns the indexed type of the array; also, the result type of the
     * expression.
     */
    public Type getType()
    {
        return type;
    }

    /**
     * Return true if this expression is a variable that is already
     * initialized. All array elements are considered to have been initialized
     * when the array is created.
     *
     * @return
     * true if this expression is a variable already initialized.
     */
    public boolean isInitialized(CExpressionContext context)
    {
        // nothing to do in array access 15.12 Array Access Expressions
        return true;
    }

    /**
     * Initialize this object.
     */
    public void initialize(CBodyContext ctxt)
    {
        // nothing to do in array access 15.12 Array Access Expressions
    }

    /**
     *
     */
    public boolean isLValue(CExpressionContext context)
    {
        // nothing to do in array access 15.12 Array Access Expressions
        return true;
    }


    /**
     * Check expression and evaluate and alter context.
     * 15.12 Array Access Expressions
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if ( arrayBoundsException == null ) {
            arrayBoundsException =
                ClassOrInterfaceType.lookup("java/lang/ArrayIndexOutOfBoundsException");

            try {
                arrayBoundsException.resolveType(null);
            }
            catch (UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }
        }

        accessor = accessor.checkExpression(context, false, false);
        Type accessorType = accessor.getType();

        check(context,
              accessorType.isNumeric() &&
              accessorType.toNumeric().unaryPromote() == StandardType.INTEGER,
              KjcMessages.ARRAY_EXPRESSION_INT,
              accessorType);

        prefix = prefix.checkExpression(context, false, false);
        check(context,
              prefix.getType().isArrayType(),
              KjcMessages.ARRAY_PREFIX);
        
        type = ((ArrayType)prefix.getType()).getIndexedType();

        CBodyContext bodyContext = context.getBodyContext();
        CThrowableInfo info = new CThrowableInfo(arrayBoundsException, this,
                                                 bodyContext.makeSnapshot(),
                                                 false); // subTypePossible
        bodyContext.addThrowable(info);
        
        // no constant folding
        // 15.12 Array Access Expressions

        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitArrayAccessExpression(this, prefix, accessor);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        prefix.genCode(code);
        accessor.genCode(code);

        int  opcode;
        switch (getType().getTypeID()) {
        case TID_CLASS:
        case TID_ARRAY:
            opcode = opc_aaload;
            break;
        case TID_INT:
            opcode = opc_iaload;
            break;
        case TID_LONG:
            opcode = opc_laload;
            break;
        case TID_FLOAT:
            opcode = opc_faload;
            break;
        case TID_DOUBLE:
            opcode = opc_daload;
            break;
        case TID_BYTE:
        case TID_BOOLEAN:
            opcode = opc_baload;
            break;
        case TID_CHAR:
            opcode = opc_caload;
            break;
        case TID_SHORT:
            opcode = opc_saload;
            break;
        default:
            fail();
            opcode = -1;
        }

        code.plantInstruction(new NoArgInstruction(opcode));
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genStartStoreCode(CodeSequence code)
    {
        prefix.genCode(code);
        accessor.genCode(code);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genEndStoreCode(CodeSequence code, boolean discardValue)
    {
        if (!discardValue) {
            if (Types.getSize(getType()) == 2) {
                code.plantInstruction(new NoArgInstruction(opc_dup2_x2));
            } else if (Types.getSize(getType()) == 1) {
                code.plantInstruction(new NoArgInstruction(opc_dup_x2));
            }
        }

        int opcode;

        switch (getType().getTypeID()) {
        case TID_CLASS:
        case TID_ARRAY:
            opcode = opc_aastore;
            break;
        case TID_DOUBLE:
            opcode = opc_dastore;
            break;
        case TID_FLOAT:
            opcode = opc_fastore;
            break;
        case TID_LONG:
            opcode = opc_lastore;
            break;
        case TID_BYTE:
        case TID_BOOLEAN:
            opcode = opc_bastore;
            break;
        case TID_SHORT:
            opcode = opc_sastore;
            break;
        case TID_CHAR:
            opcode = opc_castore;
            break;
        default:
            opcode = opc_iastore;
        }
        code.plantInstruction(new NoArgInstruction(opcode));
    }
}
