/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.MultiarrayInstruction;
import at.dms.classfile.ClassRefInstruction;
import at.dms.classfile.NewarrayInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 * This class implements a list of expression used in array initialisation
 * such as: <pre> int a[1][2][3]</pre>
 */
public class JArrayDimsAndInits
    extends JPhylum
{
    /**
     * ...
     */
    private ArrayType type;

    /**
     * ...
     */
    private JExpression[] dims;

    /**
     * ...
     */
    private JArrayInitializer init;

    /**
     * Constructor
     * arr: brackets
     */
    public JArrayDimsAndInits(TokenReference where,
                              JExpression[] dims,
                              JArrayInitializer init)
    {
        super(where);

        this.init = init;
        this.dims = dims;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public Type checkExpression(CExpressionContext context, Type type)
        throws PositionedError
    {
        boolean expr = false;
        boolean hole = true;

        for (int i = 0; i < dims.length; i++) {
            if (dims[i] != null) {
                expr = true;
                dims[i] = dims[i].checkExpression(context);
                Type dimType = dims[i].getType();
                check(context,
                      dimType.isNumeric() &&
                      ( dimType.toNumeric().unaryPromote() ==
                        StandardType.INTEGER ),
                      KjcMessages.ARRAY_BADTYPE, dimType);
                hole = false;
            }
        }

        check(context, !hole || init != null, KjcMessages.ARRAY_NEW_NOSIZE);

        if (this.type == null) {
            this.type = new ArrayType(type, dims.length);
        }

        if (init != null) {
            check(context, !expr, KjcMessages.ARRAY_NEW_NOSIZE);
            init.setType(this.type);
            init.checkExpression(context);
        }
        return this.type;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitArrayDimsAndInit(this, dims, init);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (code.discardValue()) {
            return;
        }

        if (init == null) {
            allocArray(code, type, dims);
        }
        else {
            init.genCode(code);
        }
    }


    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public static void allocArray(CodeSequence code,
                                  ArrayType type, JExpression[] dims)
    {
        for (int i = 0; i < dims.length; i++) {
            if (dims[i] != null) {
                dims[i].genCode(code);
            }
        }

        if (type.getDimensions() > 1) {
            int  count = 0;
            for (int i = 0; i < dims.length && dims[i] != null; i++) {
                count++;
            }

            String signature = type.getSignature();
            code.plantInstruction(new MultiarrayInstruction(signature, count));
        }
        else {
            if ( type.getIndexedType().isClassOrInterfaceType() ) {
                ClassOrInterfaceType indexedType =
                    type.getIndexedType().toClassOrInterface();
                String qualifiedName = indexedType.getQualifiedName();
                code.plantInstruction(new ClassRefInstruction(opc_anewarray,
                                                              qualifiedName));
            }
            else {
                byte num;

                switch (type.getIndexedType().getTypeID()) {
                case TID_BYTE:
                    num = 8;
                    break;
                case TID_BOOLEAN:
                    num = 4;
                    break;
                case TID_CHAR:
                    num = 5;
                    break;
                case TID_SHORT:
                    num = 9;
                    break;
                case TID_INT:
                    num = 10;
                    break;
                case TID_LONG:
                    num = 11;
                    break;
                case TID_FLOAT:
                    num = 6;
                    break;
                case TID_DOUBLE:
                    num = 7;
                    break;
                default:
                    fail();
                    num = -1;
                }
                code.plantInstruction(new NewarrayInstruction(num));
            }
        }
    }
}
