/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.type.*;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class implements a constant list of expressions used in array
 * initialisation. The method allowsAssignmentConversion is redefined in order to follow,
 * (JLS 5.2)
 */
public class JArrayInitializer
    extends JExpression
{

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where the line of this node in the source code
     * @param v a vector of expression {e1, e2, e3, e4}
     */
    public JArrayInitializer(TokenReference where, JExpression[] elems)
    {
        super(where);

        this.elems = elems;
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public Type getType()
    {
        return type;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param type
     * The type of this array. The type must have been checked already.
     */
    public void setType(Type type)
    {
        assert(type.isResolved(), type.toString());
        this.type = type;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        assert(type != null);
        assert(type.isArrayType());

        Type elemsType = ((ArrayType)type).getIndexedType();

        if (elemsType.isArrayType()) {
            for (int i = 0; i < elems.length; i++) {
                if (elems[i] instanceof JArrayInitializer) {
                    ((JArrayInitializer)elems[i]).setType(elemsType);
                }
                elems[i] = elems[i].checkExpression(context);
                check(context, elems[i].allowsAssignmentConversion(elemsType),
                      KjcMessages.ARRAY_INIT_BADTYPE,
                      elemsType, elems[i].getType());
            }
        }
        else {
            for (int i = 0; i < elems.length; i++) {
                elems[i] = elems[i].checkExpression(context);
                check(context, elems[i].allowsAssignmentConversion(elemsType),
                      KjcMessages.ARRAY_INIT_BADTYPE,
                      elemsType, elems[i].getType());
                elems[i] = elems[i].convertType(elemsType, context);
            }
        }

        return this;
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitArrayInitializer(this, elems);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        JArrayDimsAndInits.allocArray
            (code,
             (ArrayType)getType(),
             new JExpression[] { new JOrdinalLiteral(getTokenReference(),
                                                     elems.length,
                                                     StandardType.INTEGER)});
        int  opcode;
        
        switch (((ArrayType)getType()).getIndexedType().getTypeID()) {
        case TID_CLASS:
        case TID_ARRAY:
            opcode = opc_aastore;
            break;
        case TID_INT:
            opcode = opc_iastore;
            break;
        case TID_LONG:
            opcode = opc_lastore;
            break;
        case TID_FLOAT:
            opcode = opc_fastore;
            break;
        case TID_DOUBLE:
            opcode = opc_dastore;
            break;
        case TID_BYTE:
        case TID_BOOLEAN:
            opcode = opc_bastore;
            break;
        case TID_CHAR:
            opcode = opc_castore;
            break;
        case TID_SHORT:
            opcode = opc_sastore;
            break;
        default:
            fail();
            opcode = -1;
        }

        for (int i = 0; i < elems.length; i++) {
            //if (i != elems.length - 1)
            code.plantInstruction(new NoArgInstruction(opc_dup));
            code.plantInstruction(new PushLiteralInstruction(i));
            elems[i].genCode(code);
            code.plantInstruction(new NoArgInstruction(opcode));
        }
    }

    private Type type;
    private JExpression[] elems;
}
