/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * A 'ArrayLength' expression
 */
public class JArrayLengthExpression
    extends JExpression
{
    /**
     * Type of "java/lang/NullPointerException"
     */
    private ClassOrInterfaceType nullPointerException;

    /**
     * Syntax tree prefix of ".length" expression.
     */
    private JExpression prefix;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param prefix
     * The left expression like t.this.
     */
    public JArrayLengthExpression(TokenReference where,
                                  JExpression prefix)
    {
        super(where);
        this.prefix = prefix;
    }

    /**
     * Compute the type of this expression. Called after parsing.
     *
     * @return
     * The type of this expression.
     */
    public Type getType()
    {
        return StandardType.INTEGER;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if ( nullPointerException == null ) {
            nullPointerException =
                ClassOrInterfaceType.lookup("java/lang/NullPointerException");

            try {
                nullPointerException.resolveType(null);
            }
            catch (UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }
        }

        assert(prefix != null);
        prefix = prefix.checkExpression(context);
        check(context,
              prefix.getType().isArrayType(),
              KjcMessages.ARRAY_LENGTH_BADTYPE);

        /* Querying the length of an array can result in a
         * NullPointerException, so a catch clause for NullPointerException
         * would be reachable.
         */
        CBodyContext bodyContext = context.getBodyContext();
        CThrowableInfo info = new CThrowableInfo(nullPointerException, this,
                                                 bodyContext.makeSnapshot(),
                                                 false); // subTypePossible
        bodyContext.addThrowable(info);

        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitArrayLengthExpression(this, prefix);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());
        boolean discardValue = code.discardValue();

        code.setDiscardValue(false);
        prefix.genCode(code);
        if (discardValue) {
            code.plantInstruction(new NoArgInstruction(opc_pop));
        }
        else {
            code.plantInstruction(new NoArgInstruction(opc_arraylength));
        }
    }
}
