/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import at.dms.compiler.CWarning;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * This class represents a java class in the syntax tree.
 */
public class JClassDeclaration
    extends JTypeDeclaration
{
    /**
     * Assignment information for static fields.
     */
    private CVariableInfo staticInfo;

    /**
     * Context used for checking members of this class.
     */
    private CClassContext self;

    /**
     * Name of superclass.
     */
    protected String superName;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param modifiers
     * The modifiers of this class.
     * @param ident
     * The simple name of this class.
     * @param superName
     * The name of the super class of this class.
     * @param interfaces
     * The name of this class' interfaces.
     * @param javadoc
     * The javadoc of this class.
     * @param comment
     * Implementation comments of this class.
     */
    public JClassDeclaration(TokenReference where,
                             int modifiers,
                             String ident,
                             String superName,
                             ClassOrInterfaceType[] interfaces,
                             JFieldDeclaration[] fields,
                             JMethodDeclaration[] methods,
                             JTypeDeclaration[] inners,
                             JPhylum[] initializers,
                             JavadocComment javadoc,
                             JavaStyleComment[] comment)
    {
        super(where, modifiers, ident, interfaces, fields, methods,
              inners, initializers, javadoc, comment);
        this.superName = superName;
    }

    /**
     * Sets the super class.
     */
    public void setSuperClass(String superName)
    {
        this.superName = superName;
    }

    /**
     * Second pass (quick), check interface looks good.
     * This pass is just a tuning pass in order to create informations about
     * exported elements such as Classes, Interfaces, Methods, Constructors
     * and Fields.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInterface(final CContext context)
        throws PositionedError
    {
        boolean inInterface = false;
        CClassContext classContext = context.getClassContext();
        if (classContext != null) {
            inInterface = classContext.getCClass().isInterface();
        }

        /* Member classes defined in interfaces always have 'public' and
         * 'static' modifiers set, whether declared in the source or not.
         */
        if ( inInterface ) {
            setModifiers(getModifiers() | ACC_PUBLIC | ACC_STATIC);
            getCClass().setModifiers(getCClass().getModifiers() |
                                     ACC_PUBLIC | ACC_STATIC);
        }

        int modifiers = getModifiers();

        if (getOwner() != null) {
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC + ACC_PRIVATE + ACC_PROTECTED
                                 + ACC_ABSTRACT + ACC_FINAL + ACC_STATIC
                                 + ACC_STRICT),
                  KjcMessages.INVALID_CLASS_MODIFIERS);
        }
        else {
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC + ACC_ABSTRACT + ACC_FINAL
                                 + ACC_STRICT),
                  KjcMessages.INVALID_CLASS_MODIFIERS);
        }

        if (inInterface) {
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC + ACC_ABSTRACT + ACC_STATIC +
                                 ACC_FINAL + ACC_STRICT),
                  KjcMessages.CLASS_FLAGS_IN_INTERFACE);
        }

        check(context,
              !hasMoreThanOneFlag(modifiers, ACC_FINAL + ACC_ABSTRACT),
              KjcMessages.MODIFIERS_ABSTRACT_AND_FINAL);

        check(context,
              !hasMoreThanOneFlag(modifiers,
                                  ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE),
              KjcMessages.MULTIPLE_ACCESS_MODIFIERS);

        CClass superClass = null;
        ClassOrInterfaceType superType = null;

        if (superName == null) {
            if (getSourceClass().getType().equals(StandardType.OBJECT)) {
                // java/lang/Object
                // superClass = null;
                // superType = null;
            }
            else {
                superClass = StandardType.OBJECT.getCClass();
                superType = StandardType.OBJECT;
            }
        }
        else {
            // find name
            try {
                superType = ClassOrInterfaceType.lookup(superName);
                superType.resolveType(context);
                superClass = superType.getCClass();
            }
            catch (UnpositionedError cue) {
                cue.printStackTrace();
                throw cue.addPosition(getTokenReference());
            }
        }

        // check access
        if (superClass != null) {
            if (!superClass.isInnerClass()) {
                /* Superclass is toplevel class */
                check(context,
                      superClass.isPublic()
                      || ( getSourceClass().getPackage()
                           == superClass.getPackage() ),
                      KjcMessages.CLASS_ACCESSPARENT,
                      superName);
            }
            else {
                if (superClass.isProtected() && getCClass().isInnerClass()
                    && getOwner().descendsFrom(superClass.getOwner())) {
                    // Okay!
                }
                else {
                    check(context,
                          superClass.isPublic()
                          || ( getSourceClass().getPackage()
                               == superClass.getPackage() ),
                          KjcMessages.CLASS_ACCESSPARENT,
                          superName);
                }
            }

            check(context,
                  !superClass.isFinal(),
                  KjcMessages.CLASS_PARENT_FINAL,
                  superName);

            check(context,
                  !superClass.isInterface(),
                  KjcMessages.CLASS_EXTENDS_INTERFACE,
                  superName);
        }

        CClassContext self = createContext(context);

        JInitializerDeclaration staticInit = constructInitializers(true);
        if (staticInit != null) {
            // System.err.println("Setting a static initializer in " +
            //                   getIdent());
            staticInit.checkInterface(self);
            setStaticInitializer(staticInit);
        }
        else {
            // System.err.println("No static initializer in " + getIdent());
        }

        Iterator iterator = getMethodList().iterator();
        boolean hasConstructor = false;
        while (iterator.hasNext()) {
            JMethodDeclaration method = (JMethodDeclaration)iterator.next();
            if (method instanceof JConstructorDeclaration) {
                // usually constructors are at the top of the file
                // so its not as inefficient as it's look
                hasConstructor = true;
                break;
            }
        }

        if (!hasConstructor && getDefaultConstructor() == null) {
            setDefaultConstructor(constructDefaultConstructor());
        }

        JInitializerDeclaration instanceInit = constructInitializers(false);
        if (instanceInit != null) {
            instanceInit.checkInterface(self);
            setInstanceInitializer(instanceInit);
        }

        getCClass().setSuperClass(superType);
        super.checkInterface(context, superType);
    }

    /**
     * Check that initializers are correct.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInitializers(CContext context)
        throws PositionedError
    {
        self = createContext(context);

        JInitializerDeclaration staticInit = getStaticInitializer();

        if (staticInit != null) {
            staticInit.checkInitializer(self);
            staticInfo = self.getVariableInfo();
        }

        CClass clazz = getCClass();
        CClass superClazz = clazz.getSuperClass();

        if (superClazz != null) {
            /* A class cannot extend itself. */
            if ( clazz.equals(superClazz) ) {
                reportTrouble(context,
                              KjcMessages.CLASS_EXTENDS_SELF, getIdent());
            }
            /* If this class descends from its superclass and vice versa,
             * there is a cycle.
             */
            else if ( clazz.descendsFrom(superClazz) &&
                 superClazz.descendsFrom(clazz) ) {
                reportTrouble(context,
                              KjcMessages.CLASS_CIRCULARITY, getIdent());
            }
            /* A class cannot extend one of its member classes. */
            else if ( superClazz.descendsFromInside(clazz) ) {
                reportTrouble(context,
                              KjcMessages.CLASS_EXTENDS_MEMBER, getIdent());
            }
        }

        ClassOrInterfaceType[] interfaces = getInterfaces();
        for (int i = 0; i < interfaces.length; i++) {
            /* A class cannot implement one of its member interfaces. */
            if ( interfaces[i].getCClass().descendsFromInside(clazz) ) {
                reportTrouble(context,
                              KjcMessages.CLASS_IMPLEMENTS_MEMBER, getIdent());
            }
        }

        // Check inners
        Iterator iterator = getInnerList().iterator();
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            inner.checkInitializers(self);

        }

        super.checkInitializers(context);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkTypeBody(CContext context)
        throws PositionedError
    {
        try {
            CVariableInfo instanceInfo = null;

            /* Array of CVariableInfo objects resulting from constructors. */
            CVariableInfo[] constructorsInfo;

            /* List of CVariableInfo objects that is added to and is converted
             * to constructorsInfo when ready. */
            List constructorsInfoList = new ArrayList(5);

            /* Check the instance initializer. */
            JInitializerDeclaration instanceInit = getInstanceInitializer();
            if (instanceInit != null) {
                self.markAllStaticFieldsAsInitialized();
                instanceInit.checkInitializer(self);
                instanceInfo = self.getVariableInfo();
            }

            /* Mark all fields as 'fully declared'. */
            Iterator iterator = getFieldList().iterator();
            while (iterator.hasNext()) {
                JFieldDeclaration declaration =
                    (JFieldDeclaration)iterator.next();
                ((CSourceField)declaration.getField()).setFullyDeclared(true);
            }

            /* Check inner classes. */
            iterator = getInnerList().iterator();
            while (iterator.hasNext()) {
                JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
                try {
                    inner.checkTypeBody(self);
                }
                catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            /* Compile the constructors. */
            iterator = getMethodList().iterator();
            while (iterator.hasNext()) {
                JMethodDeclaration method =
                    (JMethodDeclaration)iterator.next();
                try {
                    /* Check if method is a constructor. */
                    if (method instanceof JConstructorDeclaration) {
                        /* Important: set assignment info from instance
                         * initializer before checking; getVariableInfo
                         * method clears old info!!
                         */
                        if (instanceInfo != null) {
                            CVariableInfo clone =
                                (CVariableInfo)instanceInfo.clone();
                            self.setVariableInfo(clone);
                        }
                        method.checkStatement(self);
                        constructorsInfoList.add(self.getVariableInfo());
                    }
                }
                catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            /* Check for circular explicit constructor invocations. */
            iterator = getMethodList().iterator();
            while (iterator.hasNext()) {
                JMethodDeclaration method =
                    (JMethodDeclaration)iterator.next();
                try {
                    /* Check if method is a constructor. */
                    if (method instanceof JConstructorDeclaration) {
                        CSourceMethod constructor = (CSourceMethod)
                            ((JConstructorDeclaration)method).getMethod();

                        checkCyclicalConstructor(method.getTokenReference(),
                                                 constructor);
                    }
                }
                catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            if (getDefaultConstructor() == null) {
                constructorsInfo = (CVariableInfo[])
                    constructorsInfoList.toArray
                    (new CVariableInfo[constructorsInfoList.size()]);
            }
            else {
                getDefaultConstructor().checkStatement(self);
                constructorsInfo =
                    new CVariableInfo[] { self.getVariableInfo() };
            }

            /* For the purpose of compiling methods, all final fields
             * are marked as initialized. They are checked later to see
             * if they really were correctly initialized in the constructors.
             */
            self.markAllFieldsAsInitialized();

            /* Compile the methods. */
            iterator = getMethodList().iterator();
            while (iterator.hasNext()) {
                JMethodDeclaration method =
                    (JMethodDeclaration)iterator.next();
                try {
                    if ( !(method instanceof JConstructorDeclaration) ) {
                        method.checkStatement(self);
                    }
                }
                catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            /* Check if methods are used. */
            iterator = getMethodList().iterator();
            while (iterator.hasNext()) {
                JMethodDeclaration method =
                    (JMethodDeclaration)iterator.next();
                CSourceMethod checkingMethod =
                    (CSourceMethod)method.getMethod();

                if (!checkingMethod.isUsed()
                    && !checkingMethod.getIdent().equals(JAV_CONSTRUCTOR)) {

                    context.reportTrouble
                        (new CWarning(method.getTokenReference(), 
                                      KjcMessages.UNUSED_PRIVATE_METHOD, 
                                      checkingMethod.getIdent()));
                }
            }

            /* Check if fields are used. */
            iterator = getFieldList().iterator();
            while (iterator.hasNext()) {
                JFieldDeclaration declaration =
                    (JFieldDeclaration)iterator.next();

                if (!((CSourceField)declaration.getField()).isUsed()) {

                    context.reportTrouble
                        (new CWarning(declaration.getTokenReference(), 
                                      KjcMessages.UNUSED_PRIVATE_FIELD, 
                                      declaration.getField().getIdent()));
                }
            }

            self.close(this, staticInfo, instanceInfo, constructorsInfo);

            super.checkTypeBody(context);
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
    
        self = null;
    }

    /**
     * Checks whether a constructor contains cyclical explicit-this
     * invocations.
     * A PositionedError is thrown in the case of such a cyclical structure
     * is detected.
     *
     * @param where
     * The TokenReference of the constructor to check.
     * @param constructor
     * The CMethod of the constructor to check.
     * @throws PositionedError
     * When a cycle is detected.
     */
    private void checkCyclicalConstructor(TokenReference where,
                                          CSourceMethod constructor)
        throws PositionedError
    {
        List calledConstructors = new ArrayList(3);
        while (constructor != null) {
            if (calledConstructors.contains(constructor)) {
                throw new PositionedError
                    (where,
                     KjcMessages.CONSTRUCTOR_CYCLICAL_EXPLICIT_CALLS,
                     constructor.getOwner().getIdent());
            }

            calledConstructors.add(constructor);
            CMethod invoked = constructor.getExplicitlyInvokedContructor();

            if ( invoked != null &&
                 invoked.getOwner().equals(constructor.getOwner()) ) {
                constructor = (CSourceMethod)invoked;
            }
            else {
                /* Any relevant call chain ends when an explicit invocation of
                 * the super class is made.
                 */
                return;
            }
        }
    }

    /**
     * Creates a CClassContext that takes this class as its class context.
     *
     * @param parent
     * The parent context or null.
     * @return
     * A CClassContext that represents this context.
     */
    public CClassContext createContext(CContext parent)
    {
        return new CClassContext(parent, getSourceClass(), this);
    }

    /**
     * Creates a default constructor for classes that don't have declared
     * constructors in their source code.
     */
    private JConstructorDeclaration constructDefaultConstructor()
    {
        TokenReference where = getTokenReference();
        JConstructorBlock defaultBlock =
            new JConstructorBlock(where, new JStatement[0]);

        /* See JLS2 8.8.7, access modifiers of default constructors.
         * Basically, access modifiers of default constructor should be
         * the same as those of the class in which it is generated,
         * wrt public/private/protected/<package>
         */
        int syntheticModifiers = ( getCClass().getModifiers() & 
                                   ( ACC_PUBLIC | ACC_PRIVATE | ACC_PROTECTED)
                                   );

        return new JConstructorDeclaration(where, syntheticModifiers,
                                           getIdent(),
                                           JFormalParameter.EMPTY,
                                           ClassOrInterfaceType.EMPTY,
                                           defaultBlock,
                                           null,
                                           null);
    }

    /**
     * Collects all initializers and builds a single method.
     *
     * @param isStatic
     * Whether to collect static or instance initializers.
     */
    private JInitializerDeclaration constructInitializers(boolean isStatic)
    {
        List statementList = new ArrayList(5);
        boolean needGen = false;

        Iterator iterator = getBodyList().iterator();
        while (iterator.hasNext()) {
            JPhylum element = (JPhylum)iterator.next();
            if (element instanceof JClassBlock) {
                JClassBlock classBlock = (JClassBlock)element;
                if (classBlock.isStaticInitializer() == isStatic) {
                    statementList.add(element);
                    needGen = true;
                }
            }
            else {
                if (element instanceof JFieldDeclaration) {
                    JFieldDeclaration declaration = (JFieldDeclaration)element;
                    if (declaration.getVariable().isStatic() == isStatic) {
                        needGen |= declaration.needInitialization();
                        statementList.add
                            (new JClassFieldDeclarator(getTokenReference(),
                                                       declaration));
                    }
                }
            }
        }

        if (statementList.isEmpty()) {
            return null;
        }

        JStatement[] statements = (JStatement[])
            statementList.toArray(new JStatement[statementList.size()]);

        return new JInitializerDeclaration
            (getTokenReference(),
             new JClassBlock(getTokenReference(), false, statements),
             isStatic, !needGen);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);

        p.visitClassDeclaration(this,
                                getCClass().getModifiers(),
                                getCClass().getIdent(),
                                superName, 
                                getInterfaces(),
                                getBody(),
                                getMethods(),
                                getInners());
    }

    /**
     * Generate the code in pure java form.
     *
     * @param p 
     * The pretty printer into which the code is generated.
     */
    public void genInnerJavaCode(KjcPrettyPrinter p)
    {
        super.accept(p);

        p.visitInnerClassDeclaration(this,
                                     getModifiers(),
                                     getIdent(),
                                     superName,
                                     getInterfaces(),
                                     getInners(),
                                     getBody(),
                                     getMethods());
    }
}
