/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.Instruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.FieldRefInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * A 'int.class' expression
 */
public class JClassExpression extends JExpression
{
    /**
     * The resulting type.
     */
    private Type type;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JClassExpression(TokenReference where, Type type)
    {
        super(where);
        this.type = type;
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JClassExpression(TokenReference where,
                            JExpression expr,
                            int bounds)
    {
        super(where);

        JNameExpression nameExpr = (JNameExpression)expr;
        String qname = nameExpr.getQualifiedName();
        type = ClassOrInterfaceType.lookup(qname.replace('.', '/'));
        if (bounds > 0) {
            type =new ArrayType(type, bounds);
        }
    }

    /**
     * Compute the type of this expression.
     *
     * @return
     * The type of this expression.
     */
    public Type getType()
    {
        return StandardType.CLASS;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        try {
            type.resolveType(context);
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }
        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitClassExpression(this, type);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        String className = "UNKNOWN";
        /* In theory, this should never throw an exception, since at 
         * compilation time it was determined that the type existed.
         * So the only thing that may be thrown is a subclass of
         * Error.
         * However, at the moment ClassNotFoundException can be thrown :-(
         * Also, the whole expression may be optimized to use a static private
         * field to look the class up only once, like javac generates.
         */
        switch (type.getTypeID()) {
        case TID_CLASS:
        case TID_ARRAY:
            ReferenceType reference = type.toReference();
            Instruction push = new PushLiteralInstruction(reference.
                                                          getQualifiedName());
            code.plantInstruction(push);
            Instruction invoke =
                new MethodRefInstruction(opc_invokestatic,
                                         "java/lang/Class",
                                         "forName",
                                         "(Ljava/lang/String;)" +
                                         "Ljava/lang/Class;");
            code.plantInstruction(invoke);
            return;
        case TID_INT:
            className = "java/lang/Integer";
            break;
        case TID_LONG:
            className = "java/lang/Long";
            break;
        case TID_FLOAT:
            className = "java/lang/Float";
            break;
        case TID_DOUBLE:
            className = "java/lang/Double";
            break;
        case TID_BYTE:
            className = "java/lang/Byte";
            break;
        case TID_BOOLEAN:
            className = "java/lang/Boolean";
            break;
        case TID_CHAR:
            className = "java/lang/Character";
            break;
        case TID_SHORT:
            className = "java/lang/Short";
            break;
        }

        Instruction primitiveClass =
            new FieldRefInstruction(opc_getstatic, className, "TYPE",
                                    StandardType.CLASS.getSignature());
        code.plantInstruction(primitiveClass);
    }
}
