/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.codegen.jvm.Types;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * A field access ('getExpression().t' or 't').
 * XXX: note: why is it necessary to private a CContext subclass object in
 * the check* methods; why not set the context once when it is know, so that
 * it is still available in the genCode methods etc.? - EJB
 */
public class JClassFieldExpression
    extends JExpression
    implements CInitializable
{
    /**
     * Type of "java/lang/NullPointerException"
     */
    private ClassOrInterfaceType nullPointerException;

    /**
     * Whether this expression had a prefix _before_ expression checking.
     */
    private boolean wasSimpleName;

    /**
     * Expression prefixing the field name.
     */
    private JExpression prefix;

    /**
     * The field name.
     */
    private String ident;

    /**
     * The field export that this expression refers to.
     */
    private CField field;

    /**
     * True if the field is not directly accessible but only by creating
     * an accessor method in the class that contains the field.
     */
    private boolean mustUseAccessorMethod = false;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JClassFieldExpression(TokenReference where,
                                 JExpression prefix, String ident)
    {
        super(where);

        if (prefix == null) {
            this.wasSimpleName = true;
        }
        else {
            this.wasSimpleName = false;
        }

        this.prefix = prefix;
        this.ident = ident;
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JClassFieldExpression(TokenReference where, String ident)
    {
        this(where, null, ident);

        this.wasSimpleName = true;
    }

    /**
     * Returns the name of the classfield that this expression refers to.
     */
    public String getIdent()
    {
        return ident;
    }

    /**
     * Compute the type of this expression.
     * Should be called after parsing.
     *
     * @return the type of this expression.
     */
    public Type getType()
    {
        return field.getType();
    }

    /**
     * Checks whether this JLS says that this expression is a constant.
     *
     * Quating JLS 15.28: 
     * <BLOCKQUOTE>
     * A compile-time constant expression is an expression denoting a value of
     * primitive type or a String that is composed using only the following:
     * <UL>
     * <LI> <EM> ... </EM>
     * <LI> Simple names that refer to final variables whose initializers are
     *      constant expressions.
     * <LI> Qualified names of the form TypeName . Identifier that refer to
     *      final variables whose initializers are constant expressions.
     *
     * @return
     * Whether this expression is constant.
     */
    public boolean isConstant()
    {
        /* Simple names are okay, and qualified names with a
         * TypeNameExpression as a prefix. Other final expressions
         * aren't constant. (See JLS2)
         */
        final boolean typeNameOrNoPrefix =
            wasSimpleName || (prefix instanceof JTypeNameExpression);

        return ( field.isFinal() && field.getValue() != null &&
                 field.getValue().isConstant() && typeNameOrNoPrefix );
    }

    /**
     * Returns true if this field (still) accepts assignment.
     * A final field only accepts assignment while it is definitely unassigned.
     */
    public boolean isLValue(CExpressionContext context)
    {
        if (!field.isFinal() || !(field instanceof CSourceField)) {
            return true;
        }

        CBodyContext bodyContext = context.getBodyContext();
        CSourceField sourceField = (CSourceField)field;

        if ( bodyContext.getClassContext().getCClass() == field.getOwner()
             && !sourceField.isFullyDeclared() ) {

            int info = bodyContext.getFieldInfo(sourceField.getPosition());
            return !CVariableInfo.mayBeInitialized(info);
        }
        else {
            return true;
        }
    }

    /**
     * Returns true if this field is already initialized.
     */
    public boolean isInitialized(CExpressionContext context)
    {
        // if (!(field instanceof CSourceField) || field.isStatic()) {
        if (!(field instanceof CSourceField)) {
            return true;
        }

        CBodyContext bodyContext = context.getBodyContext();
        CSourceField sourceField = (CSourceField)field;

        if ( bodyContext.getClassContext().getCClass() == field.getOwner()
             && !sourceField.isFullyDeclared() ) {

            return CVariableInfo.isInitialized
                (bodyContext.getFieldInfo(sourceField.getPosition()));
        }
        else {
            return true;
        }
    }

    /**
     * Signals that this field will receive a value.
     */
    public void initialize(CBodyContext context)
    {
        if ((field instanceof CSourceField)
            && (context.getClassContext().getCClass() == field.getOwner()
                && !((CSourceField)field).isFullyDeclared())) {

            context.addFieldInfo(((CSourceField)field).getPosition(),
                                 CVariableInfo.initialize());

            /* Now it is prudent to ensure a setter method; a setter method
             * isn't necessary if the field is never going to be set..
             */
            if (mustUseAccessorMethod) {
                /* Make sure that the setter method is generated in checking
                 * phase, so it will end up in the class file. It is known
                 * now because this method is called that this expression will
                 * be used for assignment; otherwise only a getter is
                 * needed.
                 */
                try {
                    ((CSourceField)field).getSetterAccessor();
                }
                catch (PositionedError e) {
                    // Shouldn't happen; no invalid inner accessors should
                    // be generated.
                    throw new RuntimeException(e.toString());
                }
            }
        }
    }

    /**
     * Returns the field-export that this expression refers to.
     */
    public CField getField()
    {
        return field;
    }

    /**
     * Returns the literal value of this field, if this is a constant
     * expression.
     */
    public JLiteral getLiteral()
    {
        return (JLiteral)field.getValue();
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if ( nullPointerException == null ) {
            nullPointerException =
                ClassOrInterfaceType.lookup("java/lang/NullPointerException");

            try {
                nullPointerException.resolveType(null);
            }
            catch (UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }
        }

        CClass local = context.getClassContext().getCClass();

        findPrefix(local, context);
        checkAccess(local, context);

        check(context,
              field.isStatic() || !(prefix instanceof JTypeNameExpression),
              KjcMessages.FIELD_NOSTATIC, ident);

        if (field.isStatic() && !(prefix instanceof JTypeNameExpression)) {
            context.reportTrouble
                (new CWarning(getTokenReference(),
                              KjcMessages.INSTANCE_PREFIXES_STATIC_FIELD,
                              ident));
        }


        /* Check if there is a chance that the final field has not yet been
         * initialized. This check is suppressed for synthetic fields, since
         * the compiler does sometimes generate initialization code later,
         * such as for outer-local variables, and that would
         * create problems if we don't suppress this check for them.
         */

        /* Strange enough, javac and jikes only detect this problem in the
         * constructor if there is no prefix or if the prefix is "this".
         * Otherwise they don't, so
         * constructor() { MyType alias = this; use(alias.unInitFinalField); }
         * works fine, while
         * constructor() { use(this.unInitFinalField); }
         * doesn't.
         * Let's call the first construct (where a prefix other than this
         * is used, even if it is an alias for this),
         * "somebody elses' problem", because it seems the easy way out.
         */
        boolean somebodyElsesProblem = false;
        if (prefix != null && !(prefix instanceof JThisExpression)) {
            somebodyElsesProblem = true;
        }
            
        check(context,
              isInitialized(context) || context.isLeftSide()
              || field.isSynthetic() || somebodyElsesProblem,
              KjcMessages.UNINITIALIZED_FINAL_FIELD,
              ident);

        if (field instanceof CSourceField && !context.isDiscardValue()) {
            ((CSourceField)field).setUsed();
        }

        if (!field.isStatic()) {
            /* Access to non-static fields can result in a
             * NullPointerException, so a catch clause for NullPointerException
             * would be reachable.
             */
            CBodyContext bodyContext = context.getBodyContext();
            CThrowableInfo info =
                new CThrowableInfo(nullPointerException, this,
                                   bodyContext.makeSnapshot(),
                                   false); // subTypePossible
            bodyContext.addThrowable(info);
        }

        if (isConstant()) {
            return field.getValue();
        }
        else {
            return this;
        }
    }

    /**
     * Find the type of the prefix.
     *
     * @exception PositionedError  Error caught as soon as possible
     */
    protected void findPrefix(CClass local, CExpressionContext context)
        throws PositionedError
    {
        if (prefix != null) {
            prefix = prefix.checkExpression(context);
            check(context,
                  prefix.getType().isClassOrInterfaceType(),
                  KjcMessages.FIELD_BADACCESS, prefix.getType());
        }
        else {
            try {
                field = context.lookupField(ident);
                if (field == null) {
                    // $$$ Why searching again
                    field = context.getClassContext().lookupOuterField(ident);
                }
            }
            catch (UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }

            check(context,
                  field != null,
                  KjcMessages.FIELD_UNKNOWN, ident);

            CClass fieldOwner = field.getOwner();

            if (!field.isStatic()) {

                check(context,
                      !local.isStatic() || local.descendsFrom(fieldOwner),
                      KjcMessages.FIELD_STATICERR, ident);

                prefix = new JThisExpression(getTokenReference(), fieldOwner);
            }
            else {
                prefix = new JTypeNameExpression(getTokenReference(),
                                                 fieldOwner.getType());
            }
            prefix = prefix.checkExpression(context);
        }
    }

    /**
     * Checks is access to prefix is okay.
     *
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public void checkAccess(CClass local, CExpressionContext context)
        throws PositionedError
    {
        CClass access = prefix.getType().toClassOrInterface().getCClass();

        try {
            field = access.lookupField(ident);
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }

        check(context,
              field != null,
              KjcMessages.FIELD_UNKNOWN, access.getIdent() + '.' + ident);

        //!!!DEBUG 000213 graf (see ClassOrInterfaceType.getCClass())
        try {
            field.getType().resolveType(context);
        }
        catch (UnpositionedError e) {
            e.printStackTrace();
            throw e.addPosition(getTokenReference());
        }
        //!!!DEBUG 000213 graf

        access = field.getOwner();

        check(context,
              field.isAccessible(local),
              KjcMessages.FIELD_NOACCESS, ident);

        check(context,
              field.getOwner().isAccessible(local),
              KjcMessages.CLASS_NOACCESS, field.getOwner().getIdent());

        if ((context.getMethodContext() instanceof CConstructorContext)
            && (prefix instanceof JThisExpression)
            && !field.isStatic()) {

            if (field.getType().isClassOrInterfaceType()
                && field.getType().toClassOrInterface().getCClass()
                == context.getClassContext().getCClass()) {

                check(context,
                      ((CConstructorContext)context.getMethodContext())
                      .isSuperConstructorCalled(),
                      KjcMessages.CONSTRUCTOR_EXPLICIT_CALL, field.getIdent());
            }
        }

        if (field.mustUseAccessorMethod(local)) {
            /* Make sure that getter method is generated in checking phase,
             * so it will end up in the class file.
             */
            ((CSourceField)field).getGetterAccessor();
            /*
            System.err.println("Must use accessor to access field "
                               + field.getOwner().getIdent() + "."
                               + field.getIdent() + " from class "
                               + local.getIdent());
            */
            this.mustUseAccessorMethod = true;
        }

    }

    public boolean equals(Object o)
    {
        return o instanceof JClassFieldExpression &&
            field.equals(((JClassFieldExpression)o).field) &&
            prefix.equals(((JClassFieldExpression)o).prefix);
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitFieldExpression(this, prefix, getIdent());
    }

    /**
     * Generates a sequence of bytecodes
     * @param code the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        boolean discardValue = code.discardValue();

        if (field.isStatic()) {
            if (prefix != null) {
                code.setDiscardValue(true);
                prefix.genCode(code);
                code.setDiscardValue(discardValue);
            }
        }
        else {
            prefix.genCode(code);
        }

        if (!discardValue) {
            if (mustUseAccessorMethod) {
                try {
                    CSourceMethod getterAccessor =
                        ((CSourceField)field).getGetterAccessor();
                    getterAccessor.genCallCode(code, false);
                }
                catch (PositionedError e) {
                    // Shouldn't happen; no invalid inner accessors should
                    // be generated.
                    throw new RuntimeException(e.toString());
                }
            }
            else {
                field.genLoad(code);
            }
        }

    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genStartStoreCode(CodeSequence code)
    {
        if (field.isStatic()) {
            if (prefix != null) {
                boolean  discardValue = code.discardValue();

                code.setDiscardValue(true);
                prefix.genCode(code);
                code.setDiscardValue(discardValue);
            }
        }
        else {
            prefix.genCode(code);
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param code the code list
     */
    public void genEndStoreCode(CodeSequence code, boolean discardValue)
    {
        if (!discardValue) {
            int opcode;

            if (field.isStatic()) {
                if (Types.getSize(getType()) == 2) {
                    opcode = opc_dup2;
                }
                else {
                    opcode = opc_dup;
                }
            }
            else {
                if (Types.getSize(getType()) == 2) {
                    opcode = opc_dup2_x1;
                }
                else {
                    opcode = opc_dup_x1;
                }
            }

            code.plantInstruction(new NoArgInstruction(opcode));
        }

        if (mustUseAccessorMethod) {
            try {
                CSourceMethod setterAccessor =
                    ((CSourceField)field).getSetterAccessor();
                setterAccessor.genCallCode(code, false);
            }
            catch (PositionedError e) {
                // Shouldn't happen; no invalid inner accessors should
                // be generated.
                throw new RuntimeException(e.toString());
            }
        }
        else {
            field.genStore(code);
        }
    }

}
