/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.List;
import java.io.File;

import at.dms.compiler.CWarning;
import at.dms.compiler.Compiler;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 * This class represents a virtual file and is the main entry point in java
 * grammar.
 */
public class JCompilationUnit
    extends JPhylum
{
    /**
     * Package name of this compilation unit, defined with
     * "package foo.bar;" at top of compilation unit.
     */
    private JPackageName packageName;

    /**
     * Imports that refer to classes (import foo.Bar;)
     */
    private JClassImport[] importedClasses;

    /**
     * Imports that refer to packages (import foo.bar.*;)
     */
    private JPackageImport[] importedPackages;

    /**
     * Type declarations in this compilation unit.
     */
    private JTypeDeclaration[] typeDeclarations;

    /**
     * Probably: all classes in compilation unit that are loaded. Not sure
     * about it.
     */
    private Map allLoadedClasses = new HashMap();

    /**
     * 
     */
    private CCompilationUnit export;

    /**
     * Constructs a CompilationUnit with the specified top level context.
     *
     * @param where
     * The position of this token.
     */
    public JCompilationUnit(TokenReference where,
                            JPackageName packageName,
                            JPackageImport[] importedPackages,
                            JClassImport[] importedClasses,
                            JTypeDeclaration[] typeDeclarations)
    {
        super(where);
        this.packageName = packageName;
        this.importedPackages = importedPackages;
        this.importedClasses = importedClasses;
        this.typeDeclarations = typeDeclarations;
    }

    /**
     * @return the package name of this compilation unit.
     */
    public String getPackageName()
    {
        assert(packageName != null);
        return packageName.getName();
    }

    /**
     * @return the name of the file associated with this compilation unit.
     */
    public String getFileName()
    {
        return getTokenReference().getFile();
    }

    /**
     * Second pass (quick), check interface looks good.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInterface(Compiler compiler)
        throws PositionedError
    {
        export = new CCompilationUnit(packageName.getName(),
                                      importedClasses,
                                      importedPackages);

        CCompilationUnitContext context =
            new CCompilationUnitContext(compiler, export);

        // 7.5.1 Single-Type-Import Declaration
        for (int i = 0; i < importedClasses.length; i++) {
            JClassImport ic = importedClasses[i];

            ClassOrInterfaceType impClass =
                ClassOrInterfaceType.lookup(ic.getClassName());

            try {
                impClass.resolveType(context);
            }
            catch (at.dms.compiler.UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }

            Object clazz = allLoadedClasses.put(ic.getClassIdent(), impClass);
            if (clazz != null) {
                if (impClass != clazz) {
                    compiler.reportTrouble
                        (new CWarning(getTokenReference(),
                                      KjcMessages.CUNIT_RENAME,
                                      impClass.getIdent()));
                }
            }
        }

        // 7.5.2 Type-Import-on-Demand Declaration
        // check uniqueness of classes
        for (int i = 0; i < typeDeclarations.length ; i++) {
            CClass object = typeDeclarations[i].getCClass();

            Object clazz = allLoadedClasses.get(object.getIdent());

            if (clazz == null) {
                allLoadedClasses.put(object.getIdent(), object.getType());
            }
            else {
                if (clazz != object) {
                    compiler.reportTrouble
                        (new PositionedError(getTokenReference(),
                                             KjcMessages.CUNIT_RENAME,
                                             object.getQualifiedName()));
                }
            }
        }

        for (int i = 0; i < typeDeclarations.length ; i++) {
            typeDeclarations[i].checkInterface(context);
        }
    }

    /**
     * Second pass (quick), check interface looks good.
     *
     * @return
     * true iff sub tree is correct enought to check code.
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInitializers(Compiler compiler, List classes)
        throws PositionedError
    {
        CCompilationUnitContext context =
            new CCompilationUnitContext(compiler, export, classes);

        for (int i = 0; i < typeDeclarations.length ; i++) {
            typeDeclarations[i].checkInitializers(context);
        }
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public void checkBody(Compiler compiler, List classes)
        throws PositionedError
    {
        CCompilationUnitContext context =
            new CCompilationUnitContext(compiler, export, classes);

        if (packageName == JPackageName.UNNAMED) {
            compiler.reportTrouble
                (new CWarning(getTokenReference(),
                              KjcMessages.PACKAGE_IS_MISSING));
        }

        for (int i = 0; i < typeDeclarations.length ; i++) {
            typeDeclarations[i].checkTypeBody(context);
        }

        // Check for unused class imports
        for (int i = 0; i < importedClasses.length; i++) {
            importedClasses[i].checkBody(compiler);
        }
        // Check for unused package imports
        for (int i = 0; i < importedPackages.length; i++) {
            importedPackages[i].checkBody(compiler, packageName);
        }
    }

    /**
     * Generate the code in pure java form.
     *
     * It is useful to debug and tune compilation process.
     */
    public void accept(Main compiler, String destination)
    {
        KjcPrettyPrinter pp;

        if ( destination == null || destination.equals("") ||
             destination.equals(".") ) {
            pp = compiler.getPrettyPrinter(getTokenReference().getName() +
                                           ".gen");
        }
        else {
            pp = compiler.getPrettyPrinter(destination + File.separatorChar
                                           + getTokenReference().getName());
        }
        accept(pp);
        pp.close();
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitCompilationUnit(this, packageName, importedPackages,
                               importedClasses, typeDeclarations);
    }
}
