/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Syntax tree node for the || (shortcut-or) operator.
 */
public class JConditionalOrExpression
    extends JBinaryExpression
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param left
     * The left operand.
     * @param right
     * The right operand.
     */
    public JConditionalOrExpression(TokenReference where,
                                    JExpression left,
                                    JExpression right)
    {
        super(where, left, right);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        Type leftType = left.getType();
        Type rightType = right.getType();

        check(context,
              leftType == StandardType.BOOLEAN &&  rightType == StandardType.BOOLEAN,
              KjcMessages.OR_BADTYPE);

        type = StandardType.BOOLEAN;

        if (left.isConstant() && right.isConstant()) {
            JBooleanLiteral leftLit = left.getBooleanLiteral();
            JBooleanLiteral rightLit = right.getBooleanLiteral();
            
            left = new JBooleanLiteral(getTokenReference(),
                                       leftLit.booleanValue()
                                       || rightLit.booleanValue());
            return left;
        }
        else if (left.isConstant()) {
            JBooleanLiteral leftLit = left.getBooleanLiteral();
            if (!leftLit.booleanValue()) {
                left = right;
            }
            /* Optimized expression, but not constant according to the JLS,
             * even though literal.
             */
            left.setConstant(false);
            return left;
        }
        else if (right.isConstant()) {
            /* When the right side is constant, the left side can have side
             * effects, even if its result will never be used because the
             * value of the right side determines the result of the expression.
             * So never throw away the left side, like you can sometimes
             * throw away the right side when the left side is constant.
             */
            right = right.getBooleanLiteral();
            return this;
        }
        else {
            return this;
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitBinaryExpression(this, "||", left, right);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditional expression.
     */
    protected void genBranch(JExpression left,
                             JExpression right,
                             boolean condition,
                             CodeSequence code,
                             CodeLabel label)
    {
        if (condition) {
            left.genBranch(true, code, label);
            right.genBranch(true, code, label);
        }
        else {
            CodeLabel  skip = new CodeLabel();
            left.genBranch(true, code, skip);
            right.genBranch(false, code, label);
            code.plantLabel(skip);
        }
    }
}
