/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 * A specialized JBlock for use in constructors.
 */
public class JConstructorBlock
    extends JBlock
{
    /**
     * If present, an explicit invocation of another constructor in this
     * class or in the super class.
     */
    private JExplicitConstructorInvocation explicitInvoke;

    /**
     * If needed, a statement that calls the instance initializer block.
     */
    private JStatement blockCall;

    /**
     * The source class that contains this constructor.
     */
    private CSourceClass sourceClass;

    /**
     * The number of parameters that this constructor has.
     */
    private int paramsLength;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param body
     * The statements in this constructor block.
     */
    public JConstructorBlock(TokenReference where,
                             JStatement[] body)
    {
        super(where, body, null);
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        CBodyContext inside = context; // !!!context.createBodyContext();
        CClassContext classContext = context.getClassContext();
        CSourceMethod sourceMethod =
            (CSourceMethod)context.getMethodContext().getCMethod();
        TokenReference where = getTokenReference();

        sourceClass = (CSourceClass)context.getClassContext().getCClass();

        boolean noSuperInvoke = false;

        if ( explicitInvoke == null ) {
            if ( body.length == 0 ) {
                noSuperInvoke = true;
            }
            else {
                 if ( !(body[0] instanceof JExpressionStatement) ) {
                     noSuperInvoke = true;
                 }
                 else {
                     JExpressionStatement first =
                         (JExpressionStatement)body[0];

                     noSuperInvoke =
                         !( first.getExpression()
                            instanceof JExplicitConstructorInvocation );
                 }
            }
        } 

        if (noSuperInvoke) {
            if ( sourceClass.getType().equals(StandardType.OBJECT) ) {
                explicitInvoke = null;
            }
            else {
                explicitInvoke =
                    new JExplicitConstructorInvocation(where, null, "super",
                                                       JExpression.EMPTY);
            }
        }
        else if ( body.length > 0 ) {
            if (body[0] instanceof JExpressionStatement) {
                JExpressionStatement first = (JExpressionStatement)body[0];
                if ( first.getExpression()
                     instanceof JExplicitConstructorInvocation ) {

                    explicitInvoke =
                        (JExplicitConstructorInvocation)first.getExpression();

                    body[0] = new JEmptyStatement(where, null);
                }
            }
        }
        if (sourceClass.isInnerClass()) {
            paramsLength = sourceMethod.getParameters().length;
        }

        if (classContext.hasInitializer() && !explicitInvoke.isThisInvoke()) {
            /* If the class has an instance initializer, and the explicit
             * constuctor invocation invokes the super class constructor,
             * then this constructor should call the initializer.
             */
            // "Block$();" ===>
            JNameExpression ident = new JNameExpression(where, JAV_INIT);
            JExpression expr = new JMethodCallExpression(where, ident,
                                                         JExpression.EMPTY);
            blockCall = new JExpressionStatement(where, expr, null);
            // "Block$();" <===
        }

        if (explicitInvoke != null) {
            explicitInvoke.checkExpression(new CExpressionContext(inside));
            if (explicitInvoke.isThisInvoke()) {
                /* Each constructor should initialize all final fields, so
                 * after a call to another constructor, no final fields can
                 * be re-assigned.
                 */
                classContext.markAllFieldsAsInitialized();
            }

            CMethod invoked = explicitInvoke.getInvokedConstructor();
            sourceMethod.setExplicitlyInvokedContructor(invoked);
        }

        if (blockCall != null) {
            blockCall.checkStatement(inside);
        }

        super.checkStatement(inside);

        inside.close(where);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        JStatement[] newBody = body;
        /* Explicit super invocations are removed from the statement array
         * internally, but for printing it needs to be inserted in the
         * array given to the pretty printing routine.
         * That's what this code does.
         */
        if (explicitInvoke != null) {
            if (body[0] instanceof JEmptyStatement) {
                newBody = new JStatement[body.length];
                System.arraycopy(body, 0, newBody, 0, body.length);
            }
            else {
                newBody = new JStatement[body.length + 1];
                System.arraycopy(body, 0, newBody, 1, body.length);
            }
            newBody[0] =
                new JExpressionStatement(explicitInvoke.getTokenReference(),
                                         explicitInvoke, null);
        }
        p.visitBlockStatement(this, newBody, getComments());
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (explicitInvoke != null) {
            explicitInvoke.genCode(code);
        }
        else if (body.length == 0) {
            new JReturnStatement(getTokenReference(), null, null)
                .genCode(code);
        }

        if (sourceClass.isInnerClass()) {
            sourceClass.genInit(code, paramsLength);
        }

        if (blockCall != null) {
            blockCall.genCode(code);
        }

        for (int i = 0; i < body.length; i++) {
            body[i].genCode(code);
        }
    }
}
