/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 * This class represents a java class in the syntax tree.
 */
public class JConstructorDeclaration
    extends JMethodDeclaration
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param modifiers
     * List of modifiers
     * @param ident
     * The name of this constructor.
     * @param parameters
     * The parameters of this constructor.
     * @param exceptions
     * The exceptions declared by this constructor.
     * @param body
     * The syntax tree of the code in the constructor.
     * @param javadoc
     * Javadoc comments.
     * @param comments
     * Implementation comments.
     */
    public JConstructorDeclaration(TokenReference where,
                                   int modifiers,
                                   String ident,
                                   JFormalParameter[] parameters,
                                   ClassOrInterfaceType[] exceptions,
                                   JConstructorBlock body,
                                   JavadocComment javadoc,
                                   JavaStyleComment[] comments)
    {
        super(where, modifiers, StandardType.VOID, ident,
              parameters, exceptions, body, javadoc, comments);
    }

    /**
     * Second pass (quick), check interface looks good.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public CSourceMethod checkInterface(CClassContext context)
        throws PositionedError
    {
        check(context,
              !hasOtherFlags(getModifiers(),
                             ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE),
              KjcMessages.CONSTRUCTOR_FLAGS,
              getIdent());

        check(context,
              getIdent() == context.getCClass().getIdent(),
              KjcMessages.CONSTRUCTOR_BAD_NAME,
              getIdent(),
              context.getCClass().getIdent());

        return super.checkInterface(context);
    }

    /**
     * Check statement and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public void checkStatement(CClassContext context)
        throws PositionedError
    {
        check(context, getBody() != null,
              KjcMessages.CONSTRUCTOR_NOBODY, getIdent());
  
        JFormalParameter[] parameters = getParameters();
        CMethodContext self = createContext(context);
        CBlockContext block = self.createBlockContext(parameters.length);
        CBodyContext inside = block.createBodyContext();
    
        if (!getMethod().isStatic()) {
            // add this local var
            block.addThisVariable();
        }
    
        CClass clazz = getMethod().getOwner();
        if (clazz.isInnerClass() && !clazz.isStatic()) {
            block.addOuterThisVariable();
        }

        for (int i = 0; i < parameters.length; i++) {
            parameters[i].checkBody(inside);
        }
    
        getBody().checkStatement(inside);
    
        inside.close(getTokenReference());
        block.close(getTokenReference());
        self.close(getTokenReference());
    }

    /**
     * Creates a CMethodContext for this constructor.
     *
     * @param parent
     * The parent context or null.
     * @return
     * Returns a CMethodContext that represents this context.
     */
    public CMethodContext createContext(CClassContext parent)
    {
        return parent.createConstructorContext(getMethod());
    }

    /**
     * Accepts the specified visitor
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        genComments(p);
        p.visitConstructorDeclaration(this,
                                      getModifiers(),
                                      getIdent(),
                                      getParameters(),
                                      getExceptions(),
                                      getBody());
    }
}
