/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a explicit call to a super or self constructor.
 */
public class JExplicitConstructorInvocation
    extends JExpression
{
    /**
     * The class in which the explicitly invoked constructor resides.
     */
    private CClass invokedClazz;

    /**
     * Which constructor is explictly invoked.
     */
    private CMethod invokedConstructor;

    /**
     * The syntax-tree prefix.
     */
    private JExpression prefix;

    /**
     * The name of the constructor that is explictly invoked.
     */
    private String ident;

    /**
     * The parameters to the invocation.
     */
    private JExpression[] params;

    /**
     * True if the constructor is not directly accessible but only by creating
     * an accessor method in the class that calls the real constructor.
     */
    private boolean mustUseAccessorMethod = false;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param prefix
     * An expression that is a field of a class representing a method.
     * @param ident
     * The constructor name.
     * @param args
     * The arguments of the call.
     */
    public JExplicitConstructorInvocation(TokenReference where,
                                          JExpression prefix,
                                          String ident,
                                          JExpression[] args)
    {
        super(where);

        this.prefix = prefix;
        this.ident = ident;
        this.params = args;
    }

    /**
     * Returns true if it's this() else it's super().
     */
    final boolean isThisInvoke()
    {
        return ident.equals(JAV_THIS);
    }

    /**
     * Returns the type of this epxression; this is a special expression
     * however, that can only occur within constructors, and calls other
     * constructors, and constructors themselves have no return type.
     * So this method returns <code>null</code>.
     */
    public Type getType()
    {
        return null;
    }

    /**
     * Returns the constructor that is invoked by this invocation.
     */
    public CMethod getInvokedConstructor()
    {
        return invokedConstructor;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        /* The explicit constructor call construct with a prefix is only
         * valid for "super" invocations.
         */
        check(context,
              prefix == null || !isThisInvoke(),
              KjcMessages.CONSTRUCTOR_INVALID_EXPLICIT_THIS_PREFIX );

        CClassContext classContext = context.getClassContext();
        CClass local = classContext.getCClass();
        CConstructorContext constructorContext =
            (CConstructorContext)context.getMethodContext();

        constructorContext.setSuperConstructorCalled(false);

        // !!! check in constructor !!!

        if (prefix != null) {
            prefix.checkExpression(context);
        }

        Type[] argsType = new Type[params.length];
        for (int i = 0; i < argsType.length; i++) {
            params[i] = params[i].checkExpression(context);

            argsType[i] = params[i].getType();
            assert(argsType[i] != null);
        }

        if (isThisInvoke()) {
            invokedClazz = local;
        }
        else {
            invokedClazz = local.getSuperClass();
        }

        assert(invokedClazz != null);
        try {
            invokedConstructor =
                invokedClazz.lookupMethod(local, JAV_CONSTRUCTOR, argsType);
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }

        check(context,
              invokedConstructor != null,
              KjcMessages.CONSTRUCTOR_NOTFOUND, invokedClazz.getIdent());

        if (invokedConstructor.getOwner() != invokedClazz) {
            /* May be an inner class with outer this. */
            if (invokedClazz.hasOuterThis()) {
                Type[] argsType2 = new Type[argsType.length + 1];
                System.arraycopy(argsType, 0, argsType2, 1, argsType.length);
                /* Outer-this is at index 0! */
                argsType2[0] = invokedClazz.getOwner().getType();

                try {
                    invokedConstructor =
                        invokedClazz.lookupMethod(local, JAV_CONSTRUCTOR,
                                                  argsType2);
                }
                catch (UnpositionedError e) {
                    throw e.addPosition(getTokenReference());
                }
            }
        }

        CClass invokedOwner = invokedConstructor.getOwner();

        if (invokedOwner != invokedClazz) {
            // do not want a super constructor !
            throw new CMethodNotFoundError(getTokenReference(), null,
                                           invokedClazz.getQualifiedName(),
                                           argsType);
        }

        check(context,
              invokedConstructor.isAccessible(local),
              KjcMessages.METHOD_NOACCESS, invokedConstructor.getIdent());

        /* If super class is an inner class, then this class must be an
         * inner class, or the call to super must have had a prefix.
         */
        check(context,
              !invokedClazz.hasOuterThis() ||
              local.hasOuterThis() || prefix != null,
              KjcMessages.INNER_INHERITENCE,
              invokedClazz.getQualifiedName(), local.getQualifiedName());

        /* It is not allowed to give a prefix to a super invocation to
         * a non-inner class.
         */
        check(context,
              invokedClazz.hasOuterThis() || prefix == null,
              KjcMessages.QUALIFIEDNEW_NOT_INNER,
              invokedClazz.getQualifiedName());

        /* Type of prefix must match type of outer class of super constructor.
         */
        if ( prefix != null &&
             !prefix.getType()
             .allowsAssignmentConversion(invokedClazz.getOwner().getType()) ) {

            throw new PositionedError
                (getTokenReference(), 
                 KjcMessages.CONSTRUCTOR_NOMATCH_EXPLICIT_PREFIX_TYPE,
                 prefix.getType(), invokedClazz.getOwner().getType());
        }

        CBodyContext bodyContext = context.getBodyContext();
        CBodyContext clonedContext = bodyContext.cloneContext();

        check(context,
              invokedOwner.isAccessible(local),
              KjcMessages.CLASS_NOACCESS, invokedOwner.getIdent());

        ClassOrInterfaceType[] exceptions = invokedConstructor.getThrowables();
        for (int i = 0; i < exceptions.length; i++) {
            CThrowableInfo info = new CThrowableInfo(exceptions[i],
                                                     this,
                                                     clonedContext,
                                                     true); // subTypePossible
            bodyContext.addThrowable(info);
        }

        /* New-object expressions can always result in a RuntimeException,
         * methods and constructors don't have to declare RuntimeExceptions.
         */
        CThrowableInfo info = new CThrowableInfo(StandardType.RUNTIMEEXCEPTION,
                                                 this, clonedContext,
                                                 true); // subTypePossible
        bodyContext.addThrowable(info);

        check(context,
              !constructorContext.getCMethod().isStatic(),
              KjcMessages.BAD_THIS_STATIC);

        argsType = invokedConstructor.getParameters();
        for (int i = 0; i < params.length; i++) {
            params[i] = params[i].convertType(argsType[i], context);
        }

        if (invokedConstructor.mustUseAccessorMethod(local)) {
            /* Make sure that accessor method is generated in checking phase,
             * so it will end up in the class file.
             */
            ((CSourceMethod)invokedConstructor).getAccessor();
            this.mustUseAccessorMethod = true;
        }

        constructorContext.setSuperConstructorCalled(true);
        return this;
    }

    /**
     * Accepts the specified visitor
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitExplicitConstructorInvocation(this, prefix, ident, params);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());
        code.plantLoadThis();

        if (invokedConstructor.getOwner().isInnerClass()) {
            invokedClazz.genSyntheticParamsFromExplicitSuper(params, code,
                                                             prefix);
        }

        for (int i = 0; i < params.length; i++) {
            params[i].genCode(code);
        }

        if (mustUseAccessorMethod) {
            code.plantInstruction(new NoArgInstruction(opc_aconst_null));
            try {
                CSourceMethod accessor =
                    ((CSourceMethod)invokedConstructor).getAccessor();
                code.plantInstruction
                    (new MethodRefInstruction(opc_invokespecial,
                                              accessor.getPrefixName(),
                                              accessor.getIdent(),
                                              accessor.getSignature()));
            }
            catch (PositionedError e) {
                // Shouldn't happen; no invalid constructor accessors should
                // be generated.
                throw new RuntimeException(e.toString());
            }
        }
        else {
            code.plantInstruction
                (new MethodRefInstruction(opc_invokespecial,
                                          invokedConstructor.getPrefixName(),
                                          invokedConstructor.getIdent(),
                                          invokedConstructor.getSignature()));
        }
    }
}
