/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.type.*;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.util.MessageDescription;

/**
 * Root class for all expressions in the syntax tree.
 */
public abstract class JExpression
    extends JPhylum
{
    /**
     * Empty expression-list constant.
     */
    public static final JExpression[] EMPTY = new JExpression[0];

    /**
     * Whether this expression is constant. By default, this field is false,
     * except for JLiteral expressions.
     */
    private boolean constant = false;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JExpression(TokenReference where)
    {
        super(where);
    }

    /**
     * Returns the type of this expression. Call after parsing only.
     */
    public abstract Type getType();

    /**
     * Returns whether this expression can result in subtypes of the type
     * returned by the <code>getType()</code> method, or whether it always
     * returns the exact type returned by <code>getType()</code>.
     * The return value of this method has no meaning unless getType() returns
     * a Class-type. (<code>Type.isClassOrInterfaceType()</code>)
     */
    public boolean isSubTypePossible()
    {
        return true;
    }

    /**
     * Returns true if this expression is constant according to the JLS.
     * Some expression may be a literal, such as the NullLiteral or the result
     * of ( true ? 10 : System.currentTimeMillis() ), but they're not constant
     * as defined by the Java Language Specification, 2nd edition.
     *
     * @return
     * true if this expression is constant.
     */
    public boolean isConstant()
    {
        return constant;
    }

    /**
     * Change this JLS-constantness of this literal.
     */
    public void setConstant(boolean constant)
    {
        this.constant = constant;
    }

    /**
     * Returns true iff this expression can be used as a statement (JLS 14.7)
     */
    public boolean isStatementExpression()
    {
        return false;
    }

    /**
     * Returns true if this expression is a literal. Many literals are also
     * constant (as defined by the JLS) but some aren't. The reverse is also
     * true, some expressions may be constant but aren't literal. Currently
     * I know only one such case, that of String constants larger than 65535
     * characters. They can only be represented as a runtime-evaluated
     * concatenation expression, but they are JLS-constant.
     *
     * <P>
     * In this particular case there is the added problem that the
     * field cannot be stored in a classfile as a compile-time constant; a
     * "final static String" constant larger than 65535
     * characters cannot be stored in the classes' constant pool.
     * When compiling against the source of a file containing such a large
     * constant, the compiler can ignore this problem.
     * The problem appears when you try to compile another program against a
     * class file with such a String constant, as the other program will not
     * consider the field a compile-time constant.
     *
     * Some expression may be a literal, such as the NullLiteral or the result
     * of ( true ? 10 : System.currentTimeMillis() ), but they're not constant
     * as defined by the Java Language Specification, 2nd edition.
     *
     * @return true if this expression is literal.
     */
    public boolean isLiteral()
    {
        return false;
    }

    /**
     * Returns the literal value of this expression.
     */
    public JLiteral getLiteral()
    {
        throw new RuntimeException("THIS IS NOT A LITERAL");
    }

    /**
     * Returns a typed literal for this expression.
     *
     * The expression must be a constant.
     */
    public final JOrdinalLiteral getOrdinalLiteral()
    {
        return (JOrdinalLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression.
     *
     * The expression must be a constant.
     */
    public final JRealLiteral getRealLiteral()
    {
        return (JRealLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression.
     *
     * The expression must be a constant.
     */
    public final JBooleanLiteral getBooleanLiteral()
    {
        return (JBooleanLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression.
     *
     * The expression must be a constant.
     */
    public final JNumberLiteral getNumberLiteral()
    {
        return (JNumberLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression.
     *
     * The expression must be a constant.
     */
    public final JStringLiteral getStringLiteral()
    {
        return (JStringLiteral)getLiteral();
    }

    /**
     * Check expression and evaluate and alter context.
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public abstract JExpression checkExpression(CExpressionContext context)
        throws PositionedError;

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @param isLeftSide
     * Whether the expression to be evaluated is at the left side of an
     * assignment operation.
     * @param discardValue
     * Whether the value ofthe expression to be evaluated will be discarded.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public final JExpression checkExpression(CExpressionContext context,
                                             boolean isLeftSide,
                                             boolean discardValue)
        throws PositionedError
    {
        boolean isLeftSideSaved = context.isLeftSide();
        boolean discardValueSaved = context.isDiscardValue();

        try {
            context.setLeftSide(isLeftSide);
            context.setDiscardValue(discardValue);
            return checkExpression(context);
        }
        finally {
            context.setLeftSide(isLeftSideSaved);
            context.setDiscardValue(discardValueSaved);
        }
    }


    /**
     * Adds a compiler error.
     *
     * @param context
     * The context in which the error occurred.
     * @param key
     * The message ident to be displayed.
     * @param params
     * The array of parameters.
     */
    protected void fail(CContext context,
                        MessageDescription key, Object[] params)
        throws PositionedError
    {
        throw new CLineError(getTokenReference(), key, params);
    }

    /**
     * Changes the type of this expression to an other type.
     *
     * @param dest
     * The destination type.
     */
    public JExpression convertType(Type dest, CExpressionContext context)
        throws PositionedError
    {
        if (getType().equals(dest)) {
            return this;
        }
        else if (isConstant() && getType().isNumeric()) {
            System.err.println("Suspicious JExpression.convertType rule "
                               + "invoked.");
            // not null
            return JLiteral.createLiteral(dest, this.getLiteral().getValue());
        }
        else {
            return new JUnaryPromote(this, dest).checkExpression(context);
        }
    }

    /**
     * Can this expression be converted to the specified type by assignment
     * conversion (JLS 5.2) ?
     *
     * @param dest
     * The destination type.
     * @return
     * True iff the casting conversion is valid.
     */
    public boolean allowsAssignmentConversion(Type dest)
    {
        return getType().allowsAssignmentConversion(dest);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public abstract void accept(KjcVisitor p);

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public abstract void genCode(CodeSequence code);

    /**
     * Generates a sequence of bytecodes to branch on a label.
     *
     * This method helps to handle heavy optimizables conditions.
     *
     * @param code
     * The code list.
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label)
    {
        genCode(code);
        code.plantLabelRef(cond ? opc_ifne : opc_ifeq, label);
    }

    /**
     * Returns this expression without parentheses.
     */
    public JExpression unParenthesize()
    {
        return this;
    }
}
