/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.type.*;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.7: Expression Statement
 *
 * Certain kinds of expressions may be used as statements by following them
 * with semicolon.
 * An expression statement is executed by evaluating the expression; if the
 * expression has a value, the value is discarded.
 */
public class JExpressionStatement
    extends JStatement
{
    /**
     * The expression that is used as a statement.
     */
    private JExpression expression;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param expression
     * The expression to evaluate.
     */
    public JExpressionStatement(TokenReference where,
                                JExpression expression,
                                JavaStyleComment[] comments)
    {
        super(where, comments);
        this.expression = expression;
    }

    /**
     * Returns the expression that is used as a statement.
     *
     * @return
     * The expression that is used as a statement.
     */
    public JExpression getExpression()
    {
        return expression;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        CExpressionContext expressionContext =
            context.createExpressionContext();

        expressionContext.setDiscardValue(true);
        expression = expression.checkExpression(expressionContext);

        check(context,
              expression.isStatementExpression(),
              KjcMessages.INVALID_EXPRESSION_STATEMENT);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitExpressionStatement(this, expression);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        boolean discardValue = code.discardValue();

        code.setDiscardValue(true);
        expression.genCode(code);
        code.setDiscardValue(discardValue);
    }
}

