/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.CWarning;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 * This class represents a field declaration in the syntax tree.
 */
public class JFieldDeclaration
    extends JMemberDeclaration
{
    /**
     * The variable definition for the field.
     */
    protected JVariableDefinition variable;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param variable
     * The variable definition.
     * @param javadoc
     * Javadoc.
     * @param comments
     * Implementation comments.
     */
    public JFieldDeclaration(TokenReference where,
                             JVariableDefinition variable,
                             JavadocComment javadoc,
                             JavaStyleComment[] comments)
    {
        super(where, javadoc, comments);

        this.variable = variable;
    }

    /**
     * Returns true if this field declarator has initializer.
     * (should be initialized)
     */
    public boolean hasInitializer()
    {
        return variable.hasInitializer();
    }

    /**
     * Returns the type of this field.
     */
    public Type getType()
    {
        return variable.getType();
    }

    /**
     * Returns true if this field need to be initialized
     * WARNING: this method return true when initial value corresponds to a
     * default value
     * ====> a second check should be made after checkStatement to ensure
     * that an initialization is really needed.
     */
    public boolean needInitialization()
    {
        if (!hasInitializer()) {
            return false;
        }

        return true;
    }

    /**
     * ...
     */
    public JVariableDefinition getVariable()
    {
        return variable;
    }

    /**
     * Second pass (quick), check interface looks good.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public CSourceField checkInterface(CClassContext context)
        throws PositionedError
    {
        int modifiers = variable.getModifiers();

        // Check modifiers
        // 8.3.1 Field Modifiers
        check(context,
              !hasOtherFlags(modifiers,
                             ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE
                             + ACC_FINAL + ACC_STATIC
                             + ACC_TRANSIENT + ACC_VOLATILE),
              KjcMessages.FIELD_FLAGS);

        if (context.getCClass().isInterface()) {
            modifiers |= ACC_PUBLIC | ACC_STATIC | ACC_FINAL;
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC + ACC_FINAL + ACC_STATIC),
                  KjcMessages.FIELD_FLAGS_IN_INTERFACE);
        }

        /* Not more than one access flag in modifier! */
        check(context,
              !hasMoreThanOneFlag(modifiers,
                                  ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE),
              KjcMessages.MULTIPLE_ACCESS_MODIFIERS);

        variable.checkInterface(context);
        setInterface(new CSourceField(context.getCClass(),
                                      modifiers,
                                      variable.getIdent(),
                                      variable.getType(),
                                      isDeprecated()));
        getField().setSynthetic(isSynthetic());
        return (CSourceField)getField();
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analysis.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        variable.checkBody(context);

        if (hasInitializer() && getField().isFinal()) {
            JExpression value = variable.getValue();

            if (value.isConstant()) {
                getField().setValue(value.getLiteral());
            }
        }
    
        if (hasInitializer()) {
            context.addFieldInfo(((CSourceField)getField()).getPosition(),
                                 CVariableInfo.initialize());
        }

        /* This warns for fields with "package" or "protected" access. */
        if (! (getField().isPublic() || getField().isPrivate()) ) {
            context.reportTrouble
                (new CWarning(getTokenReference(),
                              KjcMessages.PACKAGE_PROTECTED_ATTRIBUTE,
                              getField().getIdent()));
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitFieldDeclaration(this,
                                variable.getModifiers(),
                                variable.getType(),
                                variable.getIdent(),
                                variable.getValue());
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        if (variable.getValue() != null) {
            code.setLineNumber(getTokenReference().getLine());

            if (!getField().isStatic()) {
                code.plantLoadThis();
            }
            variable.getValue().genCode(code);
            getField().genStore(code);
        }
    }
}
