/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.compiler.PositionedError;
import at.dms.compiler.CWarning;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.10: If Statement
 *
 * The if statement executes an expression and a statement repeatedly
 * until the value of the expression is false.
 */
public class JIfStatement
    extends JStatement
{
    /**
     * Condition of if statement.
     */
    private JExpression condition;

    /**
     * The then clause of the if statement.
     */
    private JStatement thenClause;

    /**
     * The else clause of the if statement.
     */
    private JStatement elseClause;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param cond
     * The expression to evaluate.
     * @param thenClause
     * The statement to execute if the condition is true.
     * @param elseClause
     * The statement to execute if the condition is false.
     */
    public JIfStatement(TokenReference where,
                        JExpression condition,
                        JStatement thenClause,
                        JStatement elseClause,
                        JavaStyleComment[] comments)
    {
        super(where, comments);

        this.condition = condition;
        this.thenClause = thenClause;
        this.elseClause = elseClause;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        /* Check the condition expression [if (condition)] */
        condition =
            condition.checkExpression(context.createExpressionContext());

        /* Check that the type of the condition is "boolean" */
        check(context,
              condition.getType() == StandardType.BOOLEAN,
              KjcMessages.IF_COND_NOTBOOLEAN,
              condition.getType());

        CBodyContext thenContext = context.cloneContext();
        CBodyContext elseContext = context.cloneContext();

        thenClause.checkBody(thenContext);

        boolean conditionValue = false;

        if (condition.isConstant()) {
            conditionValue = condition.getBooleanLiteral().booleanValue();
        }

        if (elseClause == null) {
            /* Process if-then */
            if (condition.isConstant()) {
                if (conditionValue) {
                    context.adopt(thenContext);
                }
                else {
                    /* Breaks/continues are automatically processed,
                     * even if condition is false. This is right. I.e.
                     * "while(true) { if (false) break; } statement;"
                     * -> 'statement' is reachable.
                     * Assignment is supposed to have 'possibly' occured.
                     * Throwables of the block, even if constant false,
                     * are considered to be possible.
                     * The right solution to achieve these effects is
                     * to merge the thenContext.
                     */

                    /* Urghl. Why can't the JLS2 use formal logic in section
                     * 16 or at least use parentheses! 
                     */
                    context.merge(thenContext);
                }
            }
            else {
                /* The condition isn't constant, so it is possible that
                 * the then block doesn't get executed; any flow
                 * information from it should be merged with the parent
                 * flow, not adopted by the parent flow.
                 */
                context.merge(thenContext);
            }
            /* From the JLS:
             * An if-then statement can complete normally iff it is reachable.
             * The then-statement is reachable iff the if-then statement is
             * reachable. 
             * Since this if-then statement is checked, it must be reachable,
             * so it can complete normally.
             */
            context.setReachable(true);
            
        }
        else {
            /* Process if-then-else */
            elseClause.checkBody(elseContext);
            if (thenContext.isReachable() || elseContext.isReachable()) {
                if (condition.isConstant()) {
                    if (conditionValue) {
                        context.merge(elseContext);
                        context.adopt(thenContext);
                        /* Merge else context, see comment above in
                         * if-then case about how the flow information
                         * from a discarded context has already been
                         * passed on so this works right.
                         */
                    }
                    else {
                        context.adopt(elseContext);
                        context.setReachable(true);
                        /* Merge then context, see comment above in
                         * if-then case about how the flow information
                         * from a discarded context has already been
                         * passed on so this works right.
                         */
                        context.merge(thenContext);
                    }
                }
                else {
                    /* The condition isn't constant, so either the then
                     * or the else gets executed.
                     * So, first merge then and else, and then merge the
                     * result with the parent context.
                     * But if then or else doesn't complete normally, it
                     * has no influence on the assignment-flow in the
                     * block that contains the if-statement.
                     */
                    if (thenContext.isReachable()) {
                        if (elseContext.isReachable()) {
                            thenContext.merge(elseContext);
                        }
                        context.adopt(thenContext);
                    }
                    else {
                        context.adopt(elseContext);
                    }
                }
               
                /* Set reachable. */ 
                context.setReachable(true);
            }
            else {
                /* Neither then nor else block can complete normally.
                 * Set unreachable.
                 */ 
                context.setReachable(false);
            }
        }

        /*
        if ( condition.isConstant() ) {
            if (condition.getBooleanLiteral().booleanValue()) {
                /* Condition is constant and true, so only assignment
                 * info from then context is valid.
                 *
                context.adopt(thenContext);
                if (elseClause != null) {
                    context.merge(elseContext);
                }
            }
            else {
                /* Condition is constant and false, so only assignment
                 * info from else context is valid.
                 *

                context.adopt(elseContext);
                context.merge(thenContext);
            }
        }
        else {
            if ( elseContext.isReachable() ) {
                if ( thenContext.isReachable() ) {
                    context.adopt(thenContext);
                    context.merge(elseContext);
                }
                else {
                    context.adopt(elseContext);
                }
            }
            else if ( thenContext.isReachable() ) {
                context.adopt(thenContext);
            }
            else {
                context.setReachable(false);
            }
        }
        */

        if (condition instanceof JAssignmentExpression) {
            context.reportTrouble
                (new CWarning(getTokenReference(),
                              KjcMessages.ASSIGNMENT_IN_CONDITION));
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitIfStatement(this, condition, thenClause, elseClause);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (condition.isConstant()) {
            /* When the condition is constant, at most one of the then or
             * else clauses needs to have its code generated!
             */
            if (condition.getBooleanLiteral().booleanValue()) {
                thenClause.genCode(code);
            }
            else if (elseClause != null) {
                elseClause.genCode(code);
            }
        }
        else {
            CodeLabel elseLabel = new CodeLabel();
            CodeLabel nextLabel = new CodeLabel();

            //  COND IFEQ else
            condition.genBranch(false, code, elseLabel);
            //  THEN CODE
            thenClause.genCode(code);
            //  GOTO next
            code.plantLabelRef("if:GOTO next", opc_goto, nextLabel);
            // else:
            code.plantLabel(elseLabel);
            //  ELSE CODE
            if (elseClause != null) {
                elseClause.genCode(code);
            }
            // next ...
            code.plantLabel(nextLabel);
        }
    }
}
