/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents a Java interface in the syntax tree.
 */
public class JInterfaceDeclaration
    extends JTypeDeclaration
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param modifiers
     * The modifiers of this interface.
     * @param shortName
     * The short name of this interface.
     * @param interfaces
     * The interfaces extended by this interface.
     * @param fields
     * The fields in this interface.
     * @param methods
     * The methods in this interface.
     * @param javadoc
     * Javadoc comment.
     * @param comments
     * implementation comments.
     */
    public JInterfaceDeclaration(TokenReference where,
                                 int modifiers,
                                 String shortName,
                                 ClassOrInterfaceType[] interfaces,
                                 JFieldDeclaration[] fields,
                                 JMethodDeclaration[] methods,
                                 JTypeDeclaration[] inners,
                                 JPhylum[] initializers,
                                 JavadocComment javadoc,
                                 JavaStyleComment[] comments)
    {
        super(where, modifiers | ACC_INTERFACE | ACC_ABSTRACT,
              shortName, interfaces, fields, methods, inners, initializers,
              javadoc, comments);
    }

    /**
     * Second pass (quick), check interface looks good.
     *
     * @param context
     * The context to perform the check in.
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInterface(CContext context)
        throws PositionedError
    {
        int modifiers = getModifiers();

        if (getOwner() != null) {
            /* Inner interfaces. */
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC | ACC_PRIVATE | ACC_PROTECTED
                                 | ACC_ABSTRACT | ACC_STATIC
                                 | ACC_STRICT | ACC_INTERFACE),
                  KjcMessages.INVALID_INTERFACE_MODIFIERS);
        }
        else {
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC | ACC_ABSTRACT | ACC_STRICT
                                 | ACC_INTERFACE),
                  KjcMessages.INVALID_INTERFACE_MODIFIERS);
        }

        check(context,
              !hasMoreThanOneFlag(modifiers,
                                  ACC_PUBLIC | ACC_PROTECTED | ACC_PRIVATE),
              KjcMessages.MULTIPLE_ACCESS_MODIFIERS);

        setStaticInitializer(constructStaticInitializers());

        super.checkInterface(context, StandardType.OBJECT);
    }

    /**
     * Check that initializers are correct.
     *
     * @param context
     * The context to perform the check in.
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInitializers(CContext context)
        throws PositionedError
    {
        CInterfaceContext self = (CInterfaceContext)createContext(context);
        TokenReference where = getTokenReference();

        JInitializerDeclaration staticInit = getStaticInitializer();
        if (staticInit != null) {
            staticInit.checkInitializer(self);
        }

        // Check inners
        Iterator iterator = getInnerList().iterator();
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            inner.checkInitializers(self);
        }

        CClass clazz = getCClass();
        ClassOrInterfaceType[] interfaces = getInterfaces();
        for (int i = 0; i < interfaces.length; i++) {
            CClass superInterface = interfaces[i].getCClass();

            /* An interface cannot extend itself. */
            if (clazz.equals(superInterface)) {
                reportTrouble(context,
                              KjcMessages.INTERFACE_EXTENDS_SELF, getIdent());
            }
            /* An interface cannot extend a sub-interface. */
            else if ( clazz.descendsFrom(superInterface) &&
                      superInterface.descendsFrom(clazz) ) {
                reportTrouble(context,
                              KjcMessages.INTERFACE_CIRCULARITY, getIdent());
            }
            /* An interface cannot extend one of its member interfaces. */
            else if ( superInterface.descendsFromInside(clazz)) {
                reportTrouble(context,
                              KjcMessages.INTERFACE_EXTENDS_MEMBER,
                              getIdent());
            }
        }
        super.checkInitializers(context);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The context to perform the check in.
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkTypeBody(CContext context)
        throws PositionedError
    {
        CInterfaceContext self = (CInterfaceContext)createContext(context);

        Iterator iterator = getInnerList().iterator();
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            try {
                inner.checkTypeBody(self);
            }
            catch (CBlockError e) {
                context.reportTrouble(e);
            }
        }

        iterator = getMethodList().iterator();
        while (iterator.hasNext()) {
            JMethodDeclaration method = (JMethodDeclaration)iterator.next();
            try {
                method.checkStatement(self);
            }
            catch (PositionedError ce) {
                context.reportTrouble(ce);
            }
        }
    
        super.checkTypeBody(context);
    }

    /**
     * Creates a CInterfaceContext that takes this inteface as its context.
     *
     * @param parent
     * The parent context or null.
     * @return
     * A CInterfaceContext (which descends from CClassContext) that represents
     * this context.
     */
    public CClassContext createContext(CContext parent)
    {
        return new CInterfaceContext(parent, getSourceClass());
    }

    /**
     * Check static initializers.
     */
    private JInitializerDeclaration constructStaticInitializers()
    {
        // collect all static initializers and build a method
        List statementList = new ArrayList(5);
        boolean needGen = false;

        Iterator iterator = getBodyList().iterator();

        while (iterator.hasNext()) {
            JPhylum element = (JPhylum)iterator.next();

            if (element instanceof JClassBlock) {
                statementList.add(element);
                needGen = true;
            }
            else if (element instanceof JFieldDeclaration) {
                JFieldDeclaration declaration = (JFieldDeclaration)element;
                needGen |= declaration.needInitialization();
                statementList.add
                    (new JClassFieldDeclarator(getTokenReference(),
                                               declaration));
            }
        }
        if (statementList.isEmpty()) {
            return null;
        }

        JStatement[] statements = (JStatement[])
            statementList.toArray(new JStatement[statementList.size()]);

        return new JInitializerDeclaration(getTokenReference(),
                                           new JClassBlock(getTokenReference(),
                                                           false, statements),
                                           true, !needGen);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitInterfaceDeclaration(this,
                                    getCClass().getModifiers(),
                                    getSourceClass().getIdent(),
                                    getInterfaces(),
                                    getBody(),
                                    getMethods());
    }
}
