/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * Labeled Statement. (JLS 14.6)
 *
 * Statements may have label prefixes.
 */
public class JLabeledStatement
    extends JStatement
{
    /**
     * The name of the label in the source code.
     */
    private String label;

    /**
     * The labeled statement.
     */
    private JStatement statement;

    /**
     * The label of the first byte code instruction after this statement.
     */
    private CodeLabel endLabel = new CodeLabel();

    /**
     * Is there a break that uses this label as a target?
     */
    private boolean isBreakTarget;
    
    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code
     * @param label  the label of the enclosing labeled statement
     * @param body  the contained statements
     * @param comments comments in the source text
     */
    public JLabeledStatement(TokenReference where,
                             String label,
                             JStatement statement,
                             JavaStyleComment[] comments)
    {
        super(where, comments);

        this.label = label;
        this.statement = statement;
        this.isBreakTarget = false;
    }

    /**
     * Check statement.
     *
     * @param context  the actual context of analyse
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        check(context,
              context.getLabeledStatement(label) == null,
              KjcMessages.LABEL_ALREADY_EXISTS,
              label);

        CLabeledContext self = new CLabeledContext(context, this);
        CBodyContext body = new CBodyContext(self);

        statement.checkBody(body);

        body.close(getTokenReference());
        self.close(getTokenReference());

        if (isBreakTarget) {
            context.setReachable(true);
        }
        else {
            context.reportTrouble
                (new CWarning(getTokenReference(),
                              KjcMessages.UNUSED_LABEL, label));
        }
    }

    /**
     * Returns the label of this statement.
     */
    public String getLabel()
    {
        return label;
    }

    /**
     * Returns true iff this statement is a possible target for a break
     * statement.
     */
    public boolean acceptsBreak()
    {
        return true;
    }

    /**
     * Returns true iff this statement is a possible target for a continue
     * statement.
     */
    public boolean acceptsContinue()
    {
        if ( !(statement instanceof JLabeledStatement) ) {
            return statement.acceptsContinue();
        }
        else {
            /* A JLabeledStatement itself cannot accept a continue, only
             * its contents can. If its contents is a JLabeledStatement,
             * the outer labeled statement cannot accept continues.
             */
            return false;
        }
    }

    /**
     * Returns the statement that is labeled by this label.
     */
    public JStatement getStatement()
    {
        return statement;
    }

    /**
     * Inform statement that there is at least one break.
     */
    public void addBreak(CBodyContext context)
    {
        isBreakTarget = true;
        if (statement.acceptsBreak()) {
            statement.addBreak(context);
        }
    }

    /**
     * Inform statement that there is at least one continue.
     */
    public void addContinue(CBodyContext context)
    {
        statement.addContinue(context);
    }

    /**
     * Return the end of this block. (for break statement)
     */
    public CodeLabel getBreakLabel()
    {
        if (statement.acceptsBreak()) {
            return statement.getBreakLabel();
        }
        else {
            return endLabel;
        }
    }

    /**
     * Return the beginning of this block. (for continue statement)
     */
    public CodeLabel getContinueLabel()
    {
        if ( !(statement instanceof JLabeledStatement) ) {
            return statement.getContinueLabel();
        }
        else {
            /* A JLabeledStatement itself cannot accept a continue, only
             * its contents can. If its contents is a JLabeledStatement,
             * the outer labeled statement cannot accept continues.
             */
            throw new RuntimeException("NO CONTINUE LABEL");
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitLabeledStatement(this, label, statement);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        statement.genCode(code);
        code.plantLabel(endLabel);

        endLabel = null;
    }

}
