/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.codegen.jvm.Types;
import at.dms.compiler.TokenReference;
import at.dms.compiler.CWarning;
import at.dms.type.*;

/**
 * This class represents a local variable.
 */
public abstract class JLocalVariable
    extends JPhylum
{
    /**
     * Description: local variable.
     */
    public static final int DES_LOCAL_VAR  = 1 << 0;

    /**
     * Description: method parameter.
     */
    public static final int DES_PARAMETER  = 1 << 1;

    /**
     * Description: catch parameter.
     */
    public static final int DES_CATCH_PARAMETER = 1 << 2;

    /**
     * Description: synthetic/generated local variable.
     */
    public static final int DES_GENERATED  = 1 << 3;

    /**
     * Description bitfield.
     */
    private int desc;

    /**
     * Whether var was used.
     */
    private boolean used;

    /**
     * Whether var was assigned.
     */
    private boolean assigned;

    /**
     * Whether var was a loop variable.
     */
    private boolean loopVariable;

    /**
     * Modifiers for local var (final)
     */
    protected int modifiers;

    /**
     * Name of local var.
     */
    protected String name;

    /**
     * Type of local var.
     */
    protected Type type;

    /**
     * Initializer of local var.
     */
    protected JExpression initializer;

    /**
     * Index in block local vars array.
     */
    private int index;

    /**
     * Position in method local vars.
     */
    private int position = -1;

    /**
     * If used inside a constructor of an inner class, this field will
     * be set by JLocalVariableExpression during checking.
     */
    private CClass constructorClass;

    /**
     * Whether shifts in position during code generation have been applied.
     * They must only be applied once, so this flag is used to keep track
     * of that.
     */   
    private boolean shifted = false;

    /**
     * Constructs a local variable reference.
     *
     * @param modifiers the modifiers on this variable
     * @param name  the name of this variable
     * @param type  the type of this variable
     * @param value  the initial value
     * @param where  the location of the declaration of this variable
     */
    public JLocalVariable(TokenReference where,
                          int modifiers,
                          int desc,
                          Type type,
                          String name,
                          JExpression initializer)
    {
        super(where);

        this.modifiers = modifiers;
        this.desc = desc;
        this.type = type;
        this.name = name;
        this.initializer = initializer;
    }

    /**
     * ...
     */
    public void setConstructorClass(CClass clazz)
    {
        this.constructorClass = clazz;
    }

    /**
     * Returns if this variable is final
     */
    public boolean isFinal()
    {
        return (getModifiers() & ACC_FINAL) != 0;
    }

    /**
     * Returns if this variable is static
     */
    public boolean isStatic()
    {
        return (getModifiers() & ACC_STATIC) != 0;
    }

    /**
     * Returns the variable definied by this formal parameter
     */
    public boolean isGenerated()
    {
        return (getDescription() & DES_GENERATED) != 0;
    }

    /**
     * @return the modifiers of this variable
     */
    public int getModifiers()
    {
        return modifiers;
    }

    /**
     * @return the name of this variable
     */
    public String getIdent()
    {
        return name;
    }

    /**
     * @return the type of this variable
     */
    public Type getType()
    {
        return type;
    }

    /**
     * @return the value of this local variable at initialization
     * $$$ why not dynamically...
     */
    public JExpression getValue()
    {
        return initializer;
    }

    /**
     * @return true if this variable is constant
     */
    public boolean isConstant()
    {
        return ( hasFlag(getModifiers(), Constants.ACC_FINAL) &&
                 initializer != null && initializer.isConstant() );
        // int (final int i) {}
    }

    /**
     * @return the local index in context variable table
     */
    public int getIndex()
    {
        return index;
    }

    /**
     * @return the local index in context variable table
     */
    public void setIndex(int index)
    {
        this.index = index;
    }

    public int getDescription()
    {
        return desc;
    }

    /**
     * Returns the index in the local variable table, or if this is a class
     * field, in the field info set.
     */
    public int getPosition()
    {
        return position;
    }

    
    public void setPosition(int position)
    {
        this.position = position;
    }

    public void setUsed()
    {
        used = true;
    }

    public boolean isUsed()
    {
        return used;
    }

    public void setAssigned(TokenReference ref, CBodyContext context)
    {
        assigned = true;
        if (isLoopVariable()) {
            context.reportTrouble
                (new CWarning(ref,
                              KjcMessages.ASSIGNS_LOOP_VARIABLE, getIdent()));
        }
    }

    public boolean isAssigned()
    {
        return assigned;
    }

    public void setIsLoopVariable()
    {
        loopVariable = true;
    }

    public void unsetIsLoopVariable()
    {
        loopVariable = false;
    }

    public boolean isLoopVariable()
    {
        return loopVariable;
    }

    public boolean equals(Object o)
    {
        return o instanceof JLocalVariable &&
            index == ((JLocalVariable)o).index;
    }

    /**
     * Generates a sequence of bytecode to load
     * the value of the variable on the stack.
     *
     * @param code
     * The code sequence.
     */
    public void genLoad(CodeSequence code)
    {
        if (constructorClass != null && !isGenerated() && !shifted) {
            position += constructorClass.getOuterLocalsSize();
            shifted = true; // Make sure this doesn't happen twice.
        }
        code.plantLocalVar(Types.getLoadOpcode(type), this);
    }
  
    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code sequence.
     */
    public void genStore(CodeSequence code)
    {
        if (constructorClass != null && !isGenerated() && !shifted) {
            position += constructorClass.getOuterLocalsSize();
            shifted = true; // Make sure this doesn't happen twice.
        }
        code.plantLocalVar(Types.getStoreOpcode(type), this);
    }
}
