/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.codegen.jvm.Types;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.MessageDescription;
import at.dms.type.*;

/**
 * This class represents methods call
 * <pre> methodname(explist) </pre>
 */
public class JMethodCallExpression
    extends JExpression
{
    /**
     * Name of CloneNotSupportedException class.
     */
    private final static String CLONE_NOT_SUPPORTED =
        "java/lang/CloneNotSupportedException";

    /**
     * The prefix of the method call (the classname for a static method,
     * or an expression resulting in a reference to an instance of the class
     * for other method calls - or null if the call is to a method in the
     * same class, outer class or super class)
     */
    protected JExpression prefix;

    /**
     * The name of the method that is called by this expression.
     */
    protected String ident;

    /**
     * The arguments to the method call in this expression.
     */
    protected JExpression[] args;

    /**
     * True if the method is not directly accessible but only by creating
     * an accessor method in the class that calls the real method.
     */
    private boolean mustUseAccessorMethod = false;

    /**
     * When it is possible to determine, this field is assigned with the
     * method export that is the target of this expression.
     */
    protected CMethod method;

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where  the line of this node in the source code
     * @param prefix
     *        an expression that is a field of a class representing a method
     * @param ident  the method ident
     * @param args  the argument of the call
     */
    public JMethodCallExpression(TokenReference where,
                                 JExpression prefix,
                                 JExpression[] args)
    {
        super(where);

        this.prefix = ((JNameExpression)prefix).getPrefix();
        this.ident = ((JNameExpression)prefix).getName().intern();
        this.args = args;
    }

    /**
     * @return the type of this expression.
     */
    public Type getType()
    {
        return method.getReturnType();
    }

    /**
     * Returns true iff this expression can be used as a statement. (JLS 14.7)
     */
    public boolean isStatementExpression()
    {
        return true;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        TokenReference where = getTokenReference();

        CClass local = context.getClassContext().getCClass();
        Type[] argTypes = new Type[args.length];

        for (int i = 0; i < argTypes.length; i++) {
            args[i] = args[i].checkExpression(context, false, false);
            argTypes[i] = args[i].getType();
        }

        if (ident == null) {
            if (prefix instanceof JClassFieldExpression) {
                // condition as if-statement because of arguments to method
                // check
                check(context,
                      false,
                      KjcMessages.FIELD_AS_METHOD,
                      ((JClassFieldExpression)prefix).getIdent());
            }
            if (prefix instanceof JTypeNameExpression) {
                // condition as if-statement because of arguments to method
                // check
                check(context,
                      false,
                      KjcMessages.TYPE_AS_METHOD,
                      ((JTypeNameExpression)prefix).getType());
            }
            check(context,
                  false,
                  KjcMessages.BAD_METHOD_NAME, prefix);
        }

        if (prefix != null) {
            prefix = prefix.checkExpression(context, false, false);

            if (prefix instanceof JNameExpression) {
                // condition as if-statement because of arguments to method
                // check
                check(context,
                      false,
                      KjcMessages.BAD_METHOD_NAME,
                      ((JNameExpression)prefix).getName());
            }

            Type prefixType = prefix.getType();

            /* Check if prefix is not primitive: one cannot do method calls on
             * a primitive type.
             */
            check(context,
                  prefixType.isReference(),
                  KjcMessages.METHOD_BADPREFIX,
                  ident, prefix.getType());

            try {
                CClass prefixClass = prefixType.toReference().getCClass();
                method = prefixClass.lookupMethod(local, ident, argTypes);
            }
            catch (UnpositionedError e) {
                throw e.addPosition(where);
            }
        }
        else {
            /* No prefix, so the method must be picked from the local context.
             */
            try {
                method = context.lookupMethod(local, ident, argTypes);
            }
            catch (UnpositionedError e) {
                throw e.addPosition(where);
            }
        }

        if (method == null) {
            String prefixName;

            if (prefix instanceof JNameExpression) {
                prefixName = ((JNameExpression)prefix).getQualifiedName()
                    + ".";
            }
            else if (prefix instanceof JTypeNameExpression) {
                prefixName = ((JTypeNameExpression)prefix).getQualifiedName()
                    + ".";
            }
            else if (prefix instanceof JSuperExpression) {
                prefixName = "super.";
            }
            else {
                prefixName = "";
            }

            throw new CMethodNotFoundError(where, this,
                                           prefixName + ident, argTypes);
        }

        // check access
        CClass access = method.getOwner();

        check(context,
              method.isAccessible(local),
              KjcMessages.METHOD_NOACCESS, method.getIdent());

        ClassOrInterfaceType[] exceptions = method.getThrowables();
        CBodyContext bodyContext = context.getBodyContext();
        CBodyContext snapshot = bodyContext.makeSnapshot();

        for (int i = 0; i < exceptions.length; i++) {
            ClassOrInterfaceType exception = exceptions[i];
            String exceptionClass = exception.getCClass().getQualifiedName();

            // if ( exception.isCheckedException()
            if ( prefix != null &&
                 prefix.getType().isArrayType() &&
                 ident == Constants.JAV_CLONE &&
                 exceptionClass.equals(CLONE_NOT_SUPPORTED) ) {
                continue;
            }

            CThrowableInfo info = new CThrowableInfo(exceptions[i],
                                                     this, snapshot,
                                                     true); // subTypePossible
            bodyContext.addThrowable(info);
        }

        /* New-object expressions can always result in a RuntimeException,
         * methods and constructors don't have to declare RuntimeExceptions.
         */
        CThrowableInfo info = new CThrowableInfo(StandardType.RUNTIMEEXCEPTION,
                                                 this, snapshot,
                                                 true); // subTypePossible
        bodyContext.addThrowable(info);

        if ((prefix == null) && !method.isStatic()) {
            if (access == local) {
                prefix = new JThisExpression(where);
            }
            else {
                prefix = new JThisExpression(where, access);
            }
            prefix = prefix.checkExpression(context);
        }

        boolean prefixIsTypeName = prefix instanceof JTypeNameExpression;

        check(context,
              method.isStatic() || !prefixIsTypeName,
              KjcMessages.METHOD_STATIC_BAD, method.getIdent());

        if (method.isStatic() && prefix != null && !prefixIsTypeName) {

            context.reportTrouble(new
                CWarning(where,
                         KjcMessages.INSTANCE_PREFIXES_STATIC_METHOD,
                         method.getIdent()));
        }

        argTypes = method.getParameters();
        for ( int i = 0; i < argTypes.length; i++ ) {

            if ( args[i] instanceof JTypeNameExpression ) {

                JTypeNameExpression arg = (JTypeNameExpression)args[i];

                check(context, false,
                      KjcMessages.VAR_UNKNOWN,
                      arg.getQualifiedName());
            }
            args[i] = args[i].convertType(argTypes[i], context);
        }

        // Mark method as used if it is a source method
        if (method instanceof CSourceMethod) {
            ((CSourceMethod)method).setUsed(true);
        }

        if ( prefix instanceof JSuperExpression ) {
            JSuperExpression superPrefix = (JSuperExpression)prefix;
            if (superPrefix.getPrefix() != null) {
                /* Make sure that super accessor method is generated in
                 * checking phase, so it will end up in the class file.
                 */
                ((CSourceMethod)method).getSuperAccessor();
            }
        }

        if (method.mustUseAccessorMethod(local)) {
            /* Make sure that accessor method is generated in checking phase,
             * so it will end up in the class file.
             */
            ((CSourceMethod)method).getAccessor();
            this.mustUseAccessorMethod = true;
        }

        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        if (ident == null) {
            ident = ((JNameExpression)prefix).getName();
            prefix = ((JNameExpression)prefix).getPrefix();
        }
        p.visitMethodCallExpression(this, prefix, ident, args);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        boolean forceNonVirtual = false;
        boolean useSuperAccessor = false;
        boolean discardValue = code.discardValue();

        code.setDiscardValue(false);

        if (prefix != null) {
            if (method.isStatic()) {
                code.setDiscardValue(true);
            }
            prefix.genCode(code);

            if (prefix instanceof JSuperExpression) {
                if (((JSuperExpression)prefix).getPrefix() == null) {
                    forceNonVirtual = true;
                }
                else {
                    useSuperAccessor = true;
                    /* Plant a 'null' argument as the first argument,
                     * because the first argument of the super-accessor
                     * is a dummy 'selector' argument with the type of
                     * the class containing the method, to avoid signature
                     * clashes with accessor methods in subclasses.
                     */
                    code.plantInstruction
                        (new NoArgInstruction(opc_aconst_null));
                }
            }
        }

        code.setDiscardValue(false);

        for (int i = 0; i < args.length; i++) {
            args[i].genCode(code);
        }

        if (mustUseAccessorMethod && !forceNonVirtual) {
            try {
                CSourceMethod accessor = ((CSourceMethod)method).getAccessor();
                accessor.genCallCode(code, false);
            }
            catch (PositionedError e) {
                // Shouldn't happen; no invalid member accessors should
                // be generated.
                throw new RuntimeException(e.toString());
            }
        }
        else if (useSuperAccessor) {
            try {
                CSourceMethod superAccessor =
                    ((CSourceMethod)method).getSuperAccessor();
                superAccessor.genCallCode(code, false);
            }
            catch (PositionedError e) {
                // Shouldn't happen; no invalid super accessors should
                // be generated.
                throw new RuntimeException(e.toString());
            }
        }
        else {
            method.genCallCode(code, forceNonVirtual);
        }

        if (discardValue) {
            switch (Types.getSize(getType())) {
            case 2:
                code.plantInstruction(new NoArgInstruction(opc_pop2));
                break;
            case 1:
                code.plantInstruction(new NoArgInstruction(opc_pop));
                break;
            }
        }
    }
}
