/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.ClassRefInstruction;
import at.dms.classfile.FieldRefInstruction;
import at.dms.classfile.Instruction;
import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * This class represents a new allocation expression with a special class type:
 * an anonymous inner class.
 */
public class JNewAnonymousClassExpression
    extends JNewObjectExpression
{
    /**
     * The declaration of the anonymous inner class.
     */
    protected JClassDeclaration declaration;

    /**
     * The superclass of the anonymous class.
     */
    private CClass superClass;

    /**
     * ...
     */
    private ClassOrInterfaceType superClassType;

    /**
     * ...
     */
    private JConstructorDeclaration constructor;

    /**
     * ...
     */
    private CClass anonymousClass;

    /**
     * ...
     */
    private CClass local;

    /**
     * True if the constructor is not directly accessible but only by creating
     * an accessor method in the class that calls the real constructor.
     */
    private boolean mustUseAccessorMethod = false;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param objectType
     * The type of this object allocator.
     * @param explicitThis
     * A TypeName expression.
     * @param args
     * Parameters to be passed to constructor.
     */
    public JNewAnonymousClassExpression(TokenReference where,
                                        ClassOrInterfaceType objectType,
                                        JExpression explicitThis,
                                        String identifier,
                                        JExpression[] args,
                                        JClassDeclaration declaration)
    {
        super(where, objectType, explicitThis, identifier, args);

        this.declaration = declaration;
        this.superClassType = objectType;
    }

    /**
     * If the declaration in which this expression occurs is static, then the
     * class declaration of the anonymous inner class is also static; pass
     * it on!
     */
    public void setDeclarationIsStatic(boolean flag)
    {
        if (flag) {
            declaration.setModifiers(declaration.getModifiers() | ACC_STATIC);
        }
        else {
            declaration.setModifiers(declaration.getModifiers() & ~ACC_STATIC);
        }
    }

    /**
     * Checks types of this expression.
     */
    protected Type[] resolveTypes(CExpressionContext context)
        throws PositionedError
    {
        Type[] argsTypes = super.resolveTypes(context);

        CClassContext classContext = context.getClassContext();
        CClass owner = classContext.getCClass();    
        TokenReference where = getTokenReference();

        superClassType = getClassType();

        superClass = superClassType.getCClass();

        if (superClass.isInterface()) {
            ClassOrInterfaceType[] interfaces =
                new ClassOrInterfaceType[] { superClassType };
            declaration.setInterfaces(interfaces);
        }
        else {
            declaration.setSuperClass(superClass.getQualifiedName());
        }

        if (context.getMethodContext().getCMethod().isStatic()) {
            declaration.setModifiers(declaration.getModifiers() | ACC_STATIC);
        }

        String syntheticIdent =
            String.valueOf(classContext.getNextSyntheticIndex()).intern();
        declaration.setIdent(syntheticIdent);

        declaration.generateInterface((Main)context.getCompiler(),
                                      owner, 
                                      owner.getQualifiedName() + "$");
        anonymousClass = declaration.getCClass();
        
        owner.addInnerClass(anonymousClass.getType());

        /* Because this construct goes completely against the normal order
         * of type checking, it is necessary to temporarily set the superclass
         * with the following line, to allow the look for the superclass-
         * constructor to find protected constructors.
         */
        anonymousClass.setSuperClass(superClassType);

        CMethod superConstructor;
        try {
            superConstructor = superClass.lookupMethod(anonymousClass,
                                                       JAV_CONSTRUCTOR,
                                                       argsTypes);
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(where);
        }

        if (superConstructor == null) {
            throw new CMethodNotFoundError(where, null,
                                           superClassType.toString(),
                                           argsTypes);
        }

        Type[] parameters = superConstructor.getParameters();
        ClassOrInterfaceType[] throwables = superConstructor.getThrowables();
        
        JFormalParameter[] fargs;
        JExpression prefix = null;

        if (superClass.hasOuterThis()) {
            fargs = new JFormalParameter[parameters.length + 1];
            fargs[0] = new JFormalParameter(where,
                                            JLocalVariable.DES_GENERATED,
                                            superClass.getOwner().getType(),
                                            "super$outer", true);
            prefix = new JLocalVariableExpression(where, fargs[0]);
        }
        else {
            fargs = new JFormalParameter[parameters.length];
        }

        for ( int i = 0, j = superClass.hasOuterThis() ? 1 : 0;
              i < parameters.length;
              i++, j++ ) {

            fargs[j] = new JFormalParameter(where,
                                            JLocalVariable.DES_GENERATED,
                                            parameters[i],
                                            "p$" + i,
                                            true);
        }

        JExpression[] checkedParams = new JExpression[args.length];

        for ( int i = 0, j = superClass.hasOuterThis() ? 1 : 0;
              i < args.length;
              i++, j++ ) {

            checkedParams[i] = new JLocalVariableExpression(where, fargs[j]);
        }

        JExpression superInvoke =
            new JExplicitConstructorInvocation(where, prefix, "super",
                                               checkedParams);
        JStatement statement = new JExpressionStatement(where, superInvoke,
                                                        null);
        JConstructorBlock body =
            new JConstructorBlock(where, new JStatement[] { statement });

        // add implicit constructor
        constructor = new JConstructorDeclaration(where,
                                                  ACC_PUBLIC,
                                                  anonymousClass.getIdent(),
                                                  fargs,
                                                  throwables,
                                                  body,
                                                  null, null);
        
        declaration.setDefaultConstructor(constructor);
        /* Make sure that the interface is checked with the local context,
         * so it can access other local types and variables.
         * So don't call with context.getCompilationUnitContext as was done
         * earlier.
         */
        declaration.checkInterface(context);
        declaration.checkInitializers(context);
        declaration.checkTypeBody(context);

        type = anonymousClass.getType();

        return argsTypes;
    }

    /**
     * ...
     */
    protected JExpression checkCall(CExpressionContext context,
                                    Type[] argsType)
        throws PositionedError
    {
        TokenReference where = getTokenReference();
        local = context.getClassContext().getCClass();
        CClass superOuter = superClass.getOwner();
        CClass localOuter = local.getOwner();
        CMethod constructorMethod = constructor.getMethod();

        context.setDiscardValue(false);

        if ( superClass.hasOuterThis() && explicitThis == null ) {
            check(context,
                  !context.getMethodContext().getCMethod().isStatic(),
                  KjcMessages.INNER_INHERITENCE,
                  superClass.getType(), local.getType());

            CClass outer = local;
            explicitThis = new JThisExpression(where);
            explicitThis.checkExpression(context);

            while (outer != null) {
                if (outer.descendsFrom(superOuter)) {
                    break;
                }
                explicitThis = new JClassFieldExpression(where,
                                                         explicitThis,
                                                         JAV_OUTER_THIS);
                explicitThis.checkExpression(context);
                outer = outer.getOwner();
            }

            check(context,
                  outer != null,
                  KjcMessages.INNER_INHERITENCE,
                  superClass.getType(), local.getType());
        }

        ClassOrInterfaceType[] exceptions = constructorMethod.getThrowables();
        CBodyContext bodyContext = context.getBodyContext();
        CBodyContext snapshot = bodyContext.makeSnapshot();
        for (int i = 0; i < exceptions.length; i++) {
            CThrowableInfo info = new CThrowableInfo(exceptions[i], this,
                                                     snapshot,
                                                     true); // subTypePossible
            bodyContext.addThrowable(info);
        }

        /* New-object expressions can always result in a RuntimeException,
         * methods and constructors don't have to declare RuntimeExceptions.
         */
        CThrowableInfo info = new CThrowableInfo(StandardType.RUNTIMEEXCEPTION,
                                                 this, snapshot,
                                                 true); // subTypePossible
        bodyContext.addThrowable(info);
    
        argsType = constructorMethod.getParameters();

        for ( int i = 0, j = superClass.hasOuterThis() ? 1 : 0;
              i < args.length;
              i++, j++ ) {

            args[i] = args[i].convertType(argsType[j], context);
        }

        if (constructorMethod.mustUseAccessorMethod(local)) {
            /* Make sure that accessor method is generated in checking phase,
             * so it will end up in the class file.
             */
            ((CSourceMethod)constructorMethod).getAccessor();
            this.mustUseAccessorMethod = true;
        }

        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitNewObjectExpression(this, type, args, declaration);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        CMethod constructorMethod = constructor.getMethod();
        boolean discardValue = code.discardValue();
        code.setLineNumber(getTokenReference().getLine());

        Instruction classRef =
            new ClassRefInstruction(opc_new,
                                    anonymousClass.getQualifiedName());
        code.plantInstruction(classRef);
        if (!discardValue) {
            code.plantInstruction(new NoArgInstruction(opc_dup));
        }
        code.setDiscardValue(false);

        /* If an outer-this is needed, it will be the class that contains
         * the call; the outer-this for the superclass is done below.
         */
        anonymousClass.genSyntheticParams(code, null); 

        if (superClass.hasOuterThis()) {
            explicitThis.genCode(code);
        }

        // Generate code for non-synthetic arguments for this 'new' call.
        for (int i = 0; i < args.length; i++) {
            args[i].genCode(code);
        }

        if (mustUseAccessorMethod) {
            code.plantInstruction(new NoArgInstruction(opc_aconst_null));
            try {
                CSourceMethod accessor =
                    ((CSourceMethod)constructorMethod).getAccessor();
                code.plantInstruction
                    (new MethodRefInstruction(opc_invokespecial,
                                              accessor.getPrefixName(),
                                              accessor.getIdent(),
                                              accessor.getSignature()));
            }
            catch (PositionedError e) {
                // Shouldn't happen; no invalid constructor accessors should
                // be generated.
                throw new RuntimeException(e.toString());
            }
        }
        else {
            code.plantInstruction
                (new MethodRefInstruction(opc_invokespecial,
                                          constructorMethod.getPrefixName(),
                                          constructorMethod.getIdent(),
                                          constructorMethod.getSignature()));
        }
    }
}
