/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import at.dms.classfile.PushLiteralInstruction;

import at.dms.compiler.NumberParser;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents ordinal literals (byte, short, int, long.)
 */
public strictfp class JOrdinalLiteral
    extends JNumberLiteral
{
    /**
     * ...
     */
    protected String image;

    /**
     * ...
     */
    private Number value;

    /**
     * ...
     */
    private Type type;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param image
     * The string representation of this literal.
     */
    public JOrdinalLiteral(TokenReference where, String image)
    {
        super(where);
        this.image = image;
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JOrdinalLiteral(TokenReference where,
                           Number value,
                           NumericType type)
    {
        super(where);
        setValue(value, type);
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JOrdinalLiteral(TokenReference where,
                           long value,
                           NumericType type)
    {
        this(where, new Long(value), type);
    }

    /**
     * This is needed to handle -2147483648 that is allowed agains 2147483648
     * that is not allowed, so we need to know if there is an unary minus
     * before us.
     */
    public void setSignToMinus()
    {
        if (image != null) {
            image = "-" + image;
        }
        if (value != null) {
            setValue(new Long(-value.longValue()));
        }
    }

    /**
     * Compute the type of this expression (called after parsing.)
     *
     * @return
     * The type of this expression.
     */
    public Type getType()
    {
        assert(type != null);
        return type;
    }

    /**
     * Sets the type if it is determined by a subclass.
     */
    protected void setType(Type type)
    {
        assert(type != null);
        this.type = type;
    }

    /**
     * Get a Number that represents the value of this literal.
     * Do not call before checkBody.
     */
    public Number getNumberValue()
    {
        assert(value != null);
        return value;
    }

    /**
     * Get a Number that represents the value of this literal.
     * Do not call before checkBody.
     */
    public Object getValue()
    {
        if (getType() == StandardType.LONG) {
            return value;
        }
        else {
            return new Integer(value.intValue());
        }
    }

    public void setValue(Number value)
    {
        setValue(value, (NumericType)getType());
    }

    /**
     * @param newValue
     * The new value.
     */
    protected void setValue(Number newValue, NumericType newType)
    {
        long val = newValue.longValue();

        if (val == newValue.byteValue()) {
            type = StandardType.BYTE;
            value = new Byte(newValue.byteValue());
        }
        else if (val == newValue.shortValue()) {
            type = StandardType.SHORT;
            value = new Short(newValue.shortValue());
        }
        else if (val <= Character.MAX_VALUE && val >= Character.MIN_VALUE) {
            type = StandardType.CHAR;
            value = new Integer(newValue.intValue());
        }
        else if (val == newValue.intValue()) {
            type = StandardType.INTEGER;
            value = new Integer(newValue.intValue());
        }
        else {
            type = StandardType.LONG;
            value = newValue;
        }

        type = newType;
    }

    /**
     * Special case (see JLS2nd 3.10.1): if preceded by a unary negative
     * expression, then and only then are integer literals allowed to be
     * 2147483648 or 9223372036854775808L.
     * @return
     * true if special case was applied; in that case checkExpression
     * must not be called. Otherwise, false is returned and checkExpression
     * is done normally. Only JUnaryMinusExpression may call this method.
     */
    boolean specialUnaryMinus()
    {
        if (value == null) {
            if (image.equals("2147483648")) {
                setValue(new Integer(Integer.MIN_VALUE), StandardType.INTEGER);
                return true;
            }
            else if (image.equalsIgnoreCase("9223372036854775808L")) {
                setValue(new Long(Long.MIN_VALUE), StandardType.LONG);
                return true;
            }
        }
        return false;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     *
     * @return
     * A pure java expression including promote node.

     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if (value == null) {
            try {
                String parseImage;
                char c = image.charAt(image.length() - 1);

                if (c == 'L' || c == 'l') {
                    parseImage = image.substring(0, image.length() - 1);
                    setValue(new Long(NumberParser.decodeLong(parseImage)),
                             StandardType.LONG);
                }
                else {
                    parseImage = image;
                    setValue(new Integer(NumberParser.decodeInt(parseImage)),
                             StandardType.INTEGER);
                }
                image = null;
            }
            catch (NumberFormatException ex) {
                ex.printStackTrace();
                check(context,
                      false,
                      KjcMessages.LITER_INT_BAD, image);
            }
        }

        return this;
    }

    /**
     * Changes the type of this expression to an other.
     *
     * @param dest
     * The destination type.
     */
    public JExpression convertType(Type dest, CExpressionContext context)
        throws PositionedError
    {
        assert(type != null);
        if (type.equals(dest)) {
            return this;
        }
        else if (dest == StandardType.BYTE) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Byte((byte)value.longValue()),
                                       StandardType.BYTE);
        }
        else if (dest == StandardType.SHORT) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Short((short)value.longValue()),
                                       StandardType.SHORT);
        }
        else if (dest == StandardType.CHAR) {
            return new JCharLiteral(getTokenReference(),
                                    new Integer((int)value.longValue()));
        }
        else if (dest == StandardType.INTEGER) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Integer((int)value.longValue()),
                                       StandardType.INTEGER);
        }
        else if (dest == StandardType.LONG) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Long(value.longValue()),
                                       StandardType.LONG);
        }
        else if (dest == StandardType.FLOAT) {
            return new JRealLiteral(getTokenReference(),
                                    new Float((float)value.longValue()),
                                    StandardType.FLOAT);
        }
        else if (dest == StandardType.DOUBLE) {
            return new JRealLiteral(getTokenReference(),
                                    new Double((double)value.longValue()),
                                    StandardType.DOUBLE);
        }
        else {
            fail();
            return null;
        }
    }

    /**
     * Can this expression be converted to the specified type by assignment
     * conversion (JLS 5.2)?
     *
     * @param dest
     * The destination type.
     * @return
     * true iff the casting conversion is valid.
     */
    public boolean allowsAssignmentConversion(Type dest)
    {
        if (type != StandardType.BYTE && type != StandardType.SHORT
            && type != StandardType.CHAR && type != StandardType.INTEGER) {

            return type.allowsAssignmentConversion(dest);
        }
        else {
            long val = getNumberValue().longValue();
            if (dest == StandardType.BYTE) {
                return (byte)val == val;
            }
            else if (dest == StandardType.SHORT) {
                return (short)val == val;
            }
            else if (dest == StandardType.CHAR) {
                return (char)val == val;
            }
            else {
                return type.allowsAssignmentConversion(dest);
            }
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        assert(value != null);

        if (type == StandardType.BYTE) {
            p.visitByteLiteral(value.byteValue());
        }
        else if (type == StandardType.CHAR) {
            p.visitCharLiteral((char)value.intValue());
        }
        else if (type == StandardType.INTEGER) {
            p.visitIntLiteral(value.intValue());
        }
        else if (type == StandardType.LONG) {
            p.visitLongLiteral(value.longValue());
        }
        else if (type == StandardType.SHORT) {
            p.visitShortLiteral(value.shortValue());
        }
        else {
            fail("Type: " + type);
            // fail();
        }
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (type == StandardType.LONG) {
            code.plantInstruction
                (new PushLiteralInstruction(getNumberValue().longValue()));
        }
        else {
            code.plantInstruction
                (new PushLiteralInstruction(getNumberValue().intValue()));
        }
    }
}
