/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.IincInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.codegen.jvm.Types;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 * This class represents postfix increment and decrement expressions.
 * 15.13.2 Postfix Increment Operator ++
 * 15.13.3 Postfix Decrement Operator --
 */
public class JPostfixExpression
    extends JExpression
{
    /**
     * The type of operation.
     */
    protected int oper;

    /**
     * The expression to operate on.
     */
    protected JExpression expr;

    /**
     * The type of the result of the operation.
     */
    protected Type type;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param oper
     * The operator.
     * @param expr
     * The operand.
     */
    public JPostfixExpression(TokenReference where,
                              int oper,
                              JExpression expr)
    {
        super(where);
        this.oper = oper;
        this.expr = expr;
    }

    /**
     * Compute the type of this expression.
     *
     * @return the type of this expression.
     */
    public Type getType()
    {
        return type;
    }

    /**
     * Returns true iff this expression can be used as a statement. (JLS 14.7)
     */
    public boolean isStatementExpression()
    {
        return true;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        expr = expr.checkExpression(context, false, context.isDiscardValue());

        check(context, expr.getType().isNumeric(),
              KjcMessages.POSTFIX_BADTYPE, expr.getType());
        check(context,
              (expr instanceof CInitializable)
              && ((CInitializable)expr).isLValue(context),
              KjcMessages.POSTFIX_NOTLVALUE);
        check(context,
              ((CInitializable)expr).isInitialized(context),
              KjcMessages.POSTFIX_NOTINITIALIZED);

        type = expr.getType();

        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitPostfixExpression(this, oper, expr);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());
        boolean  discardValue = code.discardValue();
        int   val = oper == OPE_POSTINC ? 1 : -1;

        if ((expr.getType() == StandardType.INTEGER)
            && (expr instanceof JLocalVariableExpression)) {

            expr.genCode(code);
            code.plantInstruction
                (new IincInstruction(((JLocalVariableExpression)expr)
                                     .getPosition(), val));
        }
        else {
            if (discardValue) {
                code.setDiscardValue(false);
                ((CInitializable)expr).genStartStoreCode(code);
                expr.genCode(code);
            }
            else {
                ((CInitializable)expr).genStartStoreCode(code);
                expr.genCode(code);
                switch (Types.getSize(getType())) {
                case 2:
                    if ((expr instanceof JLocalVariableExpression)
                        || ((expr instanceof JClassFieldExpression)
                            && ((JClassFieldExpression)expr)
                            .getField().isStatic())) {
                        
                        code.plantInstruction(new NoArgInstruction(opc_dup2));
                    }
                    else if (expr instanceof JArrayAccessExpression) {
                        code.plantInstruction
                            (new NoArgInstruction(opc_dup2_x2));
                    }
                    else {
                        code.plantInstruction
                            (new NoArgInstruction(opc_dup2_x1));
                    }
                    break;
                case 1:
                    if ( (expr instanceof JLocalVariableExpression) ||
                         ((expr instanceof JClassFieldExpression)
                          && ((JClassFieldExpression)expr)
                          .getField().isStatic())) {
                        
                        code.plantInstruction(new NoArgInstruction(opc_dup));
                    }
                    else if (expr instanceof JArrayAccessExpression) {
                        code.plantInstruction
                            (new NoArgInstruction(opc_dup_x2));
                    }
                    else {
                        code.plantInstruction
                            (new NoArgInstruction(opc_dup_x1));
                    }
                    break;
                }
            }
            
            switch (expr.getType().getTypeID()) {
            case TID_FLOAT:
                code.plantInstruction(new PushLiteralInstruction((float)val));
                code.plantInstruction(new NoArgInstruction(opc_fadd));
                break;
            case TID_LONG:
                code.plantInstruction(new PushLiteralInstruction((long)val));
                code.plantInstruction(new NoArgInstruction(opc_ladd));
                break;
            case TID_DOUBLE:
                code.plantInstruction(new PushLiteralInstruction((double)val));
                code.plantInstruction(new NoArgInstruction(opc_dadd));
                break;
                
            case TID_BYTE:
                code.plantInstruction(new PushLiteralInstruction(val));
                code.plantInstruction(new NoArgInstruction(opc_iadd));
                code.plantInstruction(new NoArgInstruction(opc_i2b));
                break;
            case TID_CHAR:
                code.plantInstruction(new PushLiteralInstruction(val));
                code.plantInstruction(new NoArgInstruction(opc_iadd));
                code.plantInstruction(new NoArgInstruction(opc_i2c));
                break;
            case TID_SHORT:
                code.plantInstruction(new PushLiteralInstruction(val));
                code.plantInstruction(new NoArgInstruction(opc_iadd));
                code.plantInstruction(new NoArgInstruction(opc_i2s));
                break;
            case TID_INT:
                code.plantInstruction(new PushLiteralInstruction(val));
                code.plantInstruction(new NoArgInstruction(opc_iadd));
                break;
            }
            
            ((CInitializable)expr).genEndStoreCode(code, true);
        }
    }
}

