/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import java.math.BigDecimal;

import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Real literals.
 */
public strictfp class JRealLiteral
    extends JNumberLiteral
{
    /**
     * Textual image of real literal.
     */
    private String image;

    /**
     * Parsed numeric value of real literal.
     */
    private Number value;

    /**
     * Type of real literal (StandardType.FLOAT or StandardType.DOUBLE)
     */
    private Type type;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code.
     * @param image  the string representation of this literal.
     */
    public JRealLiteral(TokenReference where, String image)
    {
        super(where);
        this.image = image;
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code.
     * @param image  the string representation of this literal.
     */
    public JRealLiteral(TokenReference where, Number value, Type type)
    {
        super(where);
        assert(type == StandardType.DOUBLE || type == StandardType.FLOAT);
        setValue(value);
        this.type = type;
    }

    /**
     * Compute the type of this expression. (called after parsing)
     * @return the type of this expression.
     */
    public Type getType()
    {
        return type;
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before checkBody
     */
    public Number getNumberValue()
    {
        assert(value != null);
        return value;
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before checkBody
     */
    public Object getValue()
    {
        return getNumberValue();
    }

    /**
     * @param newValue the new value.
     */
    private void setValue(Number newValue)
    {
        type = StandardType.DOUBLE;
        value = newValue;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     */
    public strictfp JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if (value != null) {
            return this; // already checked
        }

        BigDecimal bigNum;
        BigDecimal bigZero = new BigDecimal("0");

        char c = image.charAt(image.length() - 1);
        String parseImage;

        if (c == 'f' || c == 'F' || c == 'd' | c == 'D') {
            parseImage = image.substring(0, image.length() - 1);
        }
        else {
            parseImage = image;
        }

        try {
            bigNum = new BigDecimal(parseImage);
        }
        catch (NumberFormatException ex) {
            check(context,
                  false,
                  KjcMessages.LITER_REAL_BAD, image);
            return null;
        }

        if (image.endsWith("f") || image.endsWith("F")) {
            Float value;

            try {
                value = Float.valueOf(parseImage);
            }
            catch (NumberFormatException ex) {
                check(context,
                      false,
                      KjcMessages.LITER_FLOAT_BAD, image);
                return null;
            }

            if (value.floatValue() == 0f && bigNum.compareTo(bigZero) != 0) {
                /* If conversion to float is zero but bignum is not, then
                 * value is out of range.
                 */
                check(context,
                      false,
                      KjcMessages.LITER_FLOAT_RANGE, image);
            }
            else if ( value.isInfinite() ) {
                check(context,
                      false,
                      KjcMessages.LITER_FLOAT_RANGE, image);
            }
            else {
                setValue(value);
                type = StandardType.FLOAT;
            }
        }
        else {
            Double value;

            try {
                value = Double.valueOf(parseImage);
            }
            catch (NumberFormatException ex) {
                check(context,
                      false,
                      KjcMessages.LITER_DOUBLE_BAD, image);
                return null;
            }

            if (value.doubleValue() == 0d && bigNum.compareTo(bigZero) != 0) {
                /* If conversion to double is zero but bignum is not, then
                 * value is out of range.
                 */
                check(context,
                      false,
                      KjcMessages.LITER_DOUBLE_RANGE, image);
            }
            else if ( value.isInfinite() ) {
                check(context,
                      false,
                      KjcMessages.LITER_DOUBLE_RANGE, image);
            }
            else {
                setValue(value);
                type = StandardType.DOUBLE;
            }
        }

        return this;
    }

    /**
     * Changes the type of this expression to an other type.
     *
     * @param dest
     * The destination type.
     */
    public JExpression convertType(Type dest, CExpressionContext context)
        throws PositionedError
    {
        assert(type != null);
        if (type.equals(dest)) {
            return this;
        }
        else if (dest == StandardType.BYTE) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Byte(value.byteValue()),
                                       StandardType.BYTE);
        }
        else if (dest == StandardType.SHORT) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Short(value.shortValue()),
                                       StandardType.SHORT);
        }
        else if (dest == StandardType.CHAR) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Integer((char)value.intValue()),
                                       StandardType.CHAR);
        }
        else if (dest == StandardType.INTEGER) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Integer(value.intValue()),
                                       StandardType.INTEGER);
        }
        else if (dest == StandardType.LONG) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Long(value.longValue()),
                                       StandardType.LONG);
        }
        else if (dest == StandardType.FLOAT) {
            return new JRealLiteral(getTokenReference(),
                                    new Float(value.floatValue()),
                                    StandardType.FLOAT);
        }
        else if (dest == StandardType.DOUBLE) {
            return new JRealLiteral(getTokenReference(),
                                    new Double(value.doubleValue()),
                                    StandardType.DOUBLE);
        }
        else {
            fail();
            return null;
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        assert(value != null);
        if (type == StandardType.DOUBLE) {
            p.visitDoubleLiteral(value.doubleValue());
        }
        else if (type == StandardType.FLOAT) {
            p.visitFloatLiteral(value.floatValue());
        }
        else {
            fail();
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (type == StandardType.FLOAT) {
            code.plantInstruction
                (new PushLiteralInstruction(value.floatValue()));
        }
        else {
            code.plantInstruction
                (new PushLiteralInstruction(value.doubleValue()));
        }
    }
}
