/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.codegen.jvm.Types;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 */
public class JRelationalExpression
    extends JBinaryExpression
{
    /**
     * Operation type.
     */
    protected int oper;

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where  the line of this node in the source code
     * @param oper  the operator
     * @param left  the left operand
     * @param right  the right operand
     */
    public JRelationalExpression(TokenReference where,
                                 int oper,
                                 JExpression left,
                                 JExpression right)
    {
        super(where, left, right);
        this.oper = oper;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        check(context, left.getType().isNumeric() &&
              right.getType().isNumeric(), KjcMessages.RELATIONAL_TYPE,
              left.getType(),
              right.getType());

        type = StandardType.BOOLEAN;

        NumericType numLeft = left.getType().toNumeric();
        NumericType numRight = right.getType().toNumeric();
        Type binary = numLeft.binaryPromoteWith(numRight);
        left = left.convertType(binary, context);
        right = right.convertType(binary, context);

        if (left.isConstant() && right.isConstant()) {
            return constantFolding(left.getNumberLiteral(),
                                   right.getNumberLiteral());
        }

        return this;
    }

    /**
     * @param left  the left literal
     * @param right  the right literal
     * @return a literal resulting of an operation over two literals
     */
    public JExpression constantFolding(JNumberLiteral left,
                                       JNumberLiteral right)
    {
        switch (oper) {
        case OPE_LT:
            return new JBooleanLiteral(getTokenReference(),
                                       left.getNumberValue().doubleValue() <
                                       right.getNumberValue().doubleValue());
        case OPE_LE:
            return new JBooleanLiteral(getTokenReference(),
                                       left.getNumberValue().doubleValue() <=
                                       right.getNumberValue().doubleValue());
        case OPE_GT:
            return new JBooleanLiteral(getTokenReference(),
                                       left.getNumberValue().doubleValue() >
                                       right.getNumberValue().doubleValue());
        case OPE_GE:
            return new JBooleanLiteral(getTokenReference(),
                                       left.getNumberValue().doubleValue() <=
                                       right.getNumberValue().doubleValue());
        default:
            fail();
            return null;
        }
    }

    /**
     * Accepts the specified visitor.
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitRelationalExpression(this, oper, left, right);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
                             JExpression right,
                             boolean cond,
                             CodeSequence code,
                             CodeLabel label)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (((left instanceof JOrdinalLiteral
              && ((JOrdinalLiteral)left).getNumberValue().intValue() == 0)
             || (right instanceof JOrdinalLiteral
                 && ((JOrdinalLiteral)right).getNumberValue().intValue() == 0))
            && Types.getSize(left.getType()) == 1) {

            int opcode = -1;
            if (left instanceof JOrdinalLiteral
                && ((JOrdinalLiteral)left).getNumberValue().intValue() == 0) {

                right.genCode(code);
                switch (oper) {
                case OPE_LT:
                    opcode = cond ? opc_ifgt : opc_ifle;
                    break;
                case OPE_LE:
                    opcode = cond ? opc_ifge : opc_iflt;
                    break;
                case OPE_GT:
                    opcode = cond ? opc_iflt : opc_ifge;
                    break;
                case OPE_GE:
                    opcode = cond ? opc_ifle : opc_ifgt;
                    break;
                }
            }
            else {
                left.genCode(code);
                switch (oper) {
                case OPE_LT:
                    opcode = cond ? opc_iflt : opc_ifge;
                    break;
                case OPE_LE:
                    opcode = cond ? opc_ifle : opc_ifgt;
                    break;
                case OPE_GT:
                    opcode = cond ? opc_ifgt : opc_ifle;
                    break;
                case OPE_GE:
                    opcode = cond ? opc_ifge : opc_iflt;
                    break;
                }
            }
            code.plantLabelRef("relationalExpr:genBranch/1",
                               opcode, label);
            return;
        }

        left.genCode(code);
        right.genCode(code);

        int   opcode = -1;
        switch (left.getType().getTypeID()) {
        case TID_FLOAT:
            code.plantInstruction(new NoArgInstruction(opc_fcmpl));
            break;
        case TID_LONG:
            code.plantInstruction(new NoArgInstruction(opc_lcmp));
            break;
        case TID_DOUBLE:
            code.plantInstruction(new NoArgInstruction(opc_dcmpl));
            break;
        default:
            switch (oper) {
            case OPE_LT:
                opcode = cond ? opc_if_icmplt : opc_if_icmpge;
                break;
            case OPE_LE:
                opcode = cond ? opc_if_icmple : opc_if_icmpgt;
                break;
            case OPE_GT:
                opcode = cond ? opc_if_icmpgt : opc_if_icmple;
                break;
            case OPE_GE:
                opcode = cond ? opc_if_icmpge : opc_if_icmplt;
                break;
            }
            code.plantLabelRef("relationalExpr:genBranch/2",
                               opcode, label);
            return;
        }
        switch (oper) {
        case OPE_LT:
            opcode = cond ? opc_iflt : opc_ifge;
            break;
        case OPE_LE:
            opcode = cond ? opc_ifle : opc_ifgt;
            break;
        case OPE_GT:
            opcode = cond ? opc_ifgt : opc_ifle;
            break;
        case OPE_GE:
            opcode = cond ? opc_ifge : opc_iflt;
            break;
        }

        code.plantLabelRef("relationalExpr:genBranch/3",
                           opcode, label);
    }
}
