/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;
import at.dms.codegen.jvm.Types;
import at.dms.type.*;

/**
 * Return Statement. (JLS 14.15)
 *
 * A return statement returns control to the invoker of a method or
 * constructor.
 */
public class JReturnStatement
    extends JStatement
{
    /**
     * The expression that this statement returns.
     */
    protected JExpression expression;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param expression
     * The expression to return.
     */
    public JReturnStatement(TokenReference where,
                            JExpression expression,
                            JavaStyleComment[] comments)
    {
        super(where, comments);
        this.expression = expression;
    }

    /**
     * Returns the type of this return statement.
     */
    public Type getType()
    {
        return expression != null ? expression.getType() : StandardType.VOID;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        Type returnType =
            context.getMethodContext().getCMethod().getReturnType();

        if (expression != null) {
            check(context,
                  returnType != StandardType.VOID,
                  KjcMessages.RETURN_NONEMPTY_VOID);

            CExpressionContext expressionContext =
                context.createExpressionContext();

            expression = expression.checkExpression(expressionContext);

            check(context,
                  expression.allowsAssignmentConversion(returnType),
                  KjcMessages.RETURN_BADTYPE,
                  expression.getType(), returnType);

            expression = expression.convertType(returnType, expressionContext);

        }
        else {
            check(context,
                  returnType == StandardType.VOID,
                  KjcMessages.RETURN_EMPTY_NONVOID);
        }

        context.addReturn();
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitReturnStatement(this, expression);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (expression != null) {
            expression.genCode(code);

            code.plantReturn(this);
            int opcode = Types.getReturnOpcode(expression.getType());
            code.plantInstruction(new NoArgInstruction(opcode));
        }
        else {
            code.plantReturn(this);
            code.plantInstruction(new NoArgInstruction(opc_return));
        }
    }

    /**
     * Load the value from a local var (after finally)
     */
    public void load(CodeSequence code, JLocalVariable var)
    {
        code.plantLocalVar(Types.getLoadOpcode(expression.getType()), var);
    }

    /**
     * Load the value from a local var (after finally)
     */
    public void store(CodeSequence code, JLocalVariable var)
    {
        code.plantLocalVar(Types.getStoreOpcode(expression.getType()), var);
    }
}
