/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.MessageDescription;
import at.dms.type.*;

/**
 * A representation of a bitshift operation. See JLS2 15.19
 *
 * @author DMS, Erwin Bolwidt
 */
public class JShiftExpression
    extends JBinaryArithmeticExpression
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the line of this node in the source code
     * @param oper the operator
     * @param left the left operand
     * @param right the right operand
     */
    public JShiftExpression(TokenReference where,
                            int oper, JExpression left, JExpression right)
    {
        super(where, left, right);
        this.oper = oper;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        return checkBinaryBody(context);
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     */
    public JExpression checkBinaryBody(CExpressionContext context)
        throws PositionedError
    {
        try {
            type = computeType(left.getType(), right.getType(),
                               KjcMessages.SHIFT_BADTYPE);

            if (left.isConstant() && right.isConstant()) {
                return constantFolding();
            }
            else {
                left = left.convertType(type, context);
                right = right.convertType(StandardType.INTEGER, context);
                return this;
            }
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }
    }

    /**
     * compute the type of this expression according to operands
     * @param leftType  the type of left operand
     * @param rightType  the type of right operand
     * @return the type computed for this binary operation
     * @exception UnpositionedError this error will be positioned soon
     */
    public static Type computeType(Type leftType, Type rightType,
                                   MessageDescription badTypeMessage)
        throws UnpositionedError
    {
        if (leftType.isIntegral() && rightType.isIntegral()) {
            return leftType.toNumeric().unaryPromote();
        }
        else {
            throw new UnpositionedError(badTypeMessage,
                                        leftType, rightType);
        }
    }

    /**
     * @param left  the left literal
     * @param right  the right literal
     * @return a literal resulting of an operation over two literals
     */
    public JExpression constantFolding()
    {
        Type leftType = left.getType();
        Type rightType = right.getType();
        Number leftNumber = left.getNumberLiteral().getNumberValue();
        Number rightNumber = right.getNumberLiteral().getNumberValue();


        if (left.getType() == StandardType.LONG) {
            long result = compute(leftNumber.longValue(),
                                  rightNumber.intValue());
            return new JOrdinalLiteral(getTokenReference(),
                                       new Long(result), StandardType.LONG);
        }
        else {
            int result = compute(leftNumber.intValue(),
                                 rightNumber.intValue());
            return new JOrdinalLiteral(getTokenReference(),
                                       new Integer(result), StandardType.INTEGER);
        }
    }

    /**
     * Computes the result of the operation at compile-time.
     * See JLS2 section 15.19.
     *
     * @param left the first operand
     * @param right the second operand
     * @return the result of the operation
     */
    public int compute(int left, int right)
    {
        switch (oper) {
        case OPE_SL:
            return left << right;
        case OPE_SR:
            return left >> right;
        case OPE_BSR:
            return left >>> right;
        default:
            fail();
            return 0;
        }
    }

    /**
     * Computes the result of the shift operation at compile-time.
     * See JLS2 section 15.19.
     * @param left the first operand
     * @param right the second operand
     * @return the result of the operation
     */
    public long compute(long left, int right)
    {
        switch (oper) {
        case OPE_SL:
            return left << right;
        case OPE_SR:
            return left >> right;
        case OPE_BSR:
            return left >>> right;
        default:
            fail();
            return 0L;
        }
    }

    /**
     * Cannot perform shift operation on two longs. (The right expression,
     * if it was long, has been truncated to an int before calling the
     * compute methods, which is allowed by JLS2 15.19)
     */
    public long compute(long left, long right)
    {
        fail();
        return 0l;
    }

    /**
     * Cannot perform shift operation on floats.
     */
    public float compute(float left, float right)
    {
        fail();
        return 0f;
    }

    /**
     * Cannot perform shift operation on doubles.
     */
    public double compute(double left, double right)
    {
        fail();
        return 0d;
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitShiftExpression(this, oper, left, right);
    }


    /**
     * @param type the type of result
     * @return the type of opcode for this operation
     */
    public static int getOpcode(int oper, Type type)
    {
        if (type == StandardType.LONG) {
            switch (oper) {
            case OPE_SL:
                return opc_lshl;
            case OPE_SR:
                return opc_lshr;
            case OPE_BSR:
                return opc_lushr;
            }
        }
        else {
            switch (oper) {
            case OPE_SL:
                return opc_ishl;
            case OPE_SR:
                return opc_ishr;
            case OPE_BSR:
                return opc_iushr;
            }
        }
        return -1;
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        left.genCode(code);
        right.genCode(code);

        code.plantInstruction(new NoArgInstruction(getOpcode(oper,
                                                             getType())));
    }

    protected int oper;
}
