/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import java.util.List;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents a group of statements in a switch statement, with
 * a set of switch labels label before the group.
 */
public class JSwitchGroup
    extends JPhylum
{
    /**
     * Labels before this switch group.
     */
    private JSwitchLabel[] labels;

    /**
     * Statements inside this switch group.
     */
    private JStatement[] statements;

    /**
     * Location in bytecode at which this switch group starts.
     */
    private CodeLabel pos;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param labels
     * A group of labels.
     * @param statements
     * A group of statements.
     */
    public JSwitchGroup(TokenReference where,
                        JSwitchLabel[] labels,
                        JStatement[] statements)
    {
        super(where);

        this.labels = labels;
        this.statements = statements;
    }

    /**
     * Returns a list of statements.
     */
    public JStatement[] getStatements()
    {
        return statements;
    }

    /**
     * ...
     */
    public JSwitchLabel[] getLabels()
    {
        return labels;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkBody(CSwitchBodyContext context, CBodyContext body)
        throws PositionedError
    {
        for (int i = 0; i < labels.length; i++) {
            labels[i].checkBody(context, body);
        }

        body.setReachable(true);
        for (int i = 0; i < statements.length; i++) {
            try {
                if (!body.isReachable()) {
                    throw new CLineError(statements[i].getTokenReference(),
                                         KjcMessages.STATEMENT_UNREACHABLE);
                }
                statements[i].checkBody(body);
            }
            catch (CLineError le) {
                context.reportTrouble(le);
            }
        }

        return;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitSwitchGroup(this, labels, statements);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param matches
     * A list of values to match.
     * @param targets
     * A list of target labels.
     */
    public void collectLabels(CodeLabel defaultLabel,
                              List matches, List targets)
    {
        pos = null;

        // check if one of the labels is "default:"
        for (int i = 0; pos == null && i < this.labels.length; i++) {
            if (this.labels[i].isDefault()) {
                pos = defaultLabel;
            }
        }

        if (pos == null) {
            // no default: define a new label
            pos = new CodeLabel();

            for (int i = 0; i < this.labels.length; i++) {
                matches.add(this.labels[i].getLabel());
                targets.add(pos);
            }
        }
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        code.plantLabel(pos);
        for (int i = 0; i < statements.length; i++) {
            statements[i].genCode(code);
        }
    }
}
