/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.MessageDescription;
import at.dms.type.*;

/**
 * A case: or default: label inside a switch statement.
 */
public class JSwitchLabel
    extends JPhylum
{
    /**
     * Evaluated label.
     */
    private Integer label;

    /**
     * Expression, which must be a constant expression.
     */
    private JExpression expr;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param expr
     * The expression, or null if default label.
     */
    public JSwitchLabel(TokenReference where, JExpression expr)
    {
        super(where);

        this.expr = expr;
    }

    /**
     * @return true if this label is a "default:" label
     */
    public boolean isDefault()
    {
        return expr == null;
    }

    /**
     * @return the value of this label
     */
    public Integer getLabel()
    {
        return label;
    }

    /**
     * Check body.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkBody(CSwitchBodyContext switchContext,
                          CBodyContext context)
        throws PositionedError
    {
        try {
            if (expr != null) {
                expr = expr.checkExpression(context.createExpressionContext());
                Type exprType = expr.getType();

                check(context,
                      exprType.isNumeric() &&
                      ( exprType.toNumeric().unaryPromote() ==
                        StandardType.INTEGER ),
                      KjcMessages.SWITCH_LABEL_EXPR_NOTINT, exprType);

                check(context,
                      expr.isConstant(),
                      KjcMessages.SWITCH_LABEL_EXPR_NOTCONST);

                switchContext.addLabel(expr.getOrdinalLiteral());
                label = new Integer
                    (expr.getOrdinalLiteral().getNumberValue().intValue());
            }
            else {
                switchContext.addDefault();
            }
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
    }

    /**
     * Adds a compiler error.
     *
     * @param context
     * The context in which the error occurred.
     * @param key
     * The message ident to be displayed.
     * @param params
     * The array of parameters.
     *
     */
    protected void fail(CContext context,
                        MessageDescription key, Object[] params)
        throws PositionedError
    {
        throw new CLineError(getTokenReference(), key, params);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitSwitchLabel(this, expr);
    }
}
