/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.type.*;

/**
 * A 'this' expression.
 */
public class JThisExpression
    extends JExpression
{
    /**
     * Special variable that references the outer-this variable inside a
     * constructor; in a constructor, you don't know if the outer-this
     * parameter has already been assigned to the this$0 class field,
     * so you always use the variable.
     */
    private class ConstructorOuterThis
        extends JGeneratedLocalVariable
    {
        /**
         * Construct one.
         */
        ConstructorOuterThis(ClassOrInterfaceType type)
        {
             super(null,
                   0,
                   type,
                   type.getCClass().getIdent() + ".this",
                   null);
        }

        /**
         * @return the local index in context variable
         * table.
         */
        public int getPosition()
        {
            /* outer this; 0 is inner this */
            return 1;
        }
    }

    /**
     * The class type of the prefix.
     */
    private CClass self;

    /**
     * The prefix of the "this" expression.
     */
    private JExpression prefix;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JThisExpression(TokenReference where)
    {
        super(where);
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     *
     * @param self
     * The class onto this suffix is applied.
     */
    public JThisExpression(TokenReference where, CClass self)
    {
        this(where);
        this.self = self;
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     *
     * @param prefix
     * The left expression like t.this.
     */
    public JThisExpression(TokenReference where, JExpression prefix)
    {
        this(where);
        this.prefix = prefix;
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public Type getType()
    {
        return self.getType();
    }

    /**
     * @return is this expression a lvalue ?
     */
    public boolean isLValue(CExpressionContext context)
    {
        return false;
    }

    /**
     * Returns the prefix of this JThisExpression, if it has any; otherwise,
     * returns <code>null</code>.
     */
    public JExpression getPrefix()
    {
        return prefix;
    }

    /**
     * Sets the prefix of this JThisExpression.
     */
    public void setPrefix(JExpression prefix)
    {
        this.prefix = prefix;
    }

    public CClass getSelf()
    {
        return self;
    }

    public void setSelf(CClass self)
    {
        this.self = self;
    }

    private void checkSelf(final CExpressionContext context)
        throws PositionedError
    {
        if (prefix != null) {
            prefix = prefix.checkExpression(context);
            check(context,
                  prefix.getType().isClassOrInterfaceType(),
                  KjcMessages.THIS_BADACCESS);
            self = prefix.getType().toClassOrInterface().getCClass();
        }
        else if (self == null) {
            self = context.getClassContext().getCClass();
        }
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public JExpression checkExpression(final CExpressionContext context)
        throws PositionedError
    {
        TokenReference where = getTokenReference();

        checkSelf(context);

        if (!context.getClassContext().getCClass().descendsFrom(self)) {

            // access to outer class
            JExpression expr = null;
            CClassContext classContext = context.getClassContext();
            CClass clazz = classContext.getCClass();

            if (!clazz.descendsFrom(self) && expr == null) {
                /* Check if outer class is static. In that case, it is not
                 * allowed to explicitly reference a field in it from an
                 * inner class.
                 */
                check(context, 
                      !classContext.getTypeDeclaration().getCClass()
                      .isStatic(),
                      KjcMessages.THIS_INVALID_OUTER);

                // Make sure that outer this field has been generated.
                classContext.getTypeDeclaration().addOuterThis();
 
                CMethodContext myMethod = context.getMethodContext();
                
                if (myMethod instanceof CConstructorContext) {
                    /* Inside a constructor, things are different.
                     * The outer-this is still inside a local variable,
                     * not necessarily already in a class field.
                     */
                    JLocalVariable local =
                        new ConstructorOuterThis(clazz.getOwner().getType());
                    
                    expr = new JLocalVariableExpression(where, local) {
                            public JExpression checkExpression
                                (CExpressionContext ctxt)
                            {
                                // already checked
                                return this;
                            }
                        };
                }
                else {
                    expr = new JClassFieldExpression
                        (where,
                         new JThisExpression(where),
                         JAV_OUTER_THIS);
                }

                expr = expr.checkExpression(context);
                // Go to parent context.
                classContext =
                    classContext.getParentContext().getClassContext();
                clazz = ((ClassOrInterfaceType)expr.getType()).getCClass();
            }

            /* Trying to access field defined in an outer class (direct or
             * indirect outer class): so search for the field until it is
             * found or there is no enclosing outer class anymore.
             * While doing that, an expression is built to access the field;
             * the expression looks like: this$0.this$0.this$0.field
             */
            while (!clazz.descendsFrom(self)) {

                /* Check if outer class is static. In that case, it is not
                 * allowed to explicitly reference a field in it from an
                 * inner class.
                 */
                check(context, 
                      !classContext.getTypeDeclaration().getCClass()
                      .isStatic(),
                      KjcMessages.THIS_INVALID_OUTER);

                // Make sure that outer this field has been generated.
                classContext.getTypeDeclaration().addOuterThis();
 
                expr = new JClassFieldExpression(where, expr, JAV_OUTER_THIS);

                expr = expr.checkExpression(context);
                // Go to parent context.
                classContext =
                    classContext.getParentContext().getClassContext();
                clazz = ((ClassOrInterfaceType)expr.getType()).getCClass();
            }

            if (prefix != null) {
                check(context,
                      expr.getType().equals(prefix.getType())
                      ||
                      /* May be it is an innerclass with the same name,
                       * therefore the prefix name has been wrongly assigned
                       * to the outer. So we compare the names instead:*/
                      ((ClassOrInterfaceType)expr.getType()).getCClass().getIdent()
                      .equals(((ClassOrInterfaceType)prefix.getType()).getCClass()
                              .getIdent()),
                      KjcMessages.THIS_INVALID_OUTER, prefix.getType());
            }
            return expr;
        }

        check(context,
              !context.getMethodContext().getCMethod().isStatic(),
              KjcMessages.BAD_THIS_STATIC);

        return this;
    }

    /**
     * ...
     */
    public boolean equals(Object o)
    {
        return o instanceof JThisExpression &&
            self.equals(((JThisExpression)o).self);
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitThisExpression(this, prefix);
    }

    /**
     * Generates a sequence of bytecodes.
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        if (!code.discardValue()) {
            code.setLineNumber(getTokenReference().getLine());

            code.plantLoadThis();
        }
    }
}
