/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.type.*;

import java.util.Iterator;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * Try Statement: JLS 14.18.
 *
 * A try statement executes a block.
 * If a value is thrown and the try statement has one or more catch
 * clauses that can catch it, then control will be transferred to the
 * first such catch clause.
 * If the try statement has a finally clause, then another block of code
 * is executed, no matter whether the try block completes normally or abruptly,
 * and no matter whether a catch clause is first given control.
 *
 * In this implementation, the Try Statement is split into a Try-Catch
 * Statement and a Try-Finally Statement. A Try Statement where both catch
 * and finally clauses are present is rewritten as a Try-Catch Statement
 * enclosed in a Try-Finally Statement.
 */
public class JTryCatchStatement
    extends JStatement
{
    /**
     * The try clause.
     */
    private JBlock tryClause;

    /**
     * All catch clauses in this try/catch statement.
     */
    private JCatchClause[] catchClauses;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param tryClause
     * The try body.
     * @param catchClauses
     * An array of catch clauses.
     */
    public JTryCatchStatement(TokenReference where,
                              JBlock tryClause,
                              JCatchClause[] catchClauses,
                              JavaStyleComment[] comments)
    {
        super(where, comments);

        this.tryClause = tryClause;
        this.catchClauses = catchClauses;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context for the check.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        /* Context in which the try clause is checked.
         * It is a direct clone of the context before the try statement.
         */
        CTryContext tryContext = new CTryContext(context.cloneContext());
        /* BodyContext for try clause. */
        CBodyContext tryInside = new CBodyContext(tryContext);

        /* Check the try clause. */
        tryClause.checkStatement(tryInside);
        /* Close inside context, no further checks are done with it. */
        tryInside.close(getTokenReference());

        /* Now make sure that the interface of each catch clause is exported */
        for (int i = 0; i < catchClauses.length; i++) {
            catchClauses[i].checkInterface(context);
        }
        
        checkCatchUse(context, tryContext);
        checkCatchReachability();

        /* Make a context for use when checking catch clauses. */
        CBodyContext catchContext = (CBodyContext)context.cloneContext();
        /* Adopt local variable and field assignment info (but not throwables)
         * from try context.
         */
        catchContext.adopt(tryInside);
        boolean catchCompletesOK = checkCatchBodies(context, catchContext);
        context.setReachable(tryContext.isReachable() | catchCompletesOK);
    }

    /**
     * Check the bodies of the catch clauses.
     * 
     * @param statementContext
     * The context in which the whole try/catch statement occurs.
     * @param catchContext
     * The context in which the catch clauses should be checked.
     * @return
     * <code>false</code> if none of the catch clauses complete normally,
     * <code>true</code> otherwise.
     * @exception PositionedError
     * If the check fails.
     */
    private boolean checkCatchBodies(CBodyContext statementContext,
                                     CBodyContext catchContext)
        throws PositionedError
    {
        CBodyContext accumulator = null;
        boolean normalCompletion = false;

        for (int i = 0; i < catchClauses.length; i++) {
            CBodyContext previous = accumulator;
            accumulator = catchContext.cloneContext();
            catchClauses[i].checkBody(accumulator);
            if (accumulator.isReachable()) {
                /* If at least one catch clause can complete normally,
                 * then all catch clauses combined can.
                 */
                normalCompletion = true;
            }

            if (previous != null) {
                accumulator.merge(previous);
            }
        }
        
        /* Add all exceptions that can be thrown from catch clauses to
         * the context in which the whole try/catch statement is checked.
         */
        statementContext.addThrowables(accumulator.getThrowables());
        statementContext.adopt(accumulator);
        return normalCompletion;
    }

    /**
     * Check that all catch clauses are used.
     *
     * @param statementContext
     * The context in which the whole try/catch statement occurs.
     * @param catchContext
     * The context in which the try clause was checked.
     * @exception PositionedError
     * If the check fails.
     */
    private void checkCatchUse(CBodyContext statementContext,
                               CTryContext tryContext)
        throws PositionedError
    {
        Iterator throwables = tryContext.getThrowables().iterator();

        while (throwables.hasNext()) {
            CThrowableInfo throwableInfo = (CThrowableInfo)throwables.next();
            ClassOrInterfaceType throwableType = throwableInfo.getThrowable();
            boolean caught = false;

            for (int i = 0; i < catchClauses.length; i++) {
                ClassOrInterfaceType catchType = (ClassOrInterfaceType)catchClauses[i].getType();

                
                /* Check if the catch clause catches a subclass of an
                 * exception. This is allowed (like catching
                 * FileNotFoundException when a method-call in the try block
                 * throws IOException) but does not mean that the whole
                 * IOException is hereby contained in the try/catch statement.
                 */
                if ( throwableInfo.isSubTypePossible() &&
                     !catchType.equals(throwableType) &&
                     catchType.allowsAssignmentConversion(throwableType) ) {

                    catchClauses[i].setUsed(true);
                    /* Do not mark 'exception caught' and continue looking
                     * for catch handlers for this type.
                     */
                }
                else if ( throwableType.allowsAssignmentConversion(catchType) ) {
                    catchClauses[i].setUsed(true);
                    /* Mark 'exception caught' */
                    caught = true;
                    break;
                }
            }

            /* No catch clauses catches this type of exception. */
            if (!caught) {
                /* Add this type of exception to the context in which the
                 * whole try/catch statement is checked.
                 */
                statementContext.addThrowable(throwableInfo);
            }
        }

        for (int i = 0; i < catchClauses.length; i++) {
            ClassOrInterfaceType catchType = (ClassOrInterfaceType)catchClauses[i].getType();
            if ( !catchClauses[i].isUsed() &&
                 // catchType.isCheckedException() &&
                 !catchType.allowsAssignmentConversion(StandardType.ERROR) &&
                 !catchType.equals(StandardType.THROWABLE) ) {
                 // !catchType.equals(StandardType.EXCEPTION) ) {

                tryContext.reportTrouble
                    (new CBlockError(catchClauses[i].getTokenReference(),
                                     KjcMessages.STATEMENT_UNREACHABLE)); 
            }
        }
    }

    /**
     * Check that there are no catch clauses later in list of catch clauses
     * that cannot be reached because an earlier catch clause catches the same
     * type of throwable or a supertype of the throwable.
     *
     * @exception PositionedError
     * If the check fails.
     */ 
    private void checkCatchReachability()
        throws PositionedError
    {
        if (tryClause instanceof JBlock) {
            JBlock tryBlock = (JBlock)tryClause;
            
            if (tryBlock.getBody().length == 0) {
                /* A completely empty try block can throw no exceptions.
                 * However, even a single bytecode could be the target of
                 * a ThreadDeath.
                     */
                throw new CBlockError(catchClauses[0].getTokenReference(),
                                      KjcMessages.STATEMENT_UNREACHABLE);
            }
        }
        
        for (int i = 0; i < catchClauses.length; i++) {
            Type typeI = catchClauses[i].getType();

            for (int j = i + 1; j < catchClauses.length; j++) {
                if (catchClauses[j].getType().allowsAssignmentConversion(typeI)) {

                    throw new CBlockError(catchClauses[j].getTokenReference(),
                                          KjcMessages.STATEMENT_UNREACHABLE);
                }
            }
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p  the visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitTryCatchStatement(this, tryClause, catchClauses);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        CodeLabel  nextLabel = new CodeLabel();

        int  startPC = code.getPC();
        tryClause.genCode(code);   //  TRY CODE
        int  endPC = code.getPC();
        code.plantLabelRef(opc_goto, nextLabel); //  GOTO next

        for (int i = 0; i < catchClauses.length; i++) {
            catchClauses[i].genCode(code);
            catchClauses[i].genExceptionHandler(code, startPC, endPC);
            code.plantLabelRef(opc_goto, nextLabel); //  GOTO next
        }

        code.plantLabel(nextLabel);   // next: ...
    }
}
