/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import at.dms.compiler.Compiler;
import at.dms.compiler.CWarning;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;
import at.dms.util.MessageDescription;

/**
 * The base class for representations of a java class or interface declaration
 * in the syntax tree.
 */
public abstract class JTypeDeclaration
    extends JMemberDeclaration
{
    /**
     * Access and other modifiers of this type declaration.
     */
    private int modifiers;

    /**
     * Name of type.
     */
    private String ident;

    /**
     * Interfaces implemented by this type.
     */    
    private ClassOrInterfaceType[] interfaces;

    /**
     * Method-less bodies inside the class file (instance or static
     * initializers in the source code; in the classfile, only one of
     * each is allowed, so they are combined)
     */
    private List body = new ArrayList();

    /**
     * Method-less bodies generated by syntax tree converter tools.
     */
    private List generatedBody = new ArrayList();

    /**
     * The fields declared in the class or interface.
     */
    private List fields = new ArrayList();

    /**
     * Fields generated by syntax tree converter tools.
     */
    private List generatedFields = new ArrayList();

    /**
     * The methods declared in the class or interface.
     */
    private List methods = new ArrayList();

    /**
     * Methods generated by syntax tree converter tools.
     */
    private List generatedMethods = new ArrayList();

    /**
     * The member types declared in the class or interface.
     */
    private List inners = new ArrayList();

    /**
     * Member types generated by syntax tree converter tools.
     */
    private List generatedInners = new ArrayList();
    
    /**
     * The outer-this field in this type, if it is an inner type.
     */
    private JFieldDeclaration outerThis;
    
    /**
     * The default constructor, if no constructors were declared in the source
     * code.
     */
    private JConstructorDeclaration defaultConstructor;
    
    /**
     * The static initializer.
     */
    private JInitializerDeclaration staticInit;
    
    /**
     * The instance initializer.
     */
    private JInitializerDeclaration instanceInit;

    /**
     * Class export.   
     */
    private CSourceClass sourceClass;

    /**
     * Whether the source class is unique.
     */
    private boolean uniqueSourceClass = true;

    /**
     * Constructs a type declaration in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param modifiers
     * The list of modifiers of this class.
     * @param ident
     * The short name of this class.
     * @param interfaces
     * The names of this types's interfaces.
     * @param fields
     * The fields in this class or interface.
     * @param methods
     * The methods in this class or interface.
     * @param inners
     * The member types declared in this class or interface.
     * @param initializers
     * The static and instance initializers in this class or interface.
     * @param javadoc
     * The javadoc for this class or interface.
     * @param comments
     * The implementation comments for this class or interface.
     */
    public JTypeDeclaration(TokenReference where,
                            int modifiers,
                            String ident,
                            ClassOrInterfaceType[] interfaces,
                            JFieldDeclaration[] fields,
                            JMethodDeclaration[] methods,
                            JTypeDeclaration[] inners,
                            JPhylum[] initializers,
                            JavadocComment javadoc,
                            JavaStyleComment[] comments)
    {
        super(where, javadoc, comments);
        
        this.modifiers = modifiers;
        this.ident = ident.intern();
        this.interfaces = interfaces;
        this.fields.addAll(Arrays.asList(fields));
        this.methods.addAll(Arrays.asList(methods));
        this.inners.addAll(Arrays.asList(inners));
        this.body.addAll(Arrays.asList(initializers));
    }

    /**
     * Defines an intermediate external representation of this class to use
     * internally.
     */
    public void generateInterface(Compiler compiler,
                                  CClass owner,
                                  String prefix)
    {
        // Don't do it twice.
        if (sourceClass == null) {
            sourceClass = new CSourceClass(compiler, owner, this,
                                           getTokenReference(),
                                           modifiers,
                                           ident,
                                           prefix + ident, isDeprecated());
            sourceClass.setSynthetic(isSynthetic());
        }

        setInterface(sourceClass);

        ClassOrInterfaceType[] innerClasses =
            new ClassOrInterfaceType[inners.size()];
        Iterator iterator = inners.iterator();
        int index = 0;
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            inner.generateInterface(compiler, sourceClass,
                                    sourceClass.getQualifiedName()
                                    .concat("$"));
            innerClasses[index++] = inner.getCClass().getType();
        }
        
        sourceClass.setInnerClasses(innerClasses); // prevent interface
        this.uniqueSourceClass = CTopLevel.addSourceClass(sourceClass);
    }

    /**
     * Easy way to report trouble.
     */
    protected void reportTrouble(CContext context,
                                 MessageDescription message,
                                 String param)
    {
        context.reportTrouble(new PositionedError(getTokenReference(),
                                                  message, param));
    }

    /**
     * Returns true if this class is at top level, false if not. Possible
     * reasons for not being at top level: being an inner class or interface.
     */
    public boolean isAtTopLevel()
    {
        return getCClass().getOwner() == null;
    }
   
    /**
     * Returns source-declared fields as list.
     */
    public List getFieldList()
    {
        return fields;
    }

    /**
     * Returns generated fields as list.
     */
    public List getGeneratedFieldList()
    {
        return generatedFields;
    }

    /**
     * Returns fields as array.
     */
    public JFieldDeclaration[] getFields()
    {
        return (JFieldDeclaration[])
            fields.toArray(new JFieldDeclaration[fields.size()]);
    }
    
    /**
     * Returns source-declared methods as list.
     */
    public List getMethodList()
    {
        return methods;
    }

    /**
     * Returns generated methods as list.
     */
    public List getGeneratedMethodList()
    {
        return generatedMethods;
    }

    public JMethodDeclaration[] getMethods()
    {
        return (JMethodDeclaration[])
            methods.toArray(new JMethodDeclaration[methods.size()]);
    }
    
    /**
     * Returns bodies as list.
     */
    public List getBodyList()
    {
        return body;
    }

    /**
     * Returns bodies as array. (Bodies are initializers for fields and
     * static/instance initializers. They are combined to a single instance
     * and/or static initializer)
     */
    public JPhylum[] getBody()
    {
        return (JPhylum[])body.toArray(new JPhylum[body.size()]);
    }
    
    /**
     * Returns source-declared inner classes as list.
     */
    public List getInnerList()
    {
        return inners;
    }

    /**
     * Returns generated inner classes as list.
     */
    public List getGeneratedInnerList()
    {
        return generatedInners;
    }

    /**
     * Returns inner classes as array.
     */
    public JTypeDeclaration[] getInners()
    {
        return (JTypeDeclaration[])
            inners.toArray(new JTypeDeclaration[inners.size()]);
    }

    /**
     * Sets the interfaces implemented by this type.
     */
    public void setInterfaces(ClassOrInterfaceType[] interfaces)
    {
        this.interfaces = interfaces;
    }

   
    /**
     * Returns the interfaces implemented by this type.
     */
    public ClassOrInterfaceType[] getInterfaces()
    {
        return interfaces;
    }

    /**
     * Returns the default constructor for this type.
     */
    public JConstructorDeclaration getDefaultConstructor()
    {
        return defaultConstructor;
    }
    
    /**
     * Sets the default constructor for this type.
     */
    public void setDefaultConstructor(JConstructorDeclaration constructor)
    {
        this.defaultConstructor = constructor;
    }

    /**
     * Sets the identifier for the type (the name).
     */
    public void setIdent(String ident)
    {
        this.ident = ident;
    }

    /**
     * Returns the identifier for the type (the name).
     */
    public String getIdent()
    {
        return this.ident;
    }

    /**
     * Sets the static initializer.
     */
    public void setStaticInitializer(JInitializerDeclaration initializer)
    {
        this.staticInit = initializer;
    }

    /**
     * Returns static initializer as syntax-subtree.
     */
    public JInitializerDeclaration getStaticInitializer()
    {
        return this.staticInit;
    }

    /**
     * Sets the instance initializer.
     */
    public void setInstanceInitializer(JInitializerDeclaration initializer)
    {
        this.instanceInit = initializer;
    }

    /**
     * Returns instance initializer as syntax-subtree.
     */
    public JInitializerDeclaration getInstanceInitializer()
    {
        return this.instanceInit;
    }

    /**
     * Returns export of type.
     */
    public CSourceClass getSourceClass()
    {
        return sourceClass;
    }

    /**
     * Returns modifiers of type.
     */
    public int getModifiers()
    {
        return modifiers;
    }

    /**
     * Changes modifiers of type. Be careful when to do this; it has no
     * effect after the export of the type has been generated.
     */
    public void setModifiers(int modifiers)
    {
        this.modifiers = modifiers;
    }

    /**
     * Returns owner of type.
     */
    public CClass getOwner()
    {
        return getCClass().getOwner();
    }

    /**
     * Second pass (quick), check interface looks good
     * @exception PositionedError an error with reference to the source file
     */
    public abstract void checkInterface(CContext context)
        throws PositionedError;

    /**
     * Second pass (quick), check interface looks good.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    protected void checkInterface(CContext context,
                                  ClassOrInterfaceType superClass)
        throws PositionedError
    {
        CClassContext self = new CClassContext(context, sourceClass, this);
        CSourceField[] fieldList;
        CMethod[] methodList;
        
        if (!uniqueSourceClass) {
            throw new UnpositionedError(KjcMessages.TOP_RENAME,
                                        sourceClass.getQualifiedName())
                .addPosition(getTokenReference());
        }

        // 8.1.2 Class Modifiers
        check(context,
              !isAtTopLevel()
              || !hasFlag(getCClass().getModifiers(),
                          ACC_PROTECTED + ACC_PRIVATE),
              KjcMessages.CLASS_FLAGS);

        check(context,
              !hasFlag(getCClass().getModifiers(),
                       ACC_SYNCHRONIZED + ACC_VOLATILE
                       + ACC_TRANSIENT + ACC_NATIVE),
              KjcMessages.CLASS_FLAGS);

        if (isAtTopLevel() && sourceClass.isPublic()
            && !getTokenReference().getName().startsWith(ident + ".")) {

            context.reportTrouble(new
                CWarning(getTokenReference(),
                         KjcMessages.CLASS_NAME_FILENAME,
                         ident, getTokenReference().getName()));
        }

        // resolve interfaces
        try {
            for (int i = 0; i < interfaces.length; i++) {
                interfaces[i].resolveType(context);
            }
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
        
        // Add fields of this class
        int generatedFields = getCClass().hasOuterThis() ? 1 : 0;

        fieldList = new CSourceField[fields.size() + generatedFields];
        int fieldCount = 0;

        Iterator iterator = fields.iterator();
        while (iterator.hasNext()) {
            JFieldDeclaration declaration = (JFieldDeclaration)iterator.next();
            CSourceField field = declaration.checkInterface(self);
            
            field.setPosition(fieldCount);
            fieldList[fieldCount] = field;

            for (int j = 0; j < fieldCount; j++) {
                check(context,
                      !fieldList[j].getIdent().equals(field.getIdent()),
                      KjcMessages.FIELD_RENAME, field.getIdent());
            }
            fieldCount++;
        }

        if (getCClass().hasOuterThis()) {
            createOuterThisField();
            CSourceField field = outerThis.checkInterface(self);
            
            field.setPosition(fieldCount);
            fieldList[fieldCount] = field;

            for (int j = 0; j < fieldCount; j++) {
                check(context,
                      !fieldList[j].getIdent().equals(JAV_OUTER_THIS),
                      KjcMessages.FIELD_RENAME, JAV_OUTER_THIS);
            }
            fieldCount++;
        }
        
        // Add methods of this class
        int generatedMethods = 0;
        if (defaultConstructor != null) {
            generatedMethods++;
        }
        if (staticInit != null && !staticInit.isDummy()) {
            generatedMethods++;
        }
        if (instanceInit != null && !instanceInit.isDummy()) {
            generatedMethods++;
        }
        
        methodList = new CMethod[methods.size() + generatedMethods];
        
        int methodCount = 0;
        iterator = methods.iterator();
        while (iterator.hasNext()) {
            JMethodDeclaration method = (JMethodDeclaration)iterator.next();
            CMethod selfM = method.checkInterface(self);
            methodList[methodCount] = selfM;

            for (int j = 0; j < methodCount; j++) {
                check(context,
                      !methodList[j].equals(selfM),
                      KjcMessages.METHOD_REDEFINE,
                      method.getMethod().getIdent());
            }
            methodCount++;
        }

        if (defaultConstructor != null) {
            methodList[methodCount++] =
                defaultConstructor.checkInterface(self);
        }
        if (staticInit != null) {
            if (!staticInit.isDummy()) {
                methodList[methodCount++] = staticInit.checkInterface(self);
            }
            else {
                staticInit.checkInterface(self);
            }
        }
        if (instanceInit != null) {
            if (!instanceInit.isDummy()) {
                methodList[methodCount++] = instanceInit.checkInterface(self);
            }
            else {
                instanceInit.checkInterface(self);
            }
        }
        
        // Check inners
        iterator = inners.iterator();
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            inner.checkInterface(self);
        }
        
        sourceClass.close(this.interfaces, superClass, fieldList, methodList);

        if (sourceClass.hasOuterThis()) {
            addOuterThis();
        }
    }
    
    /**
     * Checks that same interface is not specified more than once.
     *
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public void checkInitializers(CContext context)
        throws PositionedError
    {
        for (int i = 0; i < interfaces.length; i++) {
            for (int j = 0; j < i; j++) {
                check(context,
                      !interfaces[i].equals(interfaces[j]),
                      KjcMessages.INTERFACES_DUPLICATE, ident, interfaces[i]);
            }
        }

        // Checks that specified interfaces are not inherited
        /* This check is not mandated by the JLS and is not really
         * warning-worth; it may occur quite frequently.
        for (int i = 0; i < interfaces.length; i++) {
            CClass parent = sourceClass.getSuperClass();
            CClass interfaceCClass = interfaces[i].getCClass();

            if (parent != null && parent.descendsFrom(interfaceCClass)) {

                context.reportTrouble
                    (new CWarning
                        (getTokenReference(),
                         KjcMessages.INTERFACE_IMPLEMENTED_BY_SUPERCLASS,
                         interfaces[i], parent.getIdent()));
            }


            for (int j = 0; j < interfaces.length; j++) {
                if (j == i) {
                    continue;
                }
                if (interfaces[j].getCClass().descendsFrom(interfaceCClass)) {
                    context.reportTrouble
                        (new CWarning
                            (getTokenReference(),
                             KjcMessages.INTERFACE_IMPLEMENTED_BY_SUPERCLASS,
                             interfaces[i], interfaces[j]));
                }
            }
        }
         */
        sourceClass.setInitializerChecked();
    }

    /**
     * Any inner class (non-static member class) has an outer-this variable.
     * So a call to this method isn't necessary; calling it when it makes
     * no sense (on a static member class, for example) results in a runtime
     * assertion exception.
     */
    public void addOuterThis()
    {
        /* Assertion; will throw runtime exception if class cannot have
         * outer this.
         */
        sourceClass.setHasOuterThis(true);
    }

    /**
     * Does the chores of creating an outer this field.
     */
    private void createOuterThisField()
    {
        if (outerThis == null) {
            ClassOrInterfaceType ownerType = getOwner().getCClass().getType();
            JVariableDefinition outerThisVariable;
            outerThisVariable = new JVariableDefinition(getTokenReference(),
                                                        ACC_PRIVATE,
                                                        ownerType,
                                                        JAV_OUTER_THIS,
                                                        null);
            outerThis = new JFieldDeclaration(getTokenReference(),
                                              outerThisVariable, null, null);
            outerThis.setSynthetic(true);
        }
    }

    /**
     * createContext.
     *
     * @param parent
     * The parent context or null.
     * @return
     * A CClassContext that represents this context.
     * @exception UnpositionedError
     * Error caught as soon as possible.
     */
    public abstract CClassContext createContext(CContext parent)
        throws UnpositionedError;

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node
     * @exception PositionedError
     * Error caught as soon as possible (for subclasses)
     */
    public void checkTypeBody(CContext context)
        throws PositionedError
    {
        context.addSourceClass(sourceClass);
    }

    /**
     * Adds a generated inner method. The keepSeperate flag indicates whether
     * it should be stored in a seperate list or if it should be added to
     * the list of source-declared inner classes. The latter is handy for
     * source generators.
     */
    public void addGeneratedInner(JTypeDeclaration inner,
                                  boolean keepSeperate)
    {
        if (keepSeperate) {
            generatedInners.add(inner);
        }
        else {
            inners.add(inner);
        }

        inner.generateInterface(sourceClass.getCompiler(),
                                sourceClass,
                                sourceClass.getQualifiedName().concat("$"));
        sourceClass.addInnerClass(inner.getCClass().getType());
    }

    /**
     * ...
     */
    public void addGeneratedField(JFieldDeclaration field,
                                  boolean keepSeperate)
    {
        if (keepSeperate) {
            throw new RuntimeException("Not yet implemented");
            /*
            generatedFields.add(field);
            generatedBody.add(field);
            */
        }
        else {
            fields.add(field);
            body.add(field);
        }
    }

    /**
     * ...
     */
    public void addGeneratedMethod(JMethodDeclaration method,
                                   boolean keepSeperate)
    {
        if (keepSeperate) {
            generatedMethods.add(method);
        }
        else {
            methods.add(method);
        }
    }

    /**
     * ...
     */
    public void addGeneratedBodyPart(JPhylum bodyPart,
                                     boolean keepSeperate)
    {
        if (keepSeperate) {
            throw new RuntimeException("Not yet implemented");
            // generatedBody.add(bodyPart);
        }
        else {
            body.add(bodyPart);
        }
    }
}
