/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.type.*;

/**
 * Local type declaration statement. See JLS 14.3.
 *
 * A local type declaration declaration statement declares one type
 * declaration in a body of a method.
 */
public class JTypeDeclarationStatement
    extends JStatement
{
    /**
     * Type declaration inside a method.
     */
    private JTypeDeclaration declaration;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param declaration
     * The type declaration in this statement.
     */
    public JTypeDeclarationStatement(TokenReference where,
                                     JTypeDeclaration declaration)
    {
        super(where, null);
        this.declaration = declaration;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        CClass owner = context.getClassContext().getCClass();
        String prefix =
            owner.getQualifiedName() + "$"
            + context.getClassContext().getNextSyntheticIndex() + "$";

        /* If inside a static method, then the declared type is also static */
        if (context.getMethodContext().getCMethod().isStatic()) {
            declaration.setModifiers(declaration.getModifiers() | ACC_STATIC);
        }

        declaration.generateInterface(context.getCompiler(),
                                      owner, prefix);

        CClass declaredClass = declaration.getCClass();
        owner.addInnerClass(declaredClass.getType());
        try {
            context.getBlockContext().addClass(declaredClass);
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }

        declaration.checkInterface(context);
        declaration.checkInitializers(context);
        declaration.checkTypeBody(context);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitTypeDeclarationStatement(this, declaration);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        // nothing to do here
    }
}
