/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.type.*;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Representation of unary expressions: the PLUS, MINUS, LOGICAL NOT and
 * BINARY NOT expressions: +expression, -expression, !expression and
 * ~expression.
 *
 * @author DMS, Erwin Bolwidt
 */
public strictfp class JUnaryExpression
    extends JExpression
{
    /**
     * Type of operation.
     */
    protected int oper;

    /**
     * Expression to operate on.
     */
    protected JExpression expr;

    /**
     * Type of result.
     */
    protected Type type;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param oper
     * The operator.
     * @param expr
     * The operand.
     */
    public JUnaryExpression(TokenReference where,
                            int oper,
                            JExpression expr)
    {
        super(where);
        this.oper = oper;
        this.expr = expr;
    }

    /**
     * Compute the type of this expression. Call after parsing.
     *
     * @return the type of this expression.
     */
    public Type getType()
    {
        return type;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if ( (expr instanceof JOrdinalLiteral) &&
             oper == OPE_MINUS &&
             ((JOrdinalLiteral)expr).specialUnaryMinus() ) {

            oper = OPE_PLUS;
        }
        else {
            expr = expr.checkExpression(context);
        }

        switch (oper) {
        case OPE_PLUS:
        case OPE_MINUS:
            check(context, expr.getType().isNumeric(),
                  KjcMessages.UNARY_BADTYPE_PM, expr.getType());
            type = expr.getType().toNumeric().unaryPromote();

            if (expr.isConstant()) {
                if (oper == OPE_PLUS) {
                    return expr;
                }

                JNumberLiteral lit = expr.getNumberLiteral();
                if (type == StandardType.DOUBLE) {
                    lit = new JRealLiteral
                        (getTokenReference(),
                         new Double(-lit.getNumberValue().doubleValue()),
                         StandardType.DOUBLE);
                }
                else if (type == StandardType.FLOAT) {
                    lit = new JRealLiteral
                        (getTokenReference(),
                         new Float(-lit.getNumberValue().floatValue()),
                         StandardType.FLOAT);
                }
                else if (type == StandardType.LONG) {
                    lit = new JOrdinalLiteral
                        (getTokenReference(),
                         new Long(-lit.getNumberValue().longValue()),
                         StandardType.LONG);
                }
                else {
                    lit = new JOrdinalLiteral
                        (getTokenReference(),
                         new Integer(-lit.getNumberValue().intValue()),
                         StandardType.INTEGER);
                }

                return lit;
            }

            break;

        case OPE_BNOT:
            check(context, expr.getType().isIntegral(),
                  KjcMessages.UNARY_BADTYPE_BNOT, expr.getType());
            type = expr.getType().toNumeric().unaryPromote();

            if (expr.isConstant()) {
                JOrdinalLiteral lit = expr.getOrdinalLiteral();
                if (type == StandardType.LONG) {
                    lit = new JOrdinalLiteral
                        (getTokenReference(),
                         new Long(~lit.getNumberValue().longValue()),
                         StandardType.LONG);
                }
                else {
                    lit = new JOrdinalLiteral
                        (getTokenReference(),
                         new Integer(~lit.getNumberValue().intValue()),
                         StandardType.INTEGER);
                }
                return lit;
            }
            break;

        case OPE_LNOT:
            check(context, expr.getType() == StandardType.BOOLEAN,
                  KjcMessages.UNARY_BADTYPE_LNOT, expr.getType());
            type = StandardType.BOOLEAN;
            if (expr.isConstant()) {
                return new JBooleanLiteral
                    (getTokenReference(),
                     !expr.getBooleanLiteral().booleanValue());
            }
            break;

        default:
            fail();
        }

        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitUnaryExpression(this, oper, expr);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        switch (oper) {
        case OPE_PLUS:
            expr.genCode(code);
            break;

        case OPE_MINUS:
            expr.genCode(code);
            switch (type.getTypeID()) {
            case TID_FLOAT:
                code.plantInstruction(new NoArgInstruction(opc_fneg));
                break;
            case TID_LONG:
                code.plantInstruction(new NoArgInstruction(opc_lneg));
                break;
            case TID_DOUBLE:
                code.plantInstruction(new NoArgInstruction(opc_dneg));
                break;
            default:
                code.plantInstruction(new NoArgInstruction(opc_ineg));
            }
            break;

        case OPE_LNOT:
            expr.genCode(code);
            code.plantInstruction(new PushLiteralInstruction(1));
            code.plantInstruction(new NoArgInstruction(opc_ixor));
            break;

        case OPE_BNOT:
            expr.genCode(code);
            if (type == StandardType.LONG) {
                code.plantInstruction(new PushLiteralInstruction((long)-1));
                code.plantInstruction(new NoArgInstruction(opc_lxor));
            }
            else {
                code.plantInstruction(new PushLiteralInstruction((int)-1));
                code.plantInstruction(new NoArgInstruction(opc_ixor));
            }
            break;

        default:
            fail();
        }
    }

    /**
     * Generates a sequence of bytecodes to branch on a label.
     * This method helps to handle heavy optimizables conditions.
     *
     * @param code
     * The code list.
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label)
    {
        assert(oper == OPE_LNOT);

        expr.genBranch(!cond, code, label);
    }
}
