/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import at.dms.classfile.ClassFileFormatException;
import at.dms.classfile.SourceDescription;
import at.dms.compiler.Compiler;
import at.dms.compiler.CompilerMessages;
import at.dms.compiler.CWarning;
import at.dms.compiler.InputBuffer;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.WarningFilter;
import at.dms.type.*;

/**
 * The entry point of the compiler, containing the <code>main</code> method.
 */
public class Main
    extends Compiler
{
    /**
     * Files to compiler.
     */
    protected LinkedList sourceFiles = new LinkedList();

    /**
     * Was there at least one error.
     */
    protected boolean errorFound;

    /**
     * Command-line options.
     */
    protected KjcOptions options;

    /**
     * Warning filter.
     */
    private WarningFilter filter = new DefaultFilter();

    /**
     * All generated classes.
     */
    private List classes = new ArrayList(100);

    /**
     * The compilation units that are compiled in this session.
     */
    private List compilationUnits;

    /**
     * The compilation units that are loaded from the classpath during
     * checks of another compilation unit.
     */
    private LinkedList sourcePathInterfaceChecked = new LinkedList();

    /**
     * ...
     */
    private LinkedList sourcePathInitializersChecked = new LinkedList();

    /**
     * Runs a compilation session.
     *
     * @param args
     * The command line arguments.
     */
    public boolean run(String[] args)
    {
        if (!parseArguments(args)) {
            return false;
        }

        setDebug(options.debug);

        if (options.Xstdout) {
            System.setErr(System.out);
            setErr(System.out);
        }

        initialize();

        if (sourceFiles.size() == 0) {
            options.usage();
            inform(KjcMessages.NO_INPUT_FILES);
            return false;
        }

        if (verboseMode()) {
            inform(CompilerMessages.COMPILATION_STARTED,
                   new Integer(sourceFiles.size()));
        }

        try {
            List verifiedFiles = verifyFiles(sourceFiles);
            sourceFiles.clear();
            sourceFiles.addAll(verifiedFiles);
        }
        catch (UnpositionedError e) {
            inform(e);
            return false;
        }

        options.destination = checkDestination(options.destination);

        CTopLevel.initSession(this);

        compilationUnits = new ArrayList(sourceFiles.size());

        /* Don't use an Iterator, since the list of soure files may
         * be manipulated by the method that loads source files from the
         * class path, and the standard iterators don't stand that kind
         * of concurrent modification well.
         */
        while (!sourceFiles.isEmpty()) {
            File file = (File)sourceFiles.removeFirst();
            // Remove it from the list.
            JCompilationUnit unit = parseFile(file);
            if (options.debug) {
                if (unit == null) {
                    System.err.println("Parsefile failed for: " + file);
                }
            }
            if (unit != null) {
                compilationUnits.add(unit);
            }
        }

        sourceFiles = null;

        if (errorFound) {
            return false;
        }

        if (!options.beautify) {
            if (!doChecks()) {
                return false;
            }

            if (!options.java && !options.beautify) {
                compilationUnits.clear();
            }
        }

        if (!options.nowrite) {
            if ( options.java || options.beautify ) {
                Iterator units = compilationUnits.iterator();
                while (units.hasNext()) {
                    JCompilationUnit unit = (JCompilationUnit)units.next();
                    unit.accept(this, options.destination);
                }
            }
            else {
                genCode(options.destination);
            }
        }

        if (verboseMode()) {
            inform(CompilerMessages.COMPILATION_ENDED);
        }

        CodeSequence.endSession();

        return true;
    }

    /**
     * ...
     */
    private boolean doChecks()
    {
        long lastTime = System.currentTimeMillis();
        Iterator units = compilationUnits.iterator();
        while (units.hasNext()) {
            JCompilationUnit unit = (JCompilationUnit)units.next();
            checkInterface(unit);
        }


        if (verboseMode()) {
            inform(CompilerMessages.INTERFACES_CHECKED,
                   new Long(System.currentTimeMillis() - lastTime));
        }
        
        if (errorFound) {
            return false;
        }
        
        units = compilationUnits.iterator();
        while (units.hasNext()) {
            JCompilationUnit unit = (JCompilationUnit)units.next();
            checkInitializers(unit);
        }
        checkSourcePathInitializers();
        
        if (errorFound) {
            return false;
        }
       
        units = compilationUnits.iterator();
        while (units.hasNext()) {
            JCompilationUnit unit = (JCompilationUnit)units.next();
            checkBody(unit);
        }
        checkSourcePathBody();
        
        if (errorFound) {
            return false;
        }

        return true;
    }

    /**
     * Parse the argument list.
     */
    public boolean parseArguments(String[] args)
    {
        options = new KjcOptions();
        return options.parseCommandLine(args, sourceFiles);
    }

    /**
     * Generates the code from an array of compilation unit and
     * a destination
     *
     * @param destination
     * The directory where to write classfiles.
     */
    public void genCode(String destination)
    {
        CSourceClass[] classes = getClasses();

        this.classes.clear();

        int count = 0;
        try {
            for (count = 0; count < classes.length; count++) {
                long lastTime = System.currentTimeMillis();
                
                classes[count].genCode(destination);
                if (verboseMode() && !classes[count].isInnerClass()) {
                    inform(CompilerMessages.CLASSFILE_GENERATED,
                           classes[count].getQualifiedName().replace('/', '.'),
                           new Long(System.currentTimeMillis()-lastTime));
                }
                classes[count] = null;
            }
        }
        catch (ClassFileFormatException e) {
            if (options.debug) {
                System.err.println("count = " + count);
                e.printStackTrace();
            }
            inform(e);
        }
        catch (IOException e) {
            inform(e);
        }
    }

    /**
     * Initialize the compiler (read classpath, check classes.zip)
     */
    protected void initialize()
    {
        if (options.classpath == null) {
            options.classpath = System.getProperty("java.class.path");
            if (options.classpath == null) {
                options.classpath = ".";
            }
        }
        /* Initialize classpath.
         * Take note of the special nosystemlib option that allows you to
         * compile alternate system libraries like classpath and libgcj.
         */
        at.dms.classfile.ClassPath.init(options.classpath,
                                        !options.nosystemlib);
    }

    /**
     * @return true if compilation in verbose mode
     */
    public boolean verboseMode()
    {
        return options.verbose;
    }

    /**
     * @return the optimization level
     */
    public int optimizeCode()
    {
        return options.optimize;
    }

    /**
     * Shows an error/warning and set error flag to true
     */
    public KjcPrettyPrinter getPrettyPrinter(String fileName)
    {
        return new KjcPrettyPrinter(fileName);
    }

    /**
     * Entry point.
     *
     * @param args
     * The command line arguments.
     */
    public static void main(String[] args)
    {
        String repeatProperty = System.getProperty("repeat");
        boolean success;

        if (repeatProperty != null) {
            int repeat = Integer.parseInt(repeatProperty);
       
            success = false;
            while (repeat > 0) {
                long start = System.currentTimeMillis();
                success = compile(args);
                long duration = System.currentTimeMillis() - start;
                System.err.println("Duration " + repeat + ": " + duration
                                   + " ms.");
                repeat--;
            }
        }
        else {
            success = compile(args);
        }

        System.exit(success ? 0 : 1);
    }

    /**
     * Second entry point.
     */
    public static boolean compile(String[] args)
    {
        return new Main().run(args);
    }

    /**
     * ...
     */
    private void checkSourcePathInitializers()
    {
        while (sourcePathInterfaceChecked.size() != 0) {
            JCompilationUnit unit =
                (JCompilationUnit)sourcePathInterfaceChecked.removeFirst();
            checkInitializers(unit);
            sourcePathInitializersChecked.add(unit);
        }
    }
    /**
     * ...
     */
    private void checkSourcePathBody()
    {
        do {
            checkSourcePathInitializers();
            while (sourcePathInitializersChecked.size() != 0) {
                JCompilationUnit unit = (JCompilationUnit)
                    sourcePathInitializersChecked.removeFirst();
                checkBody(unit);
            }
            checkSourcePathInitializers();
        } while (sourcePathInterfaceChecked.size() != 0);
    }
        
    /**
     * ...
     */
    public void addSourceClass(CSourceClass clazz)
    {
        classes.add(clazz);
    }

    /**
     * Loads a new source file.
     */
    public void loadSourceFile(SourceDescription sourceFile)
    {
        if (options.debug) {
            System.err.println("Loading source file from classpath: " 
                               + sourceFile.getName());
        }

        try {
            JCompilationUnit unit = parseFile(sourceFile);
            sourceFile.getInput().close();

            if (unit != null) {
                checkInterface(unit);
                sourcePathInterfaceChecked.add(unit);
            }
        }
        catch (IOException e) {
            inform(e);
        }
    }

    /**
     * Parses the given file and return a compilation unit.
     * side effect: increment error number.
     *
     * @param file
     * The name of the file.
     * @return
     * The compilation unit defined by this file.
     */
    protected JCompilationUnit parseFile(File file)
    {
        InputStream fileInput = null;
        try {
            fileInput = new FileInputStream(file);
            return parseFile(new SourceDescription(file.getPath(),
                                                   file.length(),
                                                   true, fileInput));
                             
        }
        catch (IOException e) {
            inform(e);
            return null;
        }
        finally {
            if (fileInput != null) {
                try {
                    fileInput.close();
                }
                catch (IOException e) {}
            }
        }
    }

    /**
     * Parses the given file and return a compilation unit.
     * side effect: increment error number.
     *
     * @param file
     * The name of the file.
     * @return
     * The compilation unit defined by this file.
     */
    protected JCompilationUnit parseFile(SourceDescription sourceFile)
    {
        InputBuffer buffer;

        try {
            buffer = new InputBuffer(sourceFile, options.encoding);
        }
        catch (IOException e) {
            inform(e);
            errorFound = true;
            return null;
        }

        KjcScanner scanner = new KjcScanner(this, buffer);
        KjcParser parser = new KjcParser(this, scanner);
        long lastTime = System.currentTimeMillis();
        JCompilationUnit unit;

        try {
            unit = parser.jCompilationUnit();
        }
        catch (antlr.RecognitionException e) {
            if (options.debug) {
                e.printStackTrace();
            }
            inform(parser.beautifyParseError(e));
            errorFound = true;
            unit = null;
        }
        catch (Exception e) {
            if (options.debug) {
                e.printStackTrace();
            }
            errorFound = true;
            unit = null;
        }

        if (verboseMode()) {
            inform(CompilerMessages.FILE_PARSED,
                   sourceFile.getName(),
                   new Long(System.currentTimeMillis() - lastTime));
        }

        buffer.release();

        return unit;
    }

    /**
     * Checks that interface of a given compilation unit is correct.
     * side effect: increment error number
     *
     * @param cunit
     * The compilation unit.
     */
    protected void checkInterface(JCompilationUnit cunit)
    {
        try {
            cunit.checkInterface(this);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }
    }

    /**
     * Checks that interface of a given compilation unit is correct.
     * side effect: increment error number
     *
     * @param cunit
     * The compilation unit.
     */
    protected void checkInitializers(JCompilationUnit cunit)
    {
        try {
            cunit.checkInitializers(this, classes);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }
    }

    /**
     * Checks that body of a given compilation unit is correct.
     * side effect: increment error number
     *
     * @param cunit
     * The compilation unit.
     */
    protected void checkBody(JCompilationUnit cunit)
    {
        long lastTime = System.currentTimeMillis();
        try {
            cunit.checkBody(this, classes);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }

        if (verboseMode()) {
            inform(CompilerMessages.BODY_CHECKED, cunit.getFileName(),
                   new Long(System.currentTimeMillis() - lastTime));
        }
    }

    /**
     * generate the source code of parsed compilation unit
     *
     * @param cunit
     * The compilation unit.
     */
    protected void generateJavaCode(JCompilationUnit cunit)
    {
        long lastTime = System.currentTimeMillis();
    
        cunit.accept(this, options.destination);
        if (verboseMode()) {
            inform(CompilerMessages.JAVA_CODE_GENERATED,
                   cunit.getFileName(),
                   new Long(System.currentTimeMillis() - lastTime));
        }
    }

    /**
     * Reports a trouble (error or warning).
     *
     * @param trouble
     * A description of the trouble to report.
     */
    public void reportTrouble(PositionedError trouble)
    {
        if (trouble instanceof CWarning) {
            if (options.warning != 0 && filterWarning((CWarning)trouble)) {
                inform(trouble);
            }
        }
        else {
            if (trouble.getTokenReference() != TokenReference.NO_REF) {
                inform(trouble);
                errorFound = true;
            }
        }
    }

    protected boolean filterWarning(CWarning warning)
    {
        WarningFilter filter = getFilter();
        int   value = filter.filter(warning);

        switch (value) {
        case WarningFilter.FLT_REJECT:
            return false;
        case WarningFilter.FLT_FORCE:
            return true;
        case WarningFilter.FLT_ACCEPT:
            return warning.getSeverityLevel() <= options.warning;
        default:
            throw new RuntimeException("INTERNAL ERROR");
        }
    }

    protected WarningFilter getFilter()
    {
        if (filter == null) {
            if (options.filter != null) {
                try {
                    filter = (WarningFilter)Class.forName(options.filter)
                        .newInstance();
                }
                catch (Exception e) {
                    inform(KjcMessages.FILTER_NOT_FOUND, options.filter);
                }
            }
            if (filter == null) {
                filter = new DefaultFilter();
            }
        }

        return filter;
    }

    /**
     * Returns true iff comments should be parsed (false if to be skipped)
     */
    public boolean parseComments()
    {
        return options.deprecation || (options.beautify && !options.nowrite);
    }

    /**
     * Returns the classes to generate
     */
    public CSourceClass[] getClasses()
    {
        CSourceClass[] array = new CSourceClass[classes.size()];
        return (CSourceClass[])classes.toArray(array);
    }

}
