/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.optimize;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;

import at.dms.classfile.AccessorContainer;
import at.dms.classfile.AbstractInstructionAccessor;
import at.dms.classfile.ClassFileConstants;
import at.dms.classfile.HandlerInfo;
import at.dms.classfile.Instruction;
import at.dms.classfile.JumpInstruction;
import at.dms.classfile.LineNumberInfo;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.LocalVarInstruction;
import at.dms.classfile.Opcodes;
import at.dms.classfile.SwitchInstruction;

/**
 * 
 */
public class InstructionHandle
    extends AbstractInstructionAccessor
    implements ClassFileConstants, Opcodes
{
    private Instruction instruction;
    private boolean reached;

    private InstructionHandle prev;
    private InstructionHandle next;

    private List lineNumbers;
    private List accessors;

    /**
     * Creates a new instruction handle.
     *
     * @param instruction     the instruction
     * @param prev        the handle of the next instruction
     *                in textual order
     */
    public InstructionHandle(Instruction instruction, InstructionHandle prev)
    {
        this.instruction = instruction;
        this.prev = prev;
        this.next = null;

        if (prev != null) {
            prev.setNext(this);
        }
    }

    /**
     * Notifies this handle that is has been attached to the specified
     * container.
     */
    public void attachTo(AccessorContainer container)
    {
        if (container instanceof LineNumberInfo) {
            LineNumberInfo info = (LineNumberInfo)container;

            if (lineNumbers == null) {
                lineNumbers = new ArrayList();
            }
            lineNumbers.add(new Integer(info.getLine()));
        }
    }

    /**
     * Adds line number info stored for the instruction to the specified list.
     */
    public void addLineNumberInfo(List lineNumberInfo)
    {
        if (lineNumbers != null) {
            for ( Iterator iter = lineNumbers.iterator();
                  iter.hasNext(); ) {
                int line = ((Integer)iter.next()).intValue();

                lineNumberInfo.add(new LineNumberInfo((short)line, this));
            }
        }
    }

    /**
     * Returns the enclosed instruction.
     */
    public Instruction getInstruction()
    {
        return instruction;
    }

    /**
     * Returns the handle of the next instruction in textual order.
     */
    public InstructionHandle getNext()
    {
        return next;
    }

    /**
     * Returns the handle of the next instruction in textual order.
     */
    public InstructionHandle getPrevious()
    {
        return prev;
    }

    /**
     * Returns true if this instruction is a jump instruction.
     */
    public boolean isJump()
    {
        return instruction instanceof JumpInstruction;
    }

    /**
     * Returns the instruction as a Jump instruction.
     */
    public JumpInstruction getJump()
    {
        return (JumpInstruction)instruction;
    }

    /**
     * Returns the instruction as a Jump instruction.
     */
    public LocalVarInstruction getLocal()
    {
        return (LocalVarInstruction)instruction;
    }

    /**
     * Returns the target of this jump instruction.
     */
    public InstructionHandle getTarget()
    {
        return (InstructionHandle)((JumpInstruction)instruction).getTarget();
    }

    /**
     * Returns the target of this jump instruction.
     */
    public void setTarget(InstructionHandle target)
    {
        ((JumpInstruction)instruction).setTarget(target);
    }

    /**
     * Returns the target of this jump instruction.
     */
    public int getOpcode()
    {
        return instruction.getOpcode();
    }

    /**
     * Sets the handle of the next instruction in textual order.
     */
    public void setNext(InstructionHandle next)
    {
        this.next = next;
        if (next != null) {
            next.prev = this;
        }
    }

    /**
     * isReached
     */
    public boolean isReached()
    {
        return reached;
    }

    /**
     * isTarget
     */
    public boolean isTarget()
    {
        return accessors != null;
    }

    /**
     * isReached
     */
    public void set()
    {
        reached = prev == null || 
            accessors != null ||
            (prev.isReached() && prev.getInstruction().canComplete());
    }

    /**
     * destroy
     */
    public void destroy()
    {
        if (accessors != null && getNext() == null) {
            instruction = new NoArgInstruction(opc_nop);
        }
        else {
            if (prev != null) {
                prev.next = next;
            }
            if (next != null) {
                next.prev = prev;
            }
            for (int i = 0; accessors != null && i < accessors.size(); i++) {
                changeTarget(((AccessorContainer)accessors.get(i)), next);
            }
            notifyTargetOnRemove();
        }
    }

    /**
     * destroy
     */
    public boolean remove()
    {
        if (accessors != null && getNext() == null) {
            return false;
        }
        else {
            destroy();

            return true;
        }
    }

    /**
     * reset
     */
    public void reset()
    {
        if (instruction instanceof AccessorContainer) {
            if (instruction instanceof JumpInstruction) {
                ((InstructionHandle)((JumpInstruction)instruction).getTarget())
                    .addAccessor((JumpInstruction)instruction); // !!!
            }
            else if (instruction instanceof SwitchInstruction) {
                SwitchInstruction   current = (SwitchInstruction)instruction;

                for (int i = -1; i < current.getSwitchCount(); i++) {
                    ((InstructionHandle)current.getTarget(i))
                        .addAccessor(current);
                }
            }
            else {
                throw new RuntimeException(">>>>>>>" + instruction);
            }
        }
    }

    /**
     * reset
     */
    public void clean()
    {
        reached = false;
    
        if (accessors != null) {
            accessors = null;
        }
    }

    /**
     * replaceBy
     */
    public void replaceBy(Instruction instruction)
    {
        notifyTargetOnRemove();
        this.instruction = instruction;
        notifyTargetOnAdd();
        set();
    }

    /**
     * replaceBy
     */
    public void replaceBy(Instruction[] instructions)
    {
        replaceBy(instructions[0]);
    
        InstructionHandle   current = this;
        InstructionHandle   next = this.next;

        for (int i = 1; i < instructions.length; i++) {
            current = new InstructionHandle(instructions[i], current);

            current.notifyTargetOnAdd();
        }

        current.next = next;
        if (current.next != null) {
            current.next.prev = current;
        }
    }

    public void addAccessor(AccessorContainer accessor)
    {
        if (accessors == null) {
            accessors = new ArrayList(10);
        }

        accessors.add(accessor);
    }

    public AccessorContainer getAccessor(int i)
    {
        return (AccessorContainer)accessors.get(i);
    }

    public void removeAccessor(AccessorContainer accessor)
    {
        accessors.remove(accessor);
        if (accessors.size() == 0) {
            accessors = null;
        }
    }

    public void removeAccessors()
    {
        if (accessors != null) {
            accessors = null;
        }
    }

    private void notifyTargetOnRemove()
    {
        if (instruction instanceof JumpInstruction) {
            ((InstructionHandle)((JumpInstruction)instruction).getTarget())
                .removeAccessor((AccessorContainer)instruction);
        }
        else if (instruction instanceof SwitchInstruction) {
            SwitchInstruction current = (SwitchInstruction)instruction;

            for (int i = -1; i < current.getSwitchCount(); i++) {
                ((InstructionHandle)(current.getTarget(i)))
                    .removeAccessor((AccessorContainer)instruction);
            }
        }
    }

    private void notifyTargetOnAdd()
    {
        if (instruction instanceof JumpInstruction) {
            ((InstructionHandle)((JumpInstruction)instruction).getTarget())
                .addAccessor((AccessorContainer)instruction);
        }
        else if (instruction instanceof SwitchInstruction) {
            SwitchInstruction current = (SwitchInstruction)instruction;

            for (int i = -1; i < current.getSwitchCount(); i++) {
                ((InstructionHandle)(current.getTarget(i)))
                    .addAccessor((AccessorContainer)instruction);
            }
        }
    }

    private void changeTarget(AccessorContainer container,
                              InstructionHandle target)
    {
        if (container instanceof JumpInstruction) {
            ((JumpInstruction)container).setTarget(target);
            target.addAccessor(container);
        }
        else if (container instanceof SwitchInstruction) {
            SwitchInstruction current = (SwitchInstruction)container;

            for (int i = -1; i < current.getSwitchCount(); i++) {
                if (current.getTarget(i) == this) {
                    current.setTarget(i, target);
                }
            }
            target.addAccessor(container);
        }
        else if (container instanceof HandlerInfo) {
            HandlerInfo   handler = (HandlerInfo)container;

            if (handler.getStart() == this) {
                handler.setStart(target);
            }
            if (handler.getEnd() == this) {
                handler.setEnd(target);
            }
            if (handler.getHandler() == this) {
                handler.setHandler(target);
            }
        }
        else if (container instanceof Optimizer) {
            ((Optimizer)container).setCodeStart(target);
        }
        else {
            throw new RuntimeException("" + container);
        }
    }
}
