/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import at.dms.compiler.UnpositionedError;
import at.dms.compiler.FastStringBuffer;
import at.dms.kjc.CArrayClass;
import at.dms.kjc.CClass;
import at.dms.kjc.CContext;

/**
 * Abstraction of array types in Java.
 * See JLS2 chapter 10 for a description of array types.
 */
public class ArrayType
    extends ReferenceType
{
    /**
     * Virtual class for arrays.
     */
    static final CArrayClass ARRAYCLASS = new CArrayClass();

    /**
     * Base type of this array type.
     */
    private Type baseType;

    /**
     * The number of dimensions of this array type.
     */
    private int dimensions;

    /**
     * Constructs an array type.
     *
     * @param baseType
     * The base type of the array.
     * @param dimensions
     * The number of dimensions of the array.
     */
    public ArrayType(Type baseType, int dimensions)
    {
        super(TID_ARRAY);

        assert(baseType != null);
        assert(dimensions > 0);

        if (baseType.isArrayType()) {
            this.dimensions = dimensions + ((ArrayType)baseType).dimensions;
            this.baseType = ((ArrayType)baseType).baseType;
        }
        else {
            this.dimensions = dimensions;
            this.baseType = baseType;
        }
    }

    /**
     * Transforms this type to a string
     */
    public String toString()
    {
        String res = baseType.toString();
        for (int i = 0; i < dimensions; i++) {
            res += "[]";
        }
        return res;
    }

    /**
     * Returns qualified name.
     */
    public String getQualifiedName()
    {
        return getSignature();
    }

    /**
     * Appends the classfile signature of this type to the <code>buffer</code>
     * parameter.
     */
    public void appendSignature(FastStringBuffer buffer)
    {
        for (int i = 0; i < dimensions; i++) {
            buffer.append('[');
        }
        baseType.appendSignature(buffer);
    }

    /**
     * Returns true, because this is the class for array types.
     *
     * @return true, because this is the class for array types.
     */
    public boolean isArrayType()
    {
        return true;
    }

    /**
     * Returns the base class of this type.
     *
     * @return the base class of this type.
     */
    public Type getBaseType()
    {
        assert(baseType != null);
        assert(!(baseType instanceof ArrayType));
        return baseType;
    }

    /**
     * Returns the type that has one dimension less than this array type.
     *
     * @return
     * The type that has one dimension less than this array type.
     */
    public Type getIndexedType()
    {
        assert(baseType != null);
        assert(!(baseType instanceof ArrayType));
        if (dimensions == 1) {
            return baseType;
        }
        else {
            return new ArrayType(baseType, dimensions - 1);
        }
    }

    /**
     * Returns the number of dimensions of this array type.
     *
     * @return the number of dimensions of this array type.
     */
    public int getDimensions()
    {
        return dimensions;
    }

    /**
     * Returns true if this type is checked and valid.
     *
     * @return true if this type is checked and valid.
     */
    public boolean isResolved()
    {
        return baseType.isResolved();
    }

    /**
     * Returns true if this type is equal to another Type.
     */
    public boolean equalsReferenceType(ReferenceType other)
    {
        if (!other.isArrayType()) {
            return false;
        }
        else {
            ArrayType arr = (ArrayType)other;

            return ( baseType.equalsType(arr.baseType) &&
                     dimensions == arr.dimensions );
        }
    }

    /**
     * Resolve the base type of this array.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void resolveType(CContext context)
        throws UnpositionedError
    {
        if (!isResolved()) {
            baseType.resolveType(context);
        }
    }

    /**
     * Returns the CClass object associated with this type.
     * This method should not be called before the type has been resolved.
     *
     * @return
     * The CClass object associated with this type.
     */
    public CClass getCClass()
    {
        return ARRAYCLASS;
    }

    /**
     * Returns whether this type can be converted to the specified type by
     * assignment conversion. See JLS2 5.2 about assignment conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff the assignment conversion is possible.
     */
    public boolean allowsAssignmentConversion(Type to)
    {
        if (to.isArrayType()) {
            ArrayType toArray = to.toArray();
            Type toBase = to.toArray().getBaseType();

            if (dimensions == toArray.dimensions &&
                baseType.allowsAssignmentConversion(toBase)) {
                // normal folks
                return true;
            }
            else {
                if (dimensions < toArray.getDimensions()) {
                    // int[][] i = new int[]; ???
                    return false;
                }
                if (toBase.isClassOrInterfaceType()) {
                    // May be the unusal case Object[] o = new String[][]...;
                    // Or Serializable[] o = new String[][]...;
                    // Or Clonable[] o = new String[][]...;
                    ClassOrInterfaceType toBaseCorI =
                        toBase.toClassOrInterface();
                    return ARRAYCLASS.descendsFrom(toBaseCorI.getCClass());
                }
                else {
                    return false;
                }
            }
        }
        else if (to.isClassOrInterfaceType()) {
            ClassOrInterfaceType toClassOrInterface = to.toClassOrInterface();
            return ARRAYCLASS.descendsFrom(toClassOrInterface.getCClass());
        }
        else {
            return false;
        }
    }

 
    /**
     * Returns whether this type can be converted to the specified type by
     * casting conversion. See JLS2 5.5 about casting conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff casting conversion is possible.
     */
    public boolean allowsCastingConversion(Type to)
    {
        // test for array first because array types are classes

        if (to.isArrayType()) {
            // optimizes recursive call with base types (indexed types)

            Type toBase = to.toArray().getBaseType();
            int toDimensions = to.toArray().getDimensions();

            if (dimensions == toDimensions) {
                return baseType.allowsCastingConversion(toBase);
            }
            else if (dimensions < toDimensions) {
                Type substitute = new ArrayType(toBase,
                                                toDimensions - dimensions);
                return baseType.allowsCastingConversion(substitute);
            }
            else {
                // dimensions > toDimensions
                Type substitute = new ArrayType(baseType,
                                                dimensions - toDimensions);
                return substitute.allowsCastingConversion(toBase);
            }
        }
        else if (to.isClassOrInterfaceType()) {
            ClassOrInterfaceType toClassOrInterface = to.toClassOrInterface();
            return ARRAYCLASS.descendsFrom(toClassOrInterface.getCClass());
        }
        else {
            return false;
        }
    }
}
