/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import at.dms.classfile.InnerClassInfo;
import at.dms.compiler.UnpositionedError;
import at.dms.kjc.CBadClass;
import at.dms.kjc.CClass;
import at.dms.kjc.CClassContext;
import at.dms.kjc.CContext;
import at.dms.kjc.CTopLevel;
import at.dms.kjc.KjcMessages;

/**
 * ClassOrInterfaceTypes that are known to be fully qualified and are loaded
 * when needed.
 */
public class ClassFileType
    extends ClassOrInterfaceType
{
    /**
     * Whether the type has already been resolved.
     */
    private boolean resolved;

    /**
     * The qualified name.
     */
    private String qualifiedName;

    /**
     * Construct a class type.
     *
     * @param qualifiedName
     * The qualified name of the class, possible qualified.
     */
    ClassFileType(String qualifiedName)
    {
        super(qualifiedName);

        this.qualifiedName = qualifiedName;
    }

    /**
     * ...
     */
    public void setResolved(boolean flag)
    {
        this.resolved = flag;
    }

    /**
     * ...
     */
    public boolean isResolved()
    {
        return resolved;
    }

    /**
     * Returns the class object associated with this type.
     *
     * @return the class object associated with this type.
     */
    public CClass getCClass()
    {
        if (!isResolved()) {
            try {
                resolveType(null);
            }
            catch (UnpositionedError cue) {
                setCClass(new CBadClass(getQualifiedName()));
                setResolved(true);
                return getCClass();
            }
        }

        return super.getCClass();
    }

    /**
     * ...
     */
    public String getQualifiedName()
    {
        return qualifiedName;
    }

    /**
     * Checks if the type is accessible from the specified context, and
     * throws a PositionedError if it is not accessible. If the context
     * is <code>null</code>, or if there is no
     * class context to be reached from the provided context, this method
     * just succeeds.
     */
    public void checkAccess(CContext context)
        throws UnpositionedError
    {
        if (context == null) {
            return;
        }
        CClassContext classContext = context.getClassContext();
        if (classContext == null) {
            /* No class context; just return. */
            return;
        }
        CClass contextClass = classContext.getCClass();
        if (super.getCClass().isAccessible(contextClass)) {
            return;
        }

        throw new UnpositionedError(KjcMessages.CLASS_NOACCESS,
                                    getQualifiedName());
    }


    /**
     * Check that type is valid.
     * Necessary to resolve String into java/lang/String.
     *
     * @param context
     * The context (may be be null.)
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void resolveType(CContext context)
        throws UnpositionedError
    {
        if (!isResolved()) {
            /* Try to look it up at top level */
            if (CTopLevel.hasClassFile(qualifiedName)) {
                setCClassFromName(qualifiedName);
                setResolved(true);
            }
            else {
                throw new UnpositionedError(KjcMessages.TYPE_UNKNOWN,
                                            getQualifiedName());
            }
        }

        if (isResolved()) {
            /* Class has been resolved.
             * Check access to it.
             */
            checkAccess(context);
        }
    }
}
