/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import java.util.HashMap;
import java.util.Map;

import at.dms.classfile.InnerClassInfo;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.FastStringBuffer;
import at.dms.kjc.CBadClass;
import at.dms.kjc.CClass;
import at.dms.kjc.CContext;
import at.dms.kjc.CTopLevel;

/**
 * Abstraction of class and interface types in Java.
 * See JLS2 chapters 8 and 9 for a description of class and inteface types.
 */
public abstract class ClassOrInterfaceType
    extends ReferenceType
{
    /**
     * Constant for an empty aray of ClassOrInterfaceType.
     */
    public static final ClassOrInterfaceType[] EMPTY =
        new ClassOrInterfaceType[0];

    /**
     * Static map of {qualifiedname:String -> ClassOrInterfaceType}
     */
    private static Map typeCache = new HashMap(2000);

    /**
     * The class export that this is the type for.
     */
    private CClass clazz;

    /**
     * Construct a class type that is yet unresolved.
     */
    ClassOrInterfaceType()
    {
        super(TID_CLASS);
    }

    /**
     * Construct a class type that is yet unresolved but of which the
     * exact fully qualified name is known.
     */
    ClassOrInterfaceType(String qualifiedName)
    {
        super(TID_CLASS);

        typeCache.put(qualifiedName, this);
    }

    /**
     * Construct a class type.
     *
     * @param clazz
     * The class that will be represented by this type.
     */
    ClassOrInterfaceType(CClass clazz)
    {
        super(TID_CLASS);

        this.clazz = clazz;
   
        typeCache.put(clazz.getQualifiedName(), this);
    }

    /**
     * Looks up or creates objects for fully qualified names, without
     * any extra attempts at qualification or disambiguation.
     */
    public static ClassOrInterfaceType get(String qualifiedName)
    {
        ClassOrInterfaceType type =
            (ClassOrInterfaceType)typeCache.get(qualifiedName);

        if (type != null) {
            return type;
        }
        type = new ClassFileType(qualifiedName);
        typeCache.put(qualifiedName, type);
        return type;
    }

    /**
     * Looks up or creates a ClassOrInterfaceType object for the not-fully
     * qualified name.
     */
    public static ClassOrInterfaceType lookup(String className)
    {
        if (className.indexOf('/') >= 0) {
            /* The name is qualified. */
            ClassOrInterfaceType type =
                (ClassOrInterfaceType)typeCache.get(className);
            
            if (type == null) {
                type = new TypeNameType(className);
            }
            
            return type;
        }
        else {
            return new TypeNameType(className);
        }
    }

    /**
     * Returns the JVM type signature for this type.
     */
    public String getSignature()
    {
        String name = getQualifiedName();
        StringBuffer buffer = new StringBuffer(name.length() + 2);
        buffer.append('L').append(name).append(';');
        return buffer.toString();
    }

    /**
     * Appends the JVM type signature for this type to a buffer.
     */
    protected void appendSignature(FastStringBuffer buffer)
    {
        buffer.append('L');
        buffer.append(getQualifiedName());
        buffer.append(';');
    }


    /**
     * Sets the CClass export belonging to this ClassOrInterfaceType.
     */
    protected void setCClass(CClass clazz)
    {
        this.clazz = clazz;
        String name = clazz.getQualifiedName();
        if (!typeCache.containsKey(name)) {
            typeCache.put(name, this);
        }
    }

    /**
     * Tries to set the CClass export belonging to this ClassOrInterfaceType
     * by loading it from the top level.
     */
    protected void setCClassFromName(String name)
    {
        setCClass(CTopLevel.loadClass(name));
    }

    /**
     * equals.
     */
    public boolean equalsReferenceType(ReferenceType other)
    {
        if ( other.isClassOrInterfaceType() ) {
            ClassOrInterfaceType otherClass = other.toClassOrInterface();
            return getQualifiedName().equals(otherClass.getQualifiedName());
        }
        else {
            return false;
        }
    }

    /**
     * Transforms this type to a string.
     */
    public String toString()
    {
        return getQualifiedName().replace('/', '.');
        /*
        if (clazz.isInnerClass()) {
            if (clazz.isSynthetic()) {
                return "$" + clazz.getIdent();
            }
            else {
                return clazz.getIdent();
            }
        }
        else {
            return getQualifiedName().replace('/', '.');
        }
        */
    }

    /**
     * @return the short name of this class
     */
    public String getIdent()
    {
        return getCClass().getIdent();
    }

    /**
     * Returns the qualified name of the class of this class type.
     */
    public String getQualifiedName()
    {
        return getCClass().getQualifiedName();
    }

    /**
     * Check if a type is a class type.
     *
     * @return whether this is a class type.
     */
    public boolean isClassOrInterfaceType()
    {
        return true;
    }

    /**
     * Returns the class object associated with this type.
     * If this type was never checked (read from class files)
     * check it!
     *
     * @return
     * The class object associated with this type.
     */
    public CClass getCClass()
    {
        if (!isResolved()) {
            throw new IllegalStateException("resolveType has not yet been " +
                                            "called");
        }
        return clazz;
    }

    /**
     * Returns whether this type can be converted to the specified type by
     * assignment conversion. See JLS2 5.2 about assignment conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff the assignment conversion is possible.
     */
    public boolean allowsAssignmentConversion(Type to)
    {
        return ( to.isClassOrInterfaceType() &&
                 !to.isArrayType() &&
                 getCClass().descendsFrom(to.toClassOrInterface().getCClass())
                 );
    }

   
    /**
     * Returns whether this type can be converted to the specified type by
     * casting conversion. See JLS2 5.5 about casting conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff casting conversion is possible.
     */
    public boolean allowsCastingConversion(Type to)
    {
        CClass clazz = getCClass();

        // test for array first because array types are classes

        if (to.isArrayType()) {
            return ArrayType.ARRAYCLASS.descendsFrom(clazz);
        }
        else if (! to.isClassOrInterfaceType() ) {
            return false;
        }

        if ( clazz.isInterface() ) {
            CClass toClazz = to.toClassOrInterface().getCClass();
            if (toClazz.isInterface()) {
                // if T is an interface type and if T and S contain methods
                // with the same signature (JLS 8.4.2) but different return
                // types, then a compile-time error occurs.
                //!!! graf 000512: FIXME: implemt this test
                return true;
            }
            else if (! toClazz.isFinal() ) {
                return true;
            }
            else {
                return toClazz.descendsFrom(clazz);
            }
        }
        else {
            // this is a class type
            CClass toClazz = to.toClassOrInterface().getCClass();
            if (toClazz.isInterface()) {
                if (! clazz.isFinal()) {
                    return true;
                }
                else {
                    return clazz.descendsFrom(toClazz);
                }
            }
            else {
                return ( clazz.descendsFrom(toClazz) ||
                         toClazz.descendsFrom(clazz) );
            }
        }
    }

    /**
     * Whether this is an exception that should be declared.
     */
    public boolean isCheckedException()
    {
        if (!allowsAssignmentConversion(StandardType.EXCEPTION)) {
            return false;
        }
        else if ( allowsAssignmentConversion(StandardType.RUNTIMEEXCEPTION) ||
                  allowsAssignmentConversion(StandardType.ERROR) ) {
            return false;
        }
        else {
            return true;
        }
    }

    /**
     * Implements Object.hashCode()
     */
    public int hashCode()
    {
        return getQualifiedName().hashCode();
    }
}
