/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import at.dms.compiler.FastStringBuffer;
import at.dms.kjc.CContext;
import at.dms.kjc.CClass;

/**
 * Abstraction of 'null' reference type in the type structure.
 * This is not really a type according to the JLS2; it is only used
 * internally as the type of the 'null' literal.
 */
final public class NullType
    extends ReferenceType
{
    /**
     * Construct a null reference type.
     */
    NullType()
    {
        super(TID_NULL);
    }

    /**
     * Returns whether this type has been checked yet. null types don't
     * need checking; they are they same in any context. So they're always
     * checked.
     */
    public boolean isResolved()
    {
        return true;
    }

    /**
     * Returns qualified name.
     */
    public String getQualifiedName()
    {
        return "null";
    }

    /**
     * Resolve type; involves no actions for null type.
     */
    public void resolveType(CContext context)
    {
    }

    /**
     * Returns reference type equality.
     */
    public boolean equalsReferenceType(ReferenceType other)
    {
        return other.getTypeID() == TID_NULL;
    }

    /**
     * Returns the CClass object associated with this type.
     * This method should not be called before the type has been resolved.
     *
     * @return
     * The CClass object associated with this type.
     */
    public CClass getCClass()
    {
        return StandardType.OBJECT.getCClass();
    }

    /**
     * Returns whether this type can be converted to the specified type by
     * assignment conversion. See JLS2 5.2 about assignment conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff the assignment conversion is possible.
     */
    public boolean allowsAssignmentConversion(Type to)
    {
        return to.isReference();
    }

    /**
     * Returns whether this type can be converted to the specified type by
     * casting conversion. See JLS2 5.5 about casting conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff casting conversion is possible.
     */
    public boolean allowsCastingConversion(Type to)
    {
        return to.isReference();
    }

    /**
     * Returns a string that closely resembles how this type would appear
     * in java source code.
     */
    public String toString()
    {
        return "nulltype";
    }

    /**
     * Appends the JVM type signature for this type to a buffer.
     */
    protected void appendSignature(FastStringBuffer buff)
    {
        throw new RuntimeException("Null types don't have signatures.");
    }
}
