/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import at.dms.kjc.CContext;

/**
 * Abstraction of numeric types in Java.
 * See JLS2 4.2 and its subsections for a description of numeric types.
 *
 * Numeric types are: byte, short, int, long, float, double
 */
public abstract class NumericType
    extends PrimitiveType
{
    /**
     * Constructor.
     *
     * @param typeID
     * The ident (int value) of this type.
     */
    protected NumericType(int typeID)
    {
        super(typeID);
        assert(typeID == TID_BYTE ||
               typeID == TID_CHAR ||
               typeID == TID_DOUBLE ||
               typeID == TID_FLOAT ||
               typeID == TID_INT ||
               typeID == TID_LONG ||
               typeID == TID_SHORT);
    }

    /**
     * Returns whether this is a numeric type.
     */
    public boolean isNumeric()
    {
        return true;
    }

    /**
     * Returns whether this type can be converted to the specified type by
     * assignment conversion. See JLS2 5.2 about assignment conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff the assignment conversion is possible.
     */
    public boolean allowsAssignmentConversion(Type to)
    {
        /* Each type is assignable to itself. */
        if (equalsType(to)) {
            return true;
        }
        if (to.isNumeric()) {
            int fromTypeID = getTypeID();
            int toTypeID = to.getTypeID();

            /* non-const byte (signed) is not assignable to a char (unsigned)
             */
            if (fromTypeID == TID_BYTE && toTypeID == TID_CHAR) {
                return false;
            }
            /* non-const short and char are not assignable to eachother.
             * (JLS 5.1.2)
             */
            if (fromTypeID == TID_SHORT || toTypeID == TID_CHAR) {
                return toTypeID >= TID_INT;
            }
            else {
                return fromTypeID < toTypeID;
            }
        }
        else {
            return false;
        }
    }

    /**
     * Returns whether this type can be converted to the specified type by
     * casting conversion. See JLS2 5.5 about casting conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff casting conversion is possible.
     */
    public boolean allowsCastingConversion(Type to)
    {
        return to.isNumeric();
    }

    /**
     * Returns the unary promotion of this type.
     * This is done according to the rules defined in JLS2 section 5.6.1,
     * titled "Unary Numeric Promotion."
     *
     * @return
     * The unary promotion of this type.
     */
    public NumericType unaryPromote()
    {
        if ( this == StandardType.BYTE || this == StandardType.SHORT ||
             this == StandardType.CHAR ) {

            return StandardType.INTEGER;
        }
        else {
            return this;
        }
    }

    /**
     * Returns the type to which both operands in a binary expression must be
     * promoted <i>before</i> the operation is performed. This must only
     * be used for binary epxressions that are required by the JLS to perform
     * binary numeric promotion (such as the multiplicative operations).
     * The resulting type of the expression is also the promoted type.
     * Binary Numeric Promotion is described in JLS 2nd edition,
     * paragraph 5.6.2.
     *
     * @param left
     * The type of the left expression in a binary expression.
     * @param right
     * The type of the right expression in a binary expression.
     * @return
     * The promoted type.
     */
    public NumericType binaryPromoteWith(NumericType with)
    {
        if (this == StandardType.DOUBLE || with == StandardType.DOUBLE) {
            return StandardType.DOUBLE;
        }
        if (this == StandardType.FLOAT || with == StandardType.FLOAT) {
            return StandardType.FLOAT;
        }
        if (this == StandardType.LONG || with == StandardType.LONG) {
            return StandardType.LONG;
        }
        else {
            return StandardType.INTEGER;
        }
    }
}
