/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import at.dms.compiler.FastStringBuffer;
import at.dms.kjc.CContext;

/**
 * Abstraction of primitive types in Java.
 * See JLS2 4.2 for a description of primitive types.
 */
public abstract class PrimitiveType
    extends Type
{
    /**
     * Constructor.
     *
     * @param typeID
     * The ident (int value) of this type.
     */
    protected PrimitiveType(int typeID)
    {
        super(typeID);
        assert(typeID == TID_BOOLEAN ||
               typeID == TID_BYTE ||
               typeID == TID_CHAR ||
               typeID == TID_DOUBLE ||
               typeID == TID_FLOAT ||
               typeID == TID_INT ||
               typeID == TID_LONG ||
               typeID == TID_SHORT ||
               typeID == TID_VOID);
    }

    /**
     * Returns a string that closely resembles how this type would appear
     * in java source code.
     */
    public String toString()
    {
        switch (getTypeID()) {
        case TID_BOOLEAN:
            return "boolean";
        case TID_BYTE:
            return "byte";
        case TID_CHAR:
            return "char";
        case TID_DOUBLE:
            return "double";
        case TID_FLOAT:
            return "float";
        case TID_INT:
            return "int";
        case TID_LONG:
            return "long";
        case TID_SHORT:
            return "short";
        case TID_VOID:
            return "void";
        default:
            throw new IllegalStateException("Bad type ID: " + getTypeID());
        }
    }

    /**
     * Returns whether this type has been checked yet. Primitive types don't
     * need checking; they are they same in any context. So they're always
     * checked.
     */
    public boolean isResolved()
    {
        return true;
    }

    /**
     * Resolve type.
     */
    public void resolveType(CContext context)
    {
    }

    /**
     * Returns the JVM type signature for this type.
     */
    public String getSignature()
    {
        return String.valueOf(getSignatureChar());
    }

    /**
     * Returns whether this type is a primitive type.
     */
    public boolean isPrimitive()
    {
        return true;
    }

    /**
     * Appends the JVM type signature for this type to a buffer.
     */
    protected void appendSignature(FastStringBuffer buff)
    {
        buff.append(getSignatureChar());
    }

    /**
     * Returns the single character JVM type signature for this type.
     */
    private char getSignatureChar()
    {
        switch (getTypeID()) {
        case TID_BOOLEAN:
            return 'Z';
        case TID_BYTE:
            return 'B';
        case TID_CHAR:
            return 'C';
        case TID_DOUBLE:
            return 'D';
        case TID_FLOAT:
            return 'F';
        case TID_INT:
            return 'I';
        case TID_LONG:
            return 'J';
        case TID_SHORT:
            return 'S';
        case TID_VOID:
            return 'V';
        default:
            throw new IllegalStateException("Bad type ID: " + getTypeID());
        }
    }

}
