/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import java.util.ArrayList;
import java.util.List;

import at.dms.compiler.UnpositionedError;
import at.dms.compiler.FastStringBuffer;
import at.dms.kjc.CContext;
import at.dms.util.Utils;

/**
 * Root for type abstraction hierarchy.
 */
public abstract class Type
    extends Utils
{
    /**
     * Type constant for 'void' type.
     */
    public final static int TID_VOID = 1;

    /**
     * Type constant for 'byte' type.
     */
    public final static int TID_BYTE = 2;

    /**
     * Type constant for 'short' type.
     */
    public final static int TID_SHORT = 3;

    /**
     * Type constant for 'char' type.
     */
    public final static int TID_CHAR = 4;

    /**
     * Type constant for 'int' type.
     */
    public final static int TID_INT = 5;

    /**
     * Type constant for 'long' type.
     */
    public final static int TID_LONG = 6;

    /**
     * Type constant for 'float' type.
     */
    public final static int TID_FLOAT = 7;

    /**
     * Type constant for 'double' type.
     */
    public final static int TID_DOUBLE  = 8;

    /**
     * Type constant for a ClassOrInterface type.
     */
    public final static int TID_CLASS = 9;

    /**
     * Type constant for all arrays.
     */
    public final static int TID_ARRAY = 10;

    /**
     * Type constant for 'boolean' type.
     */
    public final static int TID_BOOLEAN = 11;

    /**
     * Type constant for 'null' type.
     */
    public final static int TID_NULL = 12;

    /**
     * Type identifier.
     */
    private int typeID;

    /**
     * Constructs a type baseclass with the general type identifier
     * <code>typeID</code>.
     */
    Type(int typeID)
    {
        this.typeID = typeID;
    }

    /**
     * Returns a string that closely resembles how this type would appear
     * in java source code.
     */
    public abstract String toString();

    /**
     * Returns the JVM type signature for this type.
     */
    public String getSignature()
    {
        FastStringBuffer buffer = new FastStringBuffer();

        appendSignature(buffer);
        return buffer.toString();
    }

    /**
     * Appends the JVM type signature for this type to a buffer.
     */
    protected abstract void appendSignature(FastStringBuffer buff);

    /**
     * Returns whether this type is a numeric type.
     */
    public boolean isNumeric()
    {
        return false;
    }

    /**
     * If <code>isNumeric</code> returns true, then this object, cast to
     * NumericType, is returned.
     *
     * @throws ClassCastException
     * If <code>isNumeric</code> returns false.
     */
    public NumericType toNumeric()
    {
        return (NumericType)this;
    }

    /**
     * Returns whether this is an integral type.
     */
    public boolean isIntegral()
    {
        return false;
    }

    /**
     * If <code>isIntegral</code> returns true, then this object, cast to
     * IntegralType, is returned.
     *
     * @throws ClassCastException
     * If <code>isIntegral</code> returns false.
     */
    public IntegralType toIntegral()
    {
        return (IntegralType)this;
    }

    /**
     * Returns whether this type is a floating point type.
     */
    public boolean isFloatingPoint()
    {
        return false;
    }

    /**
     * If <code>isFloatingPoint</code> returns true, then this object, cast to
     * FloatingPointType, is returned.
     *
     * @throws ClassCastException
     * If <code>isFloatingPoint</code> returns false.
     */
    public FloatingPointType toFloatingPoint()
    {
        return (FloatingPointType)this;
    }

    /**
     * Returns whether this type is a primitive type.
     */
    public boolean isPrimitive()
    {
        return false;
    }

    /**
     * If <code>isPrimitive</code> returns true, then this object, cast to
     * PrimitiveType, is returned.
     *
     * @throws ClassCastException
     * If <code>isPrimitive</code> returns false.
     */
    public PrimitiveType toPrimitive()
    {
        return (PrimitiveType)this;
    }

    /**
     * Returns whether this type is a reference type.
     */
    public boolean isReference()
    {
        return false;
    }

    /**
     * If <code>isReference</code> returns true, then this object, cast to
     * ReferenceType, is returned.
     *
     * @throws ClassCastException
     * If <code>isReference</code> returns false.
     */
    public ReferenceType toReference()
    {
        return (ReferenceType)this;
    }

    /**
     * Returns whether this type is a ClassOrInterface type.
     */
    public boolean isClassOrInterfaceType()
    {
        return false;
    }

    /**
     * If <code>isClassOrInterface</code> returns true, then this object,
     * cast to ClassOrInterfaceType, is returned.
     *
     * @throws ClassCastException
     * If <code>isClassOrInterface</code> returns false.
     */
    public ClassOrInterfaceType toClassOrInterface()
    {
        return (ClassOrInterfaceType)this;
    }

    /**
     * Returns whether this type is an array type.
     */
    public boolean isArrayType()
    {
        return false;
    }

    /**
     * If <code>isArray</code> returns true, then this object, cast to
     * ArrayType, is returned.
     *
     * @throws ClassCastException
     * If <code>isArray</code> returns false.
     */
    public ArrayType toArray()
    {
        return (ArrayType)this;
    }

    /**
     * Returns whether this type has been checked yet.
     */
    public abstract boolean isResolved();

    /**
     * Returns the internal ID number of this type.
     */
    public final int getTypeID()
    {
        return typeID;
    }

    /**
     * Resolves this type within the specified context.
     *
     * @exception UnpositionedError
     * If the type cannot be found, is inaccessible, if something wrong with
     * the context, or any other sematic error without position information.
     */
    public abstract void resolveType(CContext context)
        throws UnpositionedError;

    /**
     * Default equals for type abstractions.
     */
    public final boolean equals(Object other)
    {
        if (other != null) {
            return equalsType((Type)other);
        }
        else {
            return false;
        }
    }

    /**
     * Default equals for type abstractions.
     */
    public boolean equalsType(Type other)
    {
        return getTypeID() == other.getTypeID();
    }

    /**
     * Returns whether this type can be converted to the specified type by
     * casting conversion. See JLS2 5.5 about casting conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff casting conversion is possible.
     */
    public abstract boolean allowsCastingConversion(Type to);

    /**
     * Returns whether this type can be converted to the specified type by
     * assignment conversion. See JLS2 5.2 about assignment conversion.
     *
     * @param to
     * The destination type.
     * @return
     * true iff the assignment conversion is possible.
     */
    public abstract boolean allowsAssignmentConversion(Type to);

}
