/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import at.dms.classfile.InnerClassInfo;
import at.dms.compiler.UnpositionedError;
import at.dms.kjc.CBadClass;
import at.dms.kjc.CClass;
import at.dms.kjc.CClassContext;
import at.dms.kjc.CContext;
import at.dms.kjc.CTopLevel;
import at.dms.kjc.KjcMessages;

/**
 * ClassOrInterfaceTypes that are referred to by name and that are resolved
 * when needed.
 */
public class TypeNameType
    extends ClassOrInterfaceType
{
    /**
     * Whether the type has already been resolved.
     */
    private boolean resolved;

    /**
     * The unqualified name.
     */
    private String unqualifiedName;

    /**
     * Construct a class type.
     *
     * @param name
     * The qualified name of the class, possible qualified.
     */
    public TypeNameType(String name)
    {
        super();

        this.unqualifiedName = name.intern();
    }

    public void setResolved(boolean flag)
    {
        this.resolved = flag;
    }

    public boolean isResolved()
    {
        return resolved;
    }

    /**
     * Sets the qualified name of the type being looking for.
     */
    public void setUnqualifiedName(String name)
    {
        this.unqualifiedName = name;
    }

    /**
     * Transforms this type to a string
     */
    public String toString()
    {
        String name = getUnqualifiedName();

        return (name != null ? name.replace('/', '.') : super.toString());
    }

    /**
     * Returns the qualified name of this class.
     */
    public String getQualifiedName()
    {
        return ( isResolved()
                 ? super.getQualifiedName() : getUnqualifiedName() );
    }

    /**
     * Returns the unqualified name of this class.
     */
    public String getUnqualifiedName()
    {
        return this.unqualifiedName;
    }

    /**
     * Returns the class object associated with this type
     *
     * If this type was never checked (read from class files)
     * check it!
     *
     * @return the class object associated with this type
     */
    public CClass getCClass()
    {
        if (!isResolved()) {
            try {
                resolveType(null);
            }
            catch (UnpositionedError cue) {
                setCClass(new CBadClass(getUnqualifiedName()));
                setResolved(true);
                return getCClass();
            }
        }

        return super.getCClass();
    }

    /**
     * Try to resolve a name as the name of a member type.
     *
     * @param context
     * The context (may be be null.)
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    private boolean tryResolveInnerType(CContext context)
        throws UnpositionedError
    {
        CClass accessing = null;
       
        if (context != null) {
            CClassContext classContext = context.getClassContext();
            if (classContext != null) {
                accessing = classContext.getCClass();
            }
        }

        String name = getUnqualifiedName();

        int index = name.lastIndexOf("/");
        if (index == -1) {
            /* Definitely not an inner class */
            return false;
        }

        /* Try to resolve as if the last part of the ambiguous name was
         * an inner class with some prefix in front of it.
         */

        String outerName = name.substring(0, index).intern();
        String innerName = name.substring(index + 1).intern();
       
        ClassOrInterfaceType outer = ClassOrInterfaceType.lookup(outerName);
        try {
            outer.resolveType(context);
        }
        catch (UnpositionedError e) {
            return false;
        }
        
        CClass outerClazz = outer.getCClass();
        CClass innerClazz =
            outerClazz.lookupMemberClass(accessing, innerName);
        
        if (innerClazz != null) {
            setCClass(innerClazz);
            return true;
        }
        else {
            return false;
        }
    }

    /**
     * Checks if the type is accessible from the specified context, and
     * throws a PositionedError if it is not accessible. If the context
     * is <code>null</code>, or if there is no
     * class context to be reached from the provided context, this method
     * just succeeds.
     */
    protected void checkAccess(CContext context)
        throws UnpositionedError
    {
        if (context == null) {
            return;
        }
        CClassContext classContext = context.getClassContext();
        if (classContext == null) {
            /* No class context; just return. */
            return;
        }
        CClass contextClass = classContext.getCClass();
        if (super.getCClass().isAccessible(contextClass)) {
            return;
        }

        throw new UnpositionedError(KjcMessages.CLASS_NOACCESS,
                                    getQualifiedName());
    }


    /**
     * Check that type is valid.
     * Necessary to resolve String into java/lang/String.
     *
     * @param context
     * The context (may be be null.)
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void resolveType(CContext context)
        throws UnpositionedError
    {
        if (!isResolved()) {
            if (resolveTypeInternal(context)) {
                setResolved(true);
            }
            else {
                throw new UnpositionedError(KjcMessages.TYPE_UNKNOWN,
                                            getQualifiedName());
            }
        }

        if (isResolved()) {
            /* Class has been resolved.
             * Check access to it.
             */
            checkAccess(context);
        }
    }

    /**
     * Try to resolve class. Returns true if succeeded, false if class name
     * not resolvable.
     *
     * @param context
     * The context (may be null.)
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    protected boolean resolveTypeInternal(CContext context)
        throws UnpositionedError
    {
        /* First try if it could be a member type */
        if (tryResolveInnerType(context)) {
            return true;
        }

        String name = getUnqualifiedName();
        CClass accessing = null;
        if (context != null) {
            CClassContext classContext = context.getClassContext();
            if (classContext != null) {
                accessing = classContext.getCClass();
            }
        }

        if (context != null) {
            /* Look the type up; it may already exist */
            ClassOrInterfaceType type = context.lookupClass(accessing, name);

            if (type != null) {
                type.resolveType(context);
                /* Nice, it exists. */
                setCClass(type.getCClass());
                return true;
            }
        }

        /* Try to look it up at top level */
        if (CTopLevel.hasClassFile(name)) {
            setCClassFromName(name);
            return true;
        }

        return false;
    }

}
