/*
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.type;

import java.util.ArrayList;
import java.util.List;

import at.dms.compiler.FastStringBuffer;
import at.dms.compiler.UnpositionedError;
import at.dms.util.Utils;

/**
 * Parser for VM type and method signatures.
 *
 * Signatures are part of the Java Virtual Machine standard and describe the
 * parameter types and returns types of methods.
 */
public class TypeSignature
    extends Utils
{
    /**
     * Generate a method signature from a list of parameter types and a return
     * type of a method.
     */
    public static String genMethodSignature(Type returnType,
                                            Type[] parameters)
    {
        FastStringBuffer buffer = new FastStringBuffer();

        buffer.append('(');

        for (int i = 0; i < parameters.length; i++) {
            parameters[i].appendSignature(buffer);
        }
        buffer.append(')');

        returnType.appendSignature(buffer);

        return buffer.toString();
    }

    /**
     * Creates a type object from a JVM type signature.
     */
    public static final Type parseSignature(String signature)
        throws UnpositionedError
    {
        return parseSignature(signature, 0, signature.length());
    }

    /**
     * Parse a java type signature from the substring of
     * <code>signature</code> that starts at index <code>from</code> and
     * ends just before <code>to</code>.
     */
    public static Type parseSignature(String signature, int from, int to)
        throws UnpositionedError
    {
        Type type;
        int cursor = from;
        while (signature.charAt(cursor) == '[') {
            cursor++;
        }

        switch (signature.charAt(cursor)) {
        case 'V':
            type = StandardType.VOID;
            break;
        case 'B':
            type = StandardType.BYTE;
            break;
        case 'C':
            type = StandardType.CHAR;
            break;
        case 'D':
            type = StandardType.DOUBLE;
            break;
        case 'F':
            type = StandardType.FLOAT;
            break;
        case 'I':
            type = StandardType.INTEGER;
            break;
        case 'J':
            type = StandardType.LONG;
            break;
        case 'L':
            type = ClassOrInterfaceType.get(signature.substring(cursor + 1,
                                                                to - 1));
            break;
        case 'S':
            type = StandardType.SHORT;
            break;
        case 'Z':
            type = StandardType.BOOLEAN;
            break;
        default:
            fail();
            return null;
        }

        cursor -= from;
        if (cursor > 0) {
            type = new ArrayType(type, cursor);
        }

        return type;
    }

    /**
     * Creates an array of type objects from a JVM method signature.
     * The array consists of each parameter type, followed by the return type.
     * So the return type is the last element of the array. For methods that
     * return void, it is StandardType.VOID.
     */
    public static synchronized Type[] parseMethodSignature(String signature)
        throws UnpositionedError
    {
        List result = new ArrayList();

        // assert(sig.charAt(0) == '(');
        char[] sig = signature.toCharArray();
        int current = 1;

        while (sig[current] != ')') {
            int end = current;
            while (sig[end] == '[') {
                end += 1;
            }
            if (sig[end] != 'L') {
                end += 1;
            }
            else {
                while (sig[end] != ';') {
                    end += 1;
                }
                end += 1;
            }
            result.add(parseSignature(signature, current, end));
            current = end;
        }
        result.add(parseSignature(signature, current + 1, signature.length()));

        // store into array
        Type[] array = new Type[result.size()];

        return (Type[])result.toArray(array);
    }
}
