/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.dis;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.StringTokenizer;

import at.dms.classfile.BadAccessorException;
import at.dms.classfile.ClassInfo;
import at.dms.classfile.ClassPath;
import at.dms.classfile.ClassFileConstants;
import at.dms.classfile.CodeInfo;
import at.dms.classfile.FieldInfo;
import at.dms.classfile.HandlerInfo;
import at.dms.classfile.Instruction;
import at.dms.classfile.LocalVariableInfo;
import at.dms.classfile.Member;
import at.dms.classfile.MethodInfo;
import at.dms.classfile.Opcodes;
import at.dms.compiler.CompilerMessages;
import at.dms.compiler.UnpositionedError;
import at.dms.util.Utils;

/**
 * This class prints the classfile in ksm syntaxt.
 */
public class Disassembler
    implements ClassFileConstants, Constants
{
    /**
     * ...
     */
    private final ClassInfo classInfo;

    /**
     * ...
     */
    private final boolean optionSortMembers;

    /**
     * ...
     */
    private final boolean optionShowStack;

    /**
     * Create a disassembler object from a classinfo.
     */
    private Disassembler(ClassInfo classInfo,
                         boolean optionSortMembers, boolean optionShowStack)
    {
        this.classInfo = classInfo;
        this.optionSortMembers = optionSortMembers;
        this.optionShowStack = optionShowStack;
    }


    /**
     * Disassembles a class file.
     */
    public static void disassemble(String sourceFile,
                                   String destination,
                                   DisOptions options)
        throws UnpositionedError
    {
        ClassInfo classInfo;

        if (sourceFile.endsWith(".class")) {
            try {
                classInfo = new ClassInfo(new DataInputStream
                    (new BufferedInputStream
                        (new FileInputStream(sourceFile), 2048)), false);
            }
            catch (Exception e) {
                e.printStackTrace();
                throw new UnpositionedError(DisMessages.FILE_NOT_FOUND,
                                            new Object[] { sourceFile });
            }
        }
        else {
            classInfo = ClassPath.getClassInfo(sourceFile.replace('.', '/'),
                                               false);
        }

        if (classInfo == null) {
            throw new UnpositionedError(DisMessages.CLASS_NOT_FOUND,
                                        new Object[] { sourceFile });
        }

        writeAssemblerFile(classInfo, destination, options);
    }

    /**
     * Creates a class file from class info.
     */
    private static void writeAssemblerFile(ClassInfo classInfo,
                                           String destination,
                                           DisOptions options)
        throws UnpositionedError
    {
        String[] classPath = Utils.splitQualifiedName(classInfo.getName());

        if (classPath[0] != null) {
            /* the class is part of a package */
            String classDir =
                classPath[0].replace('.', File.separatorChar)
                .replace('/', File.separatorChar);

            if (destination != null) {
                destination += File.separator + classDir;
            }
        }

        if (!options.stdout && options.destination != null) {
            // check that destination exists or else create it
            File destDir = new File(options.destination);

            if ( !destDir.exists() ) {
                destDir.mkdirs();
            }

            if ( !destDir.isDirectory() ) {
                throw new UnpositionedError
                    (CompilerMessages.IO_EXCEPTION,
                     new Object[] { options.destination });
            }
        }

        File outputFile = null;
        try {
            PrintWriter out;

            if (options.stdout) {
                outputFile = null;
                out = new PrintWriter(System.out);
            }
            else {
                outputFile = new File(destination, classPath[1] + ".ksm");
                out = new PrintWriter(new BufferedWriter
                    (new FileWriter(outputFile)));
            }

            Disassembler  dis = new Disassembler(classInfo,
                                                 options.stack,
                                                 options.sorted);
            dis.writeClass(new IndentingWriter(out));

            out.flush();
            out.close();
        }
        catch (IOException e) {
            throw new UnpositionedError
                (CompilerMessages.IO_EXCEPTION,
                 new Object[]{ outputFile.getPath(), e.getMessage() });
        }
    }

    /**
     * Prints the class file
     */
    private void writeClass(IndentingWriter out)
    {
        out.println("/* compiler version: " + classInfo.getMajorVersion()
                    + "." + classInfo.getMinorVersion());
        out.println(" */");

        // Headers
        if (classInfo.getSourceFile() != null) {
            out.println("@source \"" + classInfo.getSourceFile() + "\"");
        }

        // super is always set (and would print "synchronized")
        writeModifiers(out, classInfo.getModifiers() & ~ACC_SUPER);
        if ((classInfo.getModifiers() & ACC_INTERFACE) > 0) {
            out.print("interface ");
        }
        else {
            out.print("class ");
        }
        out.print(convertQualifiedName(classInfo.getName()));

        if (classInfo.getSuperClass() != null) {
            out.print(" extends ");
            out.print(convertQualifiedName(classInfo.getSuperClass()));
        }

        String[] interfaces = classInfo.getInterfaces();
        if (interfaces != null && interfaces.length > 0) {
            out.print(" implements");
            for (int i = 0; i < interfaces.length; i++) {
                out.print((i == 0 ? " " : ", ")
                          + convertQualifiedName(interfaces[i]));
            }
        }
        out.println();
        out.println("{");

        FieldInfo[] fields = classInfo.getFields();
        if (optionSortMembers) {
            sort(fields);
        }
        for (int i = 0; i < fields.length; i++) {
            writeField(out, fields[i]);
        }

        out.println();

        MethodInfo[] methods = classInfo.getMethods();
        if (optionSortMembers) {
            sort(methods);
        }
        for (int i = 0; i < methods.length; i++) {
            writeMethod(out, methods[i]);
            if (i != methods.length - 1) {
                out.println();
            }
        }
        out.println();
        out.println("}");
    }

    /**
     * Prints a field.
     */
    private void writeField(IndentingWriter out, FieldInfo info)
    {
        out.incrementLevel();
        out.println();

        writeModifiers(out, info.getModifiers());
        out.print(convertFieldSignature(info.getSignature()) + " ");
        out.print(info.getName());

        Object  value = info.getConstantValue();
        if (value != null) {
            out.print(" = " + convertLiteral(value));
        }
        out.print(";");
        out.decrementLevel();
    }

    /**
     * Prints a method.
     */
    private void writeMethod(IndentingWriter out, MethodInfo info)
    {
        out.incrementLevel();
        out.println();
        out.println("/**");
        out.println(" * " + info.getName());
        out.println(" *");
        if (info.getCodeInfo() != null) {
            out.println(" * @stack " + info.getCodeInfo().getMaxStack());
            out.println(" * @locals " + info.getCodeInfo().getMaxLocals());
        }
        out.println(" */");

        writeModifiers(out, info.getModifiers());
        String[] type = convertMethodSignature(info.getSignature());
        out.print(type[1]);
        out.print(" ");
        out.print(info.getName());
        out.print(type[0]);

        String[] exceptions = info.getExceptions();
        if (exceptions != null) {
            out.println();
            out.incrementLevel();
            out.print("throws");
            for (int i = 0; i < exceptions.length; i++) {
                out.print((i == 0 ? " " : ", ")
                          + exceptions[i].replace('/', '.'));
            }
            out.decrementLevel();
        }

        if (info.getCodeInfo() == null) {
            out.print(";");
        }
        else {
            out.println();
            out.println("{");
            writeCodeInfo(out, info.getCodeInfo());
            out.println();
            out.print("}");
        }
        out.decrementLevel();
    }

    /**
     * Prints code.
     */
    private void writeCodeInfo(IndentingWriter out, CodeInfo info)
    {
        Instruction[] instructions = info.getInstructions();

        InstructionHandle[] handles =
            new InstructionHandle[instructions.length];

        for (int i = 0; i < handles.length; i++) {
            handles[i] = new InstructionHandle(instructions[i], i);
        }

        try {
            info.transformAccessors(new HandleCreator(instructions, handles));
        }
        catch (BadAccessorException e) {
            /* Eurgh.. */
            throw new RuntimeException(e.getMessage());
        }

        out.incrementLevel();
        for (int i = 0; i < handles.length; i++) {
            handles[i].write(out, optionShowStack);
        }
        writeHandlerInfo(out, info);
        writeLocalVariableInfo(out, info);
        out.decrementLevel();
    }

    private void writeModifiers(IndentingWriter out, int modifiers)
    {
        if ((modifiers & ACC_PUBLIC) != 0) {
            out.print("public ");
        }
        else if ((modifiers & ACC_PROTECTED) != 0) {
            out.print("protected ");
        }
        else if ((modifiers & ACC_PRIVATE) != 0) {
            out.print("private ");
        }

        if ((modifiers & ACC_STATIC) != 0) {
            out.print("static ");
        }
        if ((modifiers & ACC_ABSTRACT) != 0) {
            out.print("abstract ");
        }
        if ((modifiers & ACC_FINAL) != 0) {
            out.print("final ");
        }
        if ((modifiers & ACC_NATIVE) != 0) {
            out.print("native ");
        }
        if ((modifiers & ACC_SYNCHRONIZED) != 0) {
            out.print("synchronized ");
        }
        if ((modifiers & ACC_TRANSIENT) != 0) {
            out.print("transient ");
        }
        if ((modifiers & ACC_VOLATILE) != 0) {
            out.print("volatile ");
        }
    }

    /**
     * Prints exception handlers.
     */
    private void writeHandlerInfo(IndentingWriter out, CodeInfo info)
    {
        HandlerInfo[] handlers = info.getHandlers();

        for (int i = 0; i < handlers.length; i++) {
            out.println();
            out.print("@catch "
                      + ((InstructionHandle)handlers[i].getHandler())
                      .getLabel());
            if (handlers[i].getThrown() != null) {
                out.print(" " + handlers[i].getThrown().replace('/', '.'));
            }
            out.print(" ["
                      + ((InstructionHandle)handlers[i].getStart()).getLabel()
                      + ", "
                      + ((InstructionHandle)handlers[i].getEnd()).getLabel()
                      + "]");
        }
    }

    /**
     * Prints local variables.
     */
    private void writeLocalVariableInfo(IndentingWriter out, CodeInfo info)
    {
        LocalVariableInfo[] localVariables = info.getLocalVariables();

        if (localVariables == null) {
            return;
        }

        for (int i = 0; i < localVariables.length; i++) {
            if (i == 0) {
                out.println();
            }

            out.println
                ("@var " + localVariables[i].getSlot()
                 + ": " + localVariables[i].getName()
                 + " " + convertFieldSignature(localVariables[i].getType())
                 + " [ "
                 + ((InstructionHandle)localVariables[i].getStart()).getLabel()
                 + ", "
                 + ((InstructionHandle)localVariables[i].getEnd()).getLabel()
                 + "]");
        }
    }

    /**
     * Converts a field signature into ksm syntax.
     *
     * @param signature
     * The signature to convert.
     *
     * @return
     * A string in ksm syntax.
     */
    static String convertFieldSignature(String signature)
    {
        int brackets = 0;

        while (signature.charAt(brackets) == '[') {
            brackets++;
        }

        StringBuffer result = new StringBuffer(signature.length());

        switch (signature.charAt(brackets)) {
        case 'Z':
            result.append("boolean");
            break;
        case 'B':
            result.append("byte");
            break;
        case 'C':
            result.append("char");
            break;
        case 'D':
            result.append("double");
            break;
        case 'F':
            result.append("float");
            break;
        case 'I':
            result.append("int");
            break;
        case 'J':
            result.append("long");
            break;
        case 'S':
            result.append("short");
            break;
        case 'V':
            result.append("void");
            break;
        case 'L':
            result.append(signature.substring(brackets + 1,
                                              signature.length() - 1)
                          .replace('/', '.'));
            break;
        default:
            result.append("<UNEXPECTED, ILLEGAL SIGNATURE TYPE: ")
                .append(signature.charAt(brackets));
        }

        for (; brackets > 0; brackets -= 1) {
            result.append("[]");
        }
        
        return result.toString();
    }

    /**
     * Converts a method signature into ksm syntax.
     *
     * @param signature
     * The signature to convert.
     * @return
     * An array of two string in ksm syntax:
     * element 0: the argument types
     * element 1: the return type
     */
    static String[] convertMethodSignature(String signature)
    {
        if (signature.charAt(0) != '(') {
            throw new IllegalArgumentException("Not a method signature:"
                                               + signature);
        }

        StringBuffer method = new StringBuffer();
        method.append("(");

        int sigIndex = 1;

        while (signature.charAt(sigIndex) != ')') {
            int len = 0;
            while (signature.charAt(len + sigIndex) == '[') {
                len += 1;
            }
            if (signature.charAt(len + sigIndex) != 'L') {
                len += 1;
            }
            else {
                while (signature.charAt(len + sigIndex) != ';') {
                    len += 1;
                }
                len += 1;
            }
            method.append(convertFieldSignature(signature.substring(sigIndex,
                                                                    len)));
            if (signature.charAt(len + sigIndex) != ')') {
                method.append(", ");
            }
            sigIndex += len;
        }
        method.append(")");

        return new String[] {
            method.toString(),
            convertFieldSignature(signature.substring(sigIndex + 1))
        };
    }

    /**
     * Converts a qualified identifier into ksm syntax.
     *
     * @param ident
     * The identifier to convert.
     * @return
     * A string in ksm syntax.
     */
    static String convertQualifiedName(String ident)
    {
        StringTokenizer token = new StringTokenizer(ident.replace('/', '.'),
                                                    ".");
        String result;

        result = token.nextToken();
        result = isJavaIdentifier(result) ? result : "\'" + result + "\'";

        while (token.hasMoreElements()) {
            String current = token.nextToken();
            result += "."
                + ( isJavaIdentifier(current)
                    ? current : "\'" + current + "\'");
        }
        return result;
    }

    private static boolean isJavaIdentifier(String ident)
    {
        if (ident.equals("<init>") || ident.equals("<clinit>")) {
            return true;
        }
        else if (!Character.isJavaIdentifierStart(ident.charAt(0))) {
            return false;
        }
        else {
            for (int i = 1; i < ident.length(); i++) {
                if (!Character.isJavaIdentifierPart(ident.charAt(i))) {
                    return false;
                }
            }
            return true;
        }
    }

    /**
     * Converts a literal into ksm syntax.
     */
    static String convertLiteral(Object t)
    {
        if (t instanceof String) {
            return convertStringLiteral((String)t);
        }
        else if (t instanceof Float) {
            return convertFloatLiteral((Float)t);
        }
        else if (t instanceof Double) {
            return convertDoubleLiteral((Double)t);
        }
        else if (t instanceof Long) {
            return convertLongLiteral((Long)t);
        }
        else {
            return t.toString();
        }
    }

    private static String convertDoubleLiteral(Double t)
    {
        if (t.isNaN()) {
            return "<NaN>D";
        }
        else if (t.isInfinite()) {
            return t.doubleValue() > 0 ? "<Inf>D" : "-<Inf>D";
        }
        else {
            return t + "D";
        }
    }

    private static String convertFloatLiteral(Float t)
    {
        if (t.isNaN()) {
            return "<NaN>F";
        }
        else if (t.isInfinite()) {
            return t.floatValue() > 0 ? "<Inf>F" : "-<Inf>F";
        }
        else {
            return t + "F";
        }
    }

    private static String convertLongLiteral(Long t)
    {
        return t + "L";
    }

    private static String convertStringLiteral(String t)
    {
        StringBuffer buffer = new StringBuffer();

        buffer.append("\"");
        for (int i = 0; i < t.length(); i++) {
            switch (t.charAt(i)) {
            case '\t':
                buffer.append("\\t");
                break;
            case '\r':
                buffer.append("\\r");
                break;
            case '\n':
                buffer.append("\\n");
                break;
            case '\"':
                buffer.append("\\\"");
                break;
            case '\\':
                buffer.append("\\\\");
                break;
            case '\'':
                buffer.append("\\'");
                break;
            default:
                buffer.append(t.charAt(i));
            }
        }
        buffer.append("\"");
        return buffer.toString();
    }

    /**
     * Bubble sort an array of FieldInfo wrt the name.
     */
    private static final void sort(Member[] data)
    {
        for (int i = data.length; --i >= 0;) {
            for (int j = 0; j < i; j++) {
                if (data[j].getName().compareTo(data[j + 1].getName()) > 0) {
                    Member    tmp = data[j];
                    data[j] = data[j+1];
                    data[j+1] = tmp;
                }
            }
        }
    }
}
