/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.ikjc;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.Socket;
import java.net.ServerSocket;
import java.util.ArrayList;
import java.util.List;

import at.dms.compiler.Compiler;

/**
 * This class implements the server side of the kjc incremental compiler.
 */
public class IkjcServer
{
    /**
     * Port to listen on.
     */
    private static int port;

    /**
     * ServerSocket that is used to listen for client requests.
     */
    private static ServerSocket serverSocket;

    /**
     * Currently active client request.
     */
    private static Socket clientSocket;

    /**
     * Launch the server.
     */
    public static void main(final String[] args)
    {
        port = 4444;

        parseArguments(args);

        // create the socket
        try {
            if (serverSocket == null) {
                serverSocket = new ServerSocket(port);
            }
        }
        catch (IOException e) {
            System.err.println("Could not listen on port: " + port);
            System.exit(1);
        }

        /* Listen for client compile requests on a seperate thread, then
         * compile, then wait for compile request again, until the universe
         * goes cold (or the user presses control-C or the system reboots.)
         */
        Thread t = new Thread(new ThreadGroup("CompilerServer"),
                              new Runnable() {
                                      public void run() {
                                          while (true) {
                                              try {
                                                  new IkjcServer().run();
                                              }
                                              catch (Exception e) {
                                                  e.printStackTrace();
                                                  try {
                                                      clientSocket.close();
                                                  }
                                                  catch (Exception ef) {}
                                              }
                                          }
                                      }});
        t.setPriority(Thread.MIN_PRIORITY);
        t.start();
    }

    /**
     * Accept connections on compiler server socket and try to do
     * compile-on-request.
     */
    public void run()
        throws Exception
    {
        System.err.println("Server launched.");

        while (true) {
            System.err.println("Waiting for clients...");
            // wait for clients
            try {
                clientSocket = serverSocket.accept();
            }
            catch (IOException e) {
                System.err.println("Accept failed.");
                continue;
            }

            processClientRequest(clientSocket);

            System.err.println("Compilation processed, cleaning up ...");
            clientSocket.close();
            System.gc();
        }
    }

    /** 
     * Processes the request from the client, and if it is a valid request,
     * do a compilation according to that request.
     */
    private void processClientRequest(Socket client)
        throws Exception
    {
        BufferedReader in = new BufferedReader(new InputStreamReader
            (client.getInputStream()));
        PrintWriter out = new PrintWriter(client.getOutputStream(), true);

        System.err.println("Processing compilation...");

        try {
            // read arguments
            List args = new ArrayList();
            String line;
            Compiler compiler;
            String directory;
            String[] arguments;
            boolean success;

            System.err.print("Calling ikjc with arguments: ");

            while ((line = in.readLine()) != null && !line.equals("")) {
                args.add(line);
                System.err.print(line + " ");
            }
            System.err.println();

            if (args.size() <= 2) {
                System.err.println("Error: not enough arguments given.");
                return;
            }

            // get working directory
            directory = (String)args.get(0);
            args.remove(0);

            // get compiler
            String compilerName = (String)args.get(0);
            if (compilerName.equals("kjc")) {
                compiler = new at.dms.kjc.Main();
            }
            else if (compilerName.equals("xkjc")) {
                compiler =
                    (Compiler)Class.forName("at.dms.xkjc.Main").newInstance();
            }
            else if ( compilerName.equals("vkjc")
                      || compilerName.equals("kopic") ) {

                compiler =
                    (Compiler)Class.forName("at.dms.vkjc.Main").newInstance();
            }
            else if (compilerName.equals("dbi")) {
                compiler =
                    (Compiler)Class.forName("at.dms.dbi.Main").newInstance();
            }
            else {
                System.err.println("Error: unknown compiler: " + compilerName);
                return;
            }
            args.remove(0);

            // get compiler arguments
            arguments = (String[])args.toArray(new String[args.size()]);

            // run the compiler
            success = compiler.run(directory, out, arguments);

            System.err.println("compilation ended");

            out.println(IkjcClient.TERM_STRING);
            out.println(success);
        }
        finally {
            in.close();
            out.close();
        }
    }

    /**
     * Parse the argument list and set flags.
     * @return parsing error occur
     */
    private static void parseArguments(String[] args)
    {
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-port")) {
                i += 1;
                if (i == args.length) {
                    System.err.println("argument for option -port is missing");
                    System.exit(1);
                }
                try {
                    port = Integer.parseInt(args[i]);
                }
                catch (Exception e) {
                    System.err.println("Bad port number: " + args[i]);
                }
            }
            else {
                help();
            }
        }
    }

    /**
     * Shows a help message.
     */
    private static void help()
    {
        System.err.println("Usage: IkjcServer [-help] [-port p]");
        System.err.println("    -port p: set the port number of server (int)");
    }
}
