/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.ksm;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import java.util.ArrayList;
import java.util.List;

import at.dms.classfile.AccessorContainer;
import at.dms.classfile.AccessorTransformer;
import at.dms.classfile.BadAccessorException;
import at.dms.classfile.CodeInfo;
import at.dms.classfile.HandlerInfo;
import at.dms.classfile.Instruction;
import at.dms.classfile.InstructionAccessor;
import at.dms.classfile.LocalVariableInfo;
import at.dms.classfile.LineNumberInfo;
import at.dms.classfile.SwitchInstruction;

/**
 * This class represent the definition of a method body.
 */
public class MethodBody
{
    private class ResolvingTransformer
        implements AccessorTransformer
    {
        public InstructionAccessor transform(InstructionAccessor accessor,
                                             AccessorContainer container)
            throws UnresolvableLabelException
        {
            return resolveLabel(((LabelReference)accessor).getName());
        }
    }

    /**
     * ...
     */
    private List instructions = new ArrayList();

    /**
     * ...
     */
    private List handlers = new ArrayList();

    /**
     * ...
     */
    private List lines = new ArrayList();

    /**
     * ...
     */
    private List variables = new ArrayList();

    /**
     * ...
     */
    private Map labels = new HashMap();

    /**
     * To create unique labels.
     */
    private int lineNumberId = 0;

    /**
     * Add an instruction at the end of the instruction list.
     */
    public void addInstruction(Instruction instruction)
    {
        instructions.add(instruction);
    }

    /**
     * Add a label at the end of the instruction list.
     *
     * @exception KsmError
     * An assembler error.
     */
    public void addLabel(String name)
        throws KsmError
    {
        if (labels.containsKey(name)) {
            throw new KsmError(null, KsmMessages.LABEL_REDEFINED, name);
        }

        labels.put(name, new Integer(instructions.size()));
    }

    /**
     * Add an exception handler info.
     */
    public void addHandler(String thrown,
                           LabelReference start,
                           LabelReference end,
                           LabelReference handler)
    {
        handlers.add(new HandlerInfo(start, end, handler, thrown));
    }

    /**
     * Add a line number info.
     */
    public void addLineNumber(int line)
    {
        String  label;

        /* Create a unique label. Make it syntactically illegal so it cannot
         * conflict with labels in source.
         */
        label = "'" + lineNumberId++;
        try {
            addLabel(label);
        }
        catch (KsmError e) {
            throw new RuntimeException("INTERNAL ERROR");
        }
        lines.add(new LineNumberInfo((short)line, new LabelReference(label)));
    }

    /**
     * Add a local variable info.
     */
    public void addVariable(String name,
                            String type,
                            LabelReference start,
                            LabelReference end,
                            short slot)
    {
        variables.add(new LocalVariableInfo(start, end, name, type, slot));
    }

    /**
     * Returns the instruction at specified label.
     *
     * @param name
     * The label name.
     */
    public Instruction resolveLabel(String name)
        throws UnresolvableLabelException
    {
        Integer addr = (Integer)labels.get(name);

        if (addr == null) {
            throw new UnresolvableLabelException(KsmMessages.UNDEFINED_LABEL,
                                                 name);
        }
        if (addr.intValue() >= instructions.size()) {
            throw new UnresolvableLabelException(KsmMessages.NO_CODE_AT_LABEL,
                                                 name);
        }
        return (Instruction)instructions.get(addr.intValue());
    }

    /**
     * Generate a classfile code info, do not optimize.
     *
     * @exception KsmError
     * An assembler error.
     */
    public CodeInfo genCodeInfo()
        throws KsmError
    {
        Instruction[] instructionArray;
        HandlerInfo[] handlerArray;
        LineNumberInfo[]    lineArray;
        LocalVariableInfo[] variableArray;

        instructionArray = (Instruction[])
            instructions.toArray(new Instruction[instructions.size()]);
        handlerArray =
            (HandlerInfo[])handlers.toArray(new HandlerInfo[handlers.size()]);

        lineArray = (LineNumberInfo[])
            lines.toArray(new LineNumberInfo[lines.size()]);
        variableArray = (LocalVariableInfo[])
            variables.toArray(new LocalVariableInfo[variables.size()]);

        CodeInfo codeInfo = new CodeInfo(instructionArray, handlerArray,
                                         lineArray, variableArray);
        try {
            AccessorTransformer transformer = new ResolvingTransformer();
            codeInfo.transformAccessors(transformer);
        }
        catch (UnresolvableLabelException e) {
            throw new KsmError(null, e.getFormattedMessage());
        }
        catch (BadAccessorException e) {
            throw new KsmError(null, KsmMessages.UNEXPECTED_EXCEPTION,
                               e.getMessage());
        }

        //!!! graf 990903 remove asap
        for (int i = 0; i < instructionArray.length; i++) {
            if (instructionArray[i] instanceof SwitchInstruction) {
                ((SwitchInstruction)instructionArray[i]).selectSwitchType();
            }
        }
        //!!! graf 990903 remove asap

        return codeInfo;
    }
}

