/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.msggen;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

/**
 * This class is the entry point for the Message generator.
 */
public class Main
{
    /**
     * The suffix of the filename in which the output is placed.
     */
    private final static String OUTPUTFILE_SUFFIX = "Messages.java";

     /**
     * Short option for: "help"
     */
    private static final String shortopts = "h";

    /**
     * The long options, only one, also "help".
     */
    private static final LongOpt[] longopts =
    { new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h') };

    /**
     * The names of the source files.
     */
    private String[] sourceFiles;

    /**
     * The definition file objects; these are parsed source files.
     */
    private DefinitionFile[] definitions;

    /**
     * Only main can construct Main.
     */
    private Main(String[] args)
    {
        parseArgs(args);

        parseSourceFiles();
        checkIdentifiers();
        buildInterfaceFiles();

        System.exit(0);
    }

    /**
     * Parses the source files into DefinitionFile objects.
     */
    private void parseSourceFiles()
    {
        boolean errorsFound = false;

        try {
            definitions = new DefinitionFile[sourceFiles.length];

            for (int i = 0; i < sourceFiles.length; i++) {
                definitions[i] = DefinitionFile.read(sourceFiles[i]);
            }
        }
        catch (MsggenError e) {
            System.err.println(e.getMessage());
            errorsFound = true;
        }

        if (errorsFound) {
            System.exit(1);
        }
    }


    /**
     * Does some checking on the identifiers in the message definition files.
     */
    private void checkIdentifiers()
    {
        boolean errorsFound = false;

        try {
            for (int i = 0; i < definitions.length; i++) {
                definitions[i].checkIdentifiers(new HashMap());
            }
        }
        catch (MsggenError e) {
            System.err.println(e.getMessage());
            errorsFound = true;
        }

        if (errorsFound) {
            System.exit(1);
        }
    }

    /**
     * Generates the output Java source file containing constants for
     * all the messages in the source files.
     */
    private void buildInterfaceFiles()
    {
        for (int i = 0; i < definitions.length; i++) {
            File outputFile = new File(definitions[i].getPrefix()
                                       + OUTPUTFILE_SUFFIX);
            try {
                PrintWriter out = new PrintWriter(new FileWriter(outputFile));
                
                definitions[i].printFile(out);
                
                out.flush();
                out.close();
            }
            catch (java.io.IOException e) {
                System.err.println("I/O Exception on " + outputFile.getPath()
                                   + ": " + e.getMessage());
                System.exit(1);
            }
        }
    }

    /**
     * Entry point to the assembler
     */
    public static void main(String[] args)
    {
        new Main(args);
    }

    /**
     * Parse command line arguments.
     */
    private void parseArgs(String[] args)
    {
        Getopt g = new Getopt("msggen", args, shortopts, longopts, true);

    _loop_:
        for (;;) {
            switch (g.getopt()) {
            case 'h':
                printUsage();
                System.exit(0);

            case -1:
                break _loop_;
            }
        }

        this.sourceFiles = new String[args.length - g.getOptind()];
        System.arraycopy(args, g.getOptind(),
                         this.sourceFiles, 0, this.sourceFiles.length);
    }

    /**
     * Prints usage of this command.
     */
    private static void printUsage()
    {
        PrintStream err = System.err;

        err.println("at.dms.msggen.Main [-d <directory>] [-V] <file>+");
        err.println();
        err.print("This program is originally part of the KOPI suite ");
        err.println("of Java compilation tools.");
        err.println("For more information on the KOPI suite, see:");
        err.println("http://www.dms.at/kopi");
    }

}
